/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

#include "X.h"
#include "Xprotostr.h"

#include "miscstruct.h"
#include "regionstr.h"
#include "gcstruct.h"
#include "windowstr.h"
#include "pixmapstr.h"
#include "scrnintstr.h"

#include "mi.h"

#include "cfb.h"
#include "cfbmskbits.h"
#include "omronFbFunc.h"
#include "cfbfastblt.h"
#include "cfbplanebit.h"


/***							****
****		    COPYAREA MACROS	 		****
****							***/

/*
**	small area bitblt functions 
*/

/* N to N case (this function use raster operation hardware) */
#ifdef FASTGETBITS
#define	FB_bitblt_small_copy( psrcOrg, pdstOrg, xoffSrc, xoffDst, \
			    srcPlaneOffset, dstPlaneOffset, \
		   	    h, w, alu,  pmask) \
{	register unsigned int *psrc; \
	register unsigned int *pdst; \
    	register unsigned int tmpbits; \
    	register unsigned int tmpmask; \
 \
	maskpartialbits((xoffDst), (w), tmpmask); \
	sx_set_operation_all( alu, tmpmask); \
	while (h--) \
	{   register int pm = pmask; \
	    psrc = psrcOrg; \
	    pdst = pdstOrg; \
	    planeLoop( pm ) \
	    { \
		if(checkPmask(pm)) \
		{ /* check plane mask bit */ \
    		    FASTGETBITS(psrc, xoffSrc, w, tmpbits); \
		    *pdst = SCRLEFT( tmpbits, (PPW - w - xoffDst) ); \
		} \
	    	/* compute next plane address */ \
	    	psrc += srcPlaneOffset; \
	    	pdst += dstPlaneOffset; \
		updatePmask(pm); \
	    } \
	    psrcOrg += widthSrc; \
	    pdstOrg += widthDst; \
	} \
}
#else
#define	FB_bitblt_small_copy( psrcOrg, pdstOrg, xoffSrc, xoffDst, \
			    srcPlaneOffset, dstPlaneOffset, \
		   	    h, w, alu,  pmask) \
{	register unsigned int *psrc; \
	register unsigned int *pdst; \
    	register unsigned int tmpbits; \
    	register unsigned int tmpmask; \
	register int pm = pmask; \
\
	maskpartialbits((xoffDst), (w), tmpmask); \
	sx_set_operation_all( alu, tmpmask); \
\
	if((xoffSrc + w) <= PPW) \
	{ int off = xoffSrc - xoffDst;\
	    if(off > 0) \
	    { \
		while(h--) \
		{   \
		    psrc = psrcOrg; \
		    pdst = pdstOrg; \
		    planeLoop(pm) \
		    { \
			if(checkPmask(pm)) \
			{ /* check plane mask bit */ \
			    *pdst = SCRLEFT( (*psrc), off ); \
			} \
			/* compute next plane address */ \
			psrc += srcPlaneOffset; \
			pdst += dstPlaneOffset; \
			updatePmask(pm); \
		    } \
		    psrcOrg += widthSrc; \
		    pdstOrg += widthDst; \
		} \
	    } \
	    else /* off */ \
	    { \
		while(h--) \
		{   \
		    psrc = psrcOrg; \
		    pdst = pdstOrg; \
		    planeLoop(pm) \
		    { \
			if(checkPmask(pm)) \
			{ /* check plane mask bit */ \
			    *pdst = SCRRIGHT( (*psrc), (-off) ); \
			} \
			/* compute next plane address */ \
			psrc += srcPlaneOffset; \
			pdst += dstPlaneOffset; \
			updatePmask(pm); \
		    } \
		    psrcOrg += widthSrc; \
		    pdstOrg += widthDst; \
		} \
	    } \
	} \
	else \
	{   int off = xoffSrc - xoffDst;\
	    if(off > 0) \
	    { \
		while (h--) \
		{ \
		    psrc = psrcOrg; \
		    pdst = pdstOrg; \
		    planeLoop(pm) \
		    { \
			if(checkPmask(pm)) \
			{ /* check plane mask bit */ \
			    *pdst = SCRLEFT( psrc[0], off ) | \
				    SCRRIGHT( psrc[1], (PPW - off)) ; \
			} \
			/* compute next plane address */ \
			psrc += srcPlaneOffset; \
			pdst += dstPlaneOffset; \
			updatePmask(pm); \
		    } \
		    psrcOrg += widthSrc; \
		    pdstOrg += widthDst; \
		} \
	    } \
	    else /* off */\
	    { \
		while (h--) \
		{ \
		    psrc = psrcOrg; \
		    pdst = pdstOrg; \
		    planeLoop(pm) \
		    { \
			if(checkPmask(pm)) \
			{ /* check plane mask bit */ \
			    *pdst = SCRRIGHT( (*psrc), (-off) ); \
			} \
			/* compute next plane address */ \
			psrc += srcPlaneOffset; \
			pdst += dstPlaneOffset; \
			updatePmask(pm); \
		    } \
		    psrcOrg += widthSrc; \
		    pdstOrg += widthDst; \
		} \
	    } \
	} \
}
#endif

/* Nto1 1toN case */
#ifdef FASTGETBITS
#define	FB_bitblt_small_copy1( psrc, pdst, xoffSrc, xoffDst, \
		   	       h, w, alu) \
{	register unsigned int tmpbits; \
    	register unsigned int tmpmask; \
 \
	maskpartialbits((xoffDst), (w), tmpmask); \
	sx_set_operation_all( alu, tmpmask); \
	while (h--) \
	{   register unsigned int _tmpbits; \
    	    FASTGETBITS(psrc, xoffSrc, w, _tmpbits); \
	    *pdst = SCRLEFT( _tmpbits, (PPW - w - xoffDst) ); \
	    psrc += widthSrc; \
	    pdst += widthDst; \
	} \
}
#else
#define	FB_bitblt_small_copy1( psrc, pdst, xoffSrc, xoffDst, \
		   	       h, w, alu) \
{	register unsigned int tmpbits; \
	register unsigned int tmpmask; \
 \
	maskpartialbits((xoffDst), (w), tmpmask); \
	sx_set_operation_all( alu, tmpmask); \
 \
	if((xoffSrc + w) <= PPW) \
	{ int off = xoffSrc - xoffDst;\
	    if(off > 0) \
	    { \
		while (h--) \
		{ \
		    *pdst = SCRLEFT( (*psrc), off ); \
		    psrc += widthSrc; \
		    pdst += widthDst; \
		} \
	    } \
	    else /* off */ \
	    { \
		while (h--) \
		{ \
		    *pdst = SCRRIGHT( (*psrc), (-off) ); \
		    psrc += widthSrc; \
		    pdst += widthDst; \
		} \
	    } \
	} \
	else \
	{ int off = xoffSrc - xoffDst;\
	    if(off > 0) \
	    { \
		while (h--) \
		{ \
		    *pdst = SCRLEFT( psrc[0], off ) | \
			    SCRRIGHT( psrc[1], (PPW - off)) ; \
		    psrc += widthSrc; \
		    pdst += widthDst; \
		} \
	    } \
	    else /* off */\
	    { \
		while (h--) \
		{ \
		    *pdst = SCRRIGHT( (*psrc), (-off) ); \
		    psrc += widthSrc; \
		    pdst += widthDst; \
		} \
	    } \
	} \
}
#endif

/* */
#define compute_1toN_operation( alu, fg, bg, pm1, pm2, pm3, pm4, pmask) \
{ \
	unsigned long diff = fg ^ bg; \
	unsigned long same = ~diff; \
	pm1 = same & (~fg); \
	pm2 = diff & bg; \
	pm3 = diff & fg; \
	pm4 = same & fg; \
	if(conv_op_table[alu] == GXnoop) \
		pmask &= ~pm1; \
	if(invert_op_table[alu] == GXnoop) \
		pmask &= ~pm2; \
	if(const_src_op_table[alu] == GXnoop) \
		pmask &= ~pm4; \
}

#define	sx_set_1toN_operation(alu, mask, pm1, pm2, pm3, pm4, pmask) \
{ \
	sx_plane_select(pm1); \
	*sx_conv_op_cont0[alu] = mask; \
	sx_plane_select(pm2); \
	*sx_invert_op_cont0[alu] = mask; \
	sx_plane_select(pm3); \
	*sx_op_cont0[alu] = mask; \
	sx_plane_select(pm4); \
	*sx_const_src_op_cont0[alu] = mask; \
	sx_plane_select(pmask); \
}
	

/* 1toN case */
#ifdef FASTGETBITS
#define	FB_bitblt_small_copy8( psrc, pdst, xoffSrc, xoffDst, \
		   	       h, w, alu, pm1, pm2, pm3, pm4, pmask) \
{	register unsigned int tmpbits; \
    	register unsigned int tmpmask; \
	maskpartialbits((xoffDst), (w), tmpmask); \
	sx_set_1toN_operation(alu, tmpmask, pm1, pm2, pm3, pm4, pmask) \
	while (h--) \
	{   register unsigned int _tmpbits; \
    	    FASTGETBITS(psrc, xoffSrc, w, _tmpbits); \
	    *pdst = SCRLEFT( _tmpbits, (PPW - w - xoffDst) ); \
	    psrc += widthSrc; \
	    pdst += widthDst; \
	} \
}
#else
#define	FB_bitblt_small_copy8( psrc, pdst, xoffSrc, xoffDst, \
		   	       h, w, alu, pm1, pm2, pm3, pm4, pmask) \
{	register unsigned int tmpbits; \
	register unsigned int tmpmask; \
 \
	maskpartialbits((xoffDst), (w), tmpmask); \
	sx_set_1toN_operation(alu, tmpmask, pm1, pm2, pm3, pm4, pmask) \
 \
	if((xoffSrc + w) <= PPW) \
	{ int off = xoffSrc - xoffDst;\
	    if(off > 0) \
	    { \
		while (h--) \
		{ \
		    *pdst = SCRLEFT( (*psrc), off ); \
		    psrc += widthSrc; \
		    pdst += widthDst; \
		} \
	    } \
	    else /* off */ \
	    { \
		while (h--) \
		{ \
		    *pdst = SCRRIGHT( (*psrc), (-off) ); \
		    psrc += widthSrc; \
		    pdst += widthDst; \
		} \
	    } \
	} \
	else \
	{ int off = xoffSrc - xoffDst;\
	    if(off > 0) \
	    { \
		while (h--) \
		{ \
		    *pdst = SCRLEFT( psrc[0], off ) | \
			    SCRRIGHT( psrc[1], (PPW - off)) ; \
		    psrc += widthSrc; \
		    pdst += widthDst; \
		} \
	    } \
	    else /* off */\
	    { \
		while (h--) \
		{ \
		    *pdst = SCRRIGHT( (*psrc), (-off) ); \
		    psrc += widthSrc; \
		    pdst += widthDst; \
		} \
	    } \
	} \
}
#endif



/*
**	small area fill functions
*/

/* NtoN case */
#define	bitblt_fill_area(nbox, pbox, pdstBase, widthDst, dstPlaneOffset, \
			 pmask, EQ, LABEL) \
{   unsigned int *pdstLine; /* pointers to line with current src and dst */ \
                            /* following used for looping through a line */ \
    int w, h; \
    register unsigned int *pdst; \
             unsigned int *pdstOrg; \
    register int xoffDst;   \
    register int pm; \
 \
    while(nbox--) \
    { \
	w = pbox->x2 - pbox->x1; \
	h = pbox->y2 - pbox->y1; \
	pdstLine = pdstBase + (pbox->y1 * widthDst); \
	xoffDst = pbox->x1 & PIM; \
 \
	if ((xoffDst) + w <= PPW) /* width < PPW */ \
	{   register int tmpmask; \
	    maskpartialbits((xoffDst), (w), tmpmask); \
	    pdstLine += (pbox->x1 >> PWSH); \
	    while (h--) \
	    { \
		pdst = pdstLine; \
		pm = pmask; \
		planeLoop(pm) \
		{ \
		    if(checkPmask(pm)) \
		    { \
			*pdst EQ tmpmask; \
		    } \
		    updatePmask(pm); \
		    pdst += dstPlaneOffset; \
		} \
		pdstLine += widthDst; \
	    } \
	} \
	else	/* width > PPW */ \
	{   register unsigned int startmask, endmask; \
	    int nlMiddle;	/* whole longwords in dst */ \
    	    register int nl;	/* temp copy of nlMiddle */ \
 \
    	    maskbits(pbox->x1, w, startmask, endmask, nlMiddle); \
 \
	    pdstLine += (pbox->x1 >> PWSH); \
	    while (h--) \
	    { \
		pm = pmask; \
		pdstOrg = pdstLine; \
		planeLoop(pm)  \
		{ \
		    if(checkPmask(pm)) \
		    { \
		        pdst = pdstOrg; \
			if (startmask) \
			{ \
			    *pdst EQ startmask; \
			    pdst++; \
			} \
		    	nl = nlMiddle; \
		    	DuffL(nl, LABEL, *pdst++ EQ ~0;) \
			if (endmask) \
			{ \
		    	    *pdst EQ endmask; \
			} \
		    } \
		    updatePmask(pm); \
		    pdstOrg += dstPlaneOffset; \
		} \
	    	pdstLine += widthDst; \
	    }  /* h */ \
	} /* width */ \
	pbox++; \
    } /* nbox */ \
}

/******** eod of macro define *************/

extern	int InverseAlu[16];

/* DoBitblt() does multiple rectangle moves into the rectangles
   DISCLAIMER:
   this code can be made much faster; this implementation is
designed to be independent of byte/bit order, processor
instruction set, and the like.  it could probably be done
in a similarly device independent way using mask tables instead
of the getbits/putbits macros.  the narrow case (w<PPW) can be
subdivided into a case that crosses word boundaries and one that
doesn't.

   we have to cope with the dircetion on a per band basis,
rather than a per rectangle basis.  moving bottom to top
means we have to invert the order of the bands; moving right
to left requires reversing the order of the rectangles in
each band.

   if src or dst is a window, the points have already been
translated.
*/

/*
**	no operation 
*/

void
cfbDoNoop(pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
}


/*
**	alu == GXcopy
*/

/* NtoN case */
void
cfbDoBitbltCopy (pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int *psrcBase, *pdstBase;	/* start of src and dst bitmaps */
    int widthSrc, widthDst;	/* add to get to same position in next line */
    BoxPtr pbox;
    int nbox;
    BoxPtr pboxNew1, pboxNew2; /* temporaries for shuffling rectangles */
    DDXPointPtr pptNew1, pptNew2; /* shuffling boxes entails shuffling the
				     source points too */
    int w, h;
    int xdir;			/* 1 = left right, -1 = right left/ */
    int ydir;			/* 1 = top down, -1 = bottom up */
    unsigned int *psrcLine, *pdstLine;	
				/* pointers to line with current src and dst */
				/* following used for looping through a line */
    register unsigned int *psrc;
    register unsigned int *pdst;
    register          int  pm = pmask;
    unsigned int *pdstOrg;
    unsigned int *psrcOrg;
    register unsigned int bits;
    register unsigned int bits1;
    register int    xoffSrc, xoffDst;
    int		    leftShift, rightShift;
    register unsigned int startmask, endmask;
    int nlMiddle;		/* whole longwords in dst */
    register int nl;	/* temp copy of nlMiddle */

    int careful;
    int srcPlaneOffset;		/* source drawable plane size */
    int dstPlaneOffset;		/* dstnation drawble plane size */
    OmronFbInfoPtr fbInfo;
    int depth = pDstDrawable->depth;

    initPmask(pmask, pDstDrawable->depth);

    if( ! pmask ) return;

    fbInfo = getOmronFbInfo( pSrcDrawable );

    setBaseAndWidth( pSrcDrawable, psrcBase, widthSrc, srcPlaneOffset );

    if(isCommonPlane(psrcBase,fbInfo))
    {
	/* source memory is frame buffer */
	psrcBase = (unsigned int *)getPlaneAddres(fbInfo);
    }

    setBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    if(isCommonPlane(pdstBase,fbInfo))
    {
	/* dst memory is frame buffer */
	pdstBase = (unsigned int *)getPlaneAddres(fbInfo);
	/* reset raster operation hardware */
	sx_set_operation_all( GXcopy, ~0 );
    }

    /* XXX we have to err on the side of safety when both are windows,
     * because we don't know if IncludeInferiors is being used.
     */
    careful = (psrcBase == pdstBase);

    checkRevSrc(psrcBase, srcPlaneOffset, pSrcDrawable->depth);

    if(depth == 1)
    {
	/* copy from pixmap to bitmap */
	if(copyPlaneBitPlane)	/* called by copyplane */
	{
	    psrcBase += (copyPlaneBitPlane - 1) * srcPlaneOffset;
	    pmask = (pm & 1) ? bitmapPmask() : 0;
	}
    }
    else if(pSrcDrawable->depth == 1)
    {
	/* copy from bitmap to pixmap */
	srcPlaneOffset = 0;
    }

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    checkAndChangeRectBox( careful, pptSrc, pbox, nbox,
                           pboxNew1, pptNew1, pboxNew2, pptNew,
                           ydir, xdir);

    {  
	while(nbox--)
	{
	    w = pbox->x2 - pbox->x1;
	    h = pbox->y2 - pbox->y1;

	    if (ydir == -1) /* start at last scanline of rectangle */
	    {
	        psrcLine = psrcBase + ((pptSrc->y+h-1) * -widthSrc);
	        pdstLine = pdstBase + ((pbox->y2-1) * -widthDst);
	    }
	    else /* start at first scanline */
	    {
	        psrcLine = psrcBase + (pptSrc->y * widthSrc);
    		pdstLine = pdstBase + (pbox->y1 * widthDst);
	    }

	    xoffDst = pbox->x1 & PIM;

	    /* width < PPW */
#if defined(FASTGETBITS) && defined(FASTPUTBITS)
	    if (w <= PPW)
#else
	    if ((xoffDst + w) <= PPW)
#endif
	    {	
#if defined(FASTGETBITS) && defined(FASTPUTBITS)
		w &= PIM;
#else
		maskpartialbits((xoffDst), (w), bits1);	/* bits1 == mask */
#endif
		pdstLine += (pbox->x1 >> PWSH);
		psrcLine += (pptSrc->x >> PWSH);
		xoffSrc = pptSrc->x & PIM;

		while (h--)
		{  
	            pm = pmask;
		    psrc = psrcLine;
		    pdst = pdstLine;
		    psrcLine += widthSrc;
		    pdstLine += widthDst;
		    planeLoop(pm)
		    { 
			if(checkPmask(pm))
			{ 
#if	defined(FASTGETBITS) && defined(FASTPUTBITS)
		   	    FASTGETBITS(psrc, xoffSrc, w, bits);
    			    FASTPUTBITS(bits, xoffDst, w, pdst);
#else
			    getbits(psrc, xoffSrc, w, bits); 
		    	    slow_putbits_port1( bits, xoffDst, 
					        bits1, pdst );
#endif
			} 
		    	/* compute next plane address */ 
		    	psrc += srcPlaneOffset;
		    	pdst += dstPlaneOffset;
			updatePmask(pm);
		    } 
		} 
	    }
	    else	/* width > PPW */
	    {

	    	maskbits(pbox->x1, w, startmask, endmask, nlMiddle);

	    	if (xdir == 1)
	    	{
	    	    xoffSrc = pptSrc->x & PIM;
		    pdstLine += (pbox->x1 >> PWSH);
		    psrcLine += (pptSrc->x >> PWSH);

		    if (xoffSrc == xoffDst)
		    { 
	    	    	while (h--)
			{   
			    pm = pmask;
		    	    psrcOrg = psrcLine;
		    	    pdstOrg = pdstLine;
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
			    planeLoop(pm) 
			    {
				psrc = psrcOrg;
				pdst = pdstOrg;
				if(checkPmask(pm))
				{
				    if (startmask)
				    {
					*pdst = (*pdst & ~startmask) |
					  	(*psrc++ & startmask);
					pdst++;
				    }
			    	    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
			            psrc += nl & (UNROLL-1);
			            pdst += nl & (UNROLL-1);
#define BodyOdd(n) pdst[-n] = psrc[-n];
#define BodyEven(n) pdst[-n] = psrc[-n];
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n)  *pdst++ = *psrc++;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
			    	    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    	    DuffL(nl, label1, *pdst++ = *psrc++;)
#endif
				    if (endmask)
				    {
					*pdst = (*pdst & ~endmask) |
					  	(*psrc++ & endmask);
				    }
				} /* checkPmask() */
				updatePmask(pm);
				psrcOrg += srcPlaneOffset;
				pdstOrg += dstPlaneOffset;
			    } /* endLoop */
			} /* h */
		    }
		    else /* xoffSrc == xoffDst */
		    {
		    	if (xoffSrc > xoffDst)
			{
			    leftShift = (xoffSrc - xoffDst);
			    rightShift = PPW - leftShift;
			}
		    	else
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}

		    	while (h--)
		    	{
			    pm = pmask;
			    psrcOrg = psrcLine;
			    pdstOrg = pdstLine;
			    pdstLine += widthDst;
			    psrcLine += widthSrc;
			    planeLoop(pm)
			    {
				psrc = psrcOrg;
				pdst = pdstOrg;
				if(checkPmask(pm))
				{
				    bits = 0;
			 	    if (xoffSrc > xoffDst)
			    		bits = *psrc++;
				    if (startmask)
				    {
				    	bits1 = SCRLEFT(bits,leftShift);
				    	bits = *psrc++;
				    	bits1 |= SCRRIGHT(bits,rightShift);
				    	*pdst = (*pdst & ~startmask) |
					  	(bits1 & startmask);
				    	pdst++;
				    }
				    nl = nlMiddle;

#ifdef LARGE_INSTRUCTION_CACHE
				    bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
				    psrc += nl & (UNROLL-1);
				    pdst += nl & (UNROLL-1);
#define BodyOdd(n) \
    bits = psrc[-n]; \
    psdt[-n] = BitLeft(bits1, leftShift)|BitRight(bits, rightShift);
#define BodyEven(n) \
    bits1 = psrc[-n]; \
    pdst[-n] = BitLeft(bits, leftShift)|BitRight(bits1, rightShift);
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n) \
    bits = *psrc++; \
    *psdt++ = BitLeft(bits1, leftShift)|BitRight(bits, rightShift);
#define BodyEven(n) \
    bits1 = *psrc++; \
    *pdst++ = BitLeft(bits, leftShift)|BitRight(bits1, rightShift);
#define LoopReset   ;
#endif	/* !FAST_CONSTANT_OFFSET_MODE */
				    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				    DuffL (nl,label2,
					bits1 = BitLeft(bits, leftShift);
					bits = *psrc++;
					*pdst++ = bits1 | BitRight(bits, rightShift);
			    	    )
#endif
				    if (endmask)
				    {
				    	bits1 = SCRLEFT(bits, leftShift);
				    	if (SCRLEFT(endmask, rightShift))
				    	{
					    bits = *psrc++;
					    bits1 |= SCRRIGHT(bits, rightShift);
				    	}
				    	*pdst = (*pdst & ~endmask) |
					  	(bits1 & endmask);
				    }
				} /* checkPmask() */
				updatePmask(pm);
				psrcOrg += srcPlaneOffset;
				pdstOrg += dstPlaneOffset;
			    } /* endLoop */
		    	} /* h */
		    } /* xoffSrc == xoffDst */
	    	}
	    	else	/* xdir == -1 */
	    	{
	    	    xoffSrc = (pptSrc->x + w - 1) & PIM;
	    	    xoffDst = (pbox->x2 - 1) & PIM;
		    pdstLine += ((pbox->x2-1) >> PWSH) + 1;
		    psrcLine += ((pptSrc->x+w - 1) >> PWSH) + 1;
		    if (xoffSrc == xoffDst)
		    {
	    	    	while (h--)
	    	    	{
			    pm = pmask;
			    psrcOrg = psrcLine;
			    pdstOrg = pdstLine;
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
			    planeLoop(pm)
			    {
				psrc = psrcOrg;
				pdst = pdstOrg;
				if(checkPmask(pm))
				{
				    if (endmask)
				    {
				    	pdst--;
				    	*pdst = (*pdst & ~endmask) |
					  	(*--psrc & endmask);
				    }
				    nl = nlMiddle;

#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
				    psrc -= nl & (UNROLL - 1);
				    pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) pdst[n-1] = psrc[n-1];
#define BodyEven(n) pdst[n-1] = psrc[n-1];
#define LoopReset \
    pdst -= UNROLL;\
    psrc -= UNROLL;
#else
#define BodyOdd(n)  *--pdst = *--psrc;
		    
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
				    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				    DuffL(nl,label3, *--pdst = *--psrc;)
#endif
				    if (startmask)
				    {
				    	--pdst;
				    	*pdst = (*pdst & ~startmask) |
						(*--psrc & startmask);
				    }
				} /* checkPmask() */
				updatePmask(pm);
				psrcOrg += srcPlaneOffset;
				pdstOrg += dstPlaneOffset;
			    } /* endLoop */
		    	} /* h */
		    }
		    else /* xoffSrc == xoffDst */
		    {
			if (xoffDst > xoffSrc)
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}
			else
			{
		    	    leftShift = (xoffSrc - xoffDst);
		    	    rightShift = PPW - leftShift;
			}
	    	    	while (h--)
	    	    	{
			    pm = pmask;
			    psrcOrg = psrcLine;
			    pdstOrg = pdstLine;
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
			    planeLoop( pm )
			    {
				psrc = psrcOrg;
				pdst = pdstOrg;
				if(checkPmask(pm))
				{
				    bits = 0;
				    if (xoffDst > xoffSrc)
					bits = *--psrc;
				    if (endmask)
				    {
				    	bits1 = SCRRIGHT(bits, rightShift);
				    	bits = *--psrc;
				    	bits1 |= SCRLEFT(bits, leftShift);
				    	pdst--;
				    	*pdst = ((*pdst) & ~endmask) |
					  	(bits1 & endmask);
				    }
				    nl = nlMiddle;
	
#ifdef LARGE_INSTRUCTION_CACHE
				    bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
				    psrc -= nl & (UNROLL - 1);
				    pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) \
    bits = psrc[n-1]; \
    pdst[n-1] = BitRight(bits1, rightShift) | BitLeft(bits, leftShift);
#define BodyEven(n) \
    bits1 = psrc[n-1]; \
    pdst[n-1] = BitRight(bits, rightShift) | BitLeft(bits1, leftShift);
#define LoopReset \
    pdst -= UNROLL; \
    psrc -= UNROLL;
#else
#define BodyOdd(n) \
    bits = *--psrc;\
    *--pdst = BitRight(bits1, rightShift) | BitLeft(bits, leftShift);
#define BodyEven(n) \
    bits1 = *--psrc;\
    *--pdst =  BitRight(bits, rightShift) | BitLeft(bits1, leftShift);
#define LoopReset   ;
#endif
				    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				    DuffL (nl, label4,
					bits1 = BitRight(bits, rightShift);
					bits = *--psrc;
					*--pdst = bits1 | BitLeft(bits, leftShift);
				    )
#endif
				    if (startmask)
				    {
				    	bits1 = SCRRIGHT(bits, rightShift);
				    	if (SCRRIGHT (startmask, leftShift))
				    	{
					    bits = *--psrc;
					    bits1 |= SCRLEFT(bits, leftShift);
				    	}
					--pdst;
				    	*pdst = (*pdst & ~startmask) |
					   	(bits1 & startmask);
				    }
				} /* checkPmask() */
				updatePmask(pm);
				psrcOrg += srcPlaneOffset;
				pdstOrg += dstPlaneOffset;
			    } /* endLoop */
		    	} /* h */
		    } /* xoffSrc == xoffDst */
	    	} /* xdir */
	    } /* width */
	    pbox++;
	    pptSrc++;
	} /* nbox */
    }

    /* free up stuff */
    if (pboxNew2)
    {
	DEALLOCATE_LOCAL(pptNew2);
	DEALLOCATE_LOCAL(pboxNew2);
    }
    if (pboxNew1)
    {
	DEALLOCATE_LOCAL(pptNew1);
	DEALLOCATE_LOCAL(pboxNew1);
    }
}


/* Nto1 case */
void
cfbDoBitbltCopy1 (pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int *psrcBase, *pdstBase;	/* start of src and dst bitmaps */
    int widthSrc, widthDst;	/* add to get to same position in next line */
    BoxPtr pbox;
    int nbox;
    BoxPtr pboxNew1, pboxNew2; /* temporaries for shuffling rectangles */
    DDXPointPtr pptNew1, pptNew2; /* shuffling boxes entails shuffling the
				     source points too */
    int w, h;
    int xdir;			/* 1 = left right, -1 = right left/ */
    int ydir;			/* 1 = top down, -1 = bottom up */
    unsigned int *psrcLine, *pdstLine;	
				/* pointers to line with current src and dst */
				/* following used for looping through a line */
    register unsigned int *psrc;
    register unsigned int *pdst;
    register unsigned int bits;
    register unsigned int bits1;
    register int    xoffSrc, xoffDst;
    int		    leftShift, rightShift;

    int careful;
    int srcPlaneOffset;		/* source drawable plane size */
    int dstPlaneOffset;		/* dstnation drawble plane size */
    OmronFbInfoPtr fbInfo;
    int depth = pDstDrawable->depth;

    if( copyPlaneBitPlane <= 0 ) return;

    initPmask(pmask, depth);

    if( ! pmask ) return;

    fbInfo = getOmronFbInfo( pSrcDrawable );

    setBaseAndWidth( pSrcDrawable, psrcBase, widthSrc, srcPlaneOffset );

    if(isCommonPlane(psrcBase,fbInfo))
    {
	psrcBase = (unsigned int *)getPlaneAddres(fbInfo);
    }

    setBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    if(isCommonPlane(pdstBase,fbInfo))
    {
	pdstBase = (unsigned int *)getPlaneAddres(fbInfo);
	sx_set_operation_all( GXcopy, ~0 );
    }

    /* XXX we have to err on the side of safety when both are windows,
     * because we don't know if IncludeInferiors is being used.
     */
    careful = (psrcBase == pdstBase);

    checkRevSrc(psrcBase, srcPlaneOffset, pSrcDrawable->depth);

    psrcBase += (copyPlaneBitPlane - 1) * srcPlaneOffset;

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    checkAndChangeRectBox( careful, pptSrc, pbox, nbox,
                           pboxNew1, pptNew1, pboxNew2, pptNew,
                           ydir, xdir);

    {
	while(nbox--)
	{
	    w = pbox->x2 - pbox->x1;
	    h = pbox->y2 - pbox->y1;

	    if (ydir == -1) /* start at last scanline of rectangle */
	    {
	        psrcLine = psrcBase + ((pptSrc->y+h-1) * -widthSrc);
	        pdstLine = pdstBase + ((pbox->y2-1) * -widthDst);
	    }
	    else /* start at first scanline */
	    {
	        psrcLine = psrcBase + (pptSrc->y * widthSrc);
    		pdstLine = pdstBase + (pbox->y1 * widthDst);
	     }

	    xoffDst = pbox->x1 & PIM;

	    /* width < PPW */
#if defined(FASTGETBITS) && defined(FASTPUTBITS)
	    if (w <= PPW)
#else
	    if ((xoffDst + w) <= PPW)
#endif
	    {	
#if	defined(FASTGETBITS) && defined(FASTPUTBITS)
		w &= PIM;
#else
		maskpartialbits((xoffDst), (w), bits1);	/* bits1 == mask */
#endif
		pdstLine += (pbox->x1 >> PWSH);
		psrcLine += (pptSrc->x >> PWSH);
		xoffSrc = pptSrc->x & PIM;

		while (h--)
		{  
		    psrc = psrcLine;
		    pdst = pdstLine;
#if	(defined(FASTGETBITS) && defined(FASTPUTBITS))
	   	    FASTGETBITS(psrc, xoffSrc, w, bits);
    		    FASTPUTBITS(bits, xoffDst, w, pdst);
#else
		    getbits(psrc, xoffSrc, w, bits); 
	    	    slow_putbits_port1( bits, xoffDst, bits1, pdst );
#endif
		    psrcLine += widthSrc;
		    pdstLine += widthDst;
		} 
	    }
	    else	/* width > PPW */
	    {
		unsigned int startmask, endmask;
    		int nlMiddle;		/* whole longwords in dst */
    		register int nl;	/* temp copy of nlMiddle */

	    	maskbits(pbox->x1, w, startmask, endmask, nlMiddle);

	    	if (xdir == 1)
	    	{
	    	    xoffSrc = pptSrc->x & PIM;
		    pdstLine += (pbox->x1 >> PWSH);
		    psrcLine += (pptSrc->x >> PWSH);

		    if (xoffSrc == xoffDst)
		    { 
	    	    	while (h--)
			{   
		    	    psrc = psrcLine;
		    	    pdst = pdstLine;
			    if (startmask)
			    {
				*pdst = ((*pdst) & ~startmask) |
				  	((*psrc++) & startmask);
				pdst++;
			    }
		    	    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
		            psrc += nl & (UNROLL-1);
		            pdst += nl & (UNROLL-1);
#define BodyOdd(n) pdst[-n] = psrc[-n];
#define BodyEven(n) pdst[-n] = psrc[-n];
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n)  *pdst++ = *psrc++;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
		    	    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
		    	    DuffL(nl, label1, *pdst++ = *psrc++;)
#endif
			    if (endmask)
			    {
				*pdst = ((*pdst) & ~endmask) |
					((*psrc++) & endmask);
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
			} /* h */
		    }
		    else /* xoffSrc == xoffDst */
		    {
		    	if (xoffSrc > xoffDst)
			{
			    leftShift = (xoffSrc - xoffDst);
			    rightShift = PPW - leftShift;
			}
		    	else
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}

		    	while (h--)
		    	{
			    psrc = psrcLine;
			    pdst = pdstLine;
			    bits = 0;
		 	    if (xoffSrc > xoffDst)
		    		bits = *psrc++;
			    if (startmask)
			    {
			    	bits1 = SCRLEFT(bits,leftShift);
			    	bits = *psrc++;
			    	bits1 |= SCRRIGHT(bits,rightShift);
			    	*pdst = ((*pdst) & ~startmask) |
				  	(bits1 & startmask);
			    	pdst++;
			    }
			    nl = nlMiddle;

#ifdef LARGE_INSTRUCTION_CACHE
			    bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
			    psrc += nl & (UNROLL-1);
			    pdst += nl & (UNROLL-1);
#define BodyOdd(n) \
    bits = psrc[-n]; \
    psdt[-n] = BitLeft(bits1, leftShift)|BitRight(bits, rightShift);
#define BodyEven(n) \
    bits1 = psrc[-n]; \
    pdst[-n] = BitLeft(bits, leftShift)|BitRight(bits1, rightShift);
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n) \
    bits = *psrc++; \
    *psdt++ = BitLeft(bits1, leftShift)|BitRight(bits, rightShift);
#define BodyEven(n) \
    bits1 = *psrc++; \
    *pdst++ = BitLeft(bits, leftShift)|BitRight(bits1, rightShift);
#define LoopReset   ;
#endif	/* !FAST_CONSTANT_OFFSET_MODE */
			    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    DuffL (nl,label2,
				bits1 = BitLeft(bits, leftShift);
				bits = *psrc++;
				*pdst++ = bits1 | BitRight(bits, rightShift);
		    	    )
#endif
			    if (endmask)
			    {
			    	bits1 = SCRLEFT(bits, leftShift);
			    	if (SCRLEFT(endmask, rightShift))
			    	{
				    bits = *psrc++;
				    bits1 |= SCRRIGHT(bits, rightShift);
			    	}
			    	*pdst = ((*pdst) & ~endmask) |
				  	(bits1 & endmask);
			    }
			    pdstLine += widthDst;
			    psrcLine += widthSrc;
		    	} /* h */
		    } /* xoffSrc == xoffDst */
	    	}
	    	else	/* xdir == -1 */
	    	{
	    	    xoffSrc = (pptSrc->x + w - 1) & PIM;
	    	    xoffDst = (pbox->x2 - 1) & PIM;
		    pdstLine += ((pbox->x2-1) >> PWSH) + 1;
		    psrcLine += ((pptSrc->x+w - 1) >> PWSH) + 1;
		    if (xoffSrc == xoffDst)
		    {
	    	    	while (h--)
	    	    	{
			    psrc = psrcLine;
			    pdst = pdstLine;
			    if (endmask)
			    {
			    	pdst--;
			    	*pdst = ((*pdst) & ~endmask) |
				  	((*--psrc) & endmask);
			    }
			    nl = nlMiddle;

#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
			    psrc -= nl & (UNROLL - 1);
			    pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) pdst[n-1] = psrc[n-1];
#define BodyEven(n) pdst[n-1] = psrc[n-1];
#define LoopReset \
    pdst -= UNROLL;\
    psrc -= UNROLL;
#else
#define BodyOdd(n)  *--pdst = *--psrc;
		    
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
			    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    DuffL(nl,label3, *--pdst = *--psrc;)
#endif
			    if (startmask)
			    {
			    	--pdst;
			    	*pdst = (*pdst & ~startmask) |
					(*--psrc & startmask);
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    }
		    else /* xoffSrc == xoffDst */
		    {
			if (xoffDst > xoffSrc)
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}
			else
			{
		    	    leftShift = (xoffSrc - xoffDst);
		    	    rightShift = PPW - leftShift;
			}
	    	    	while (h--)
	    	    	{
			    psrc = psrcLine;
			    pdst = pdstLine;
			    bits = 0;
			    if (xoffDst > xoffSrc)
				bits = *--psrc;
			    if (endmask)
			    {
			    	bits1 = SCRRIGHT(bits, rightShift);
			    	bits = *--psrc;
			    	bits1 |= SCRLEFT(bits, leftShift);
			    	pdst--;
			    	*pdst = ((*pdst) & ~endmask) |
				  	(bits1 & endmask);
			    }
			    nl = nlMiddle;
	
#ifdef LARGE_INSTRUCTION_CACHE
			    bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
			    psrc -= nl & (UNROLL - 1);
			    pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) \
    bits = psrc[n-1]; \
    pdst[n-1] = BitRight(bits1, rightShift) | BitLeft(bits, leftShift);
#define BodyEven(n) \
    bits1 = psrc[n-1]; \
    pdst[n-1] = BitRight(bits, rightShift) | BitLeft(bits1, leftShift);
#define LoopReset \
    pdst -= UNROLL; \
    psrc -= UNROLL;
#else
#define BodyOdd(n) \
    bits = *--psrc;  pdst--;\
    *pdst = BitRight(bits1, rightShift) | BitLeft(bits, leftShift);
#define BodyEven(n) \
    bits1 = *--psrc;  pdst--;\
    *pdst =  BitRight(bits, rightShift) | BitLeft(bits1, leftShift);
#define LoopReset   ;
#endif
			    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    DuffL (nl, label4,
				bits1 = BitRight(bits, rightShift);
				bits = *--psrc;
				*--pdst = bits1 | BitLeft(bits, leftShift);
			    )
#endif
			    if (startmask)
			    {
			    	bits1 = SCRRIGHT(bits, rightShift);
			    	if (SCRRIGHT (startmask, leftShift))
			    	{
				    bits = *--psrc;
				    bits1 |= SCRLEFT(bits, leftShift);
			    	}
				--pdst;
			    	*pdst = ((*pdst) & ~startmask) |
				   	(bits1 & startmask);
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    } /* xoffSrc == xoffDst */
	    	} /* xdir */
	    } /* width */
	    pbox++;
	    pptSrc++;
	} /* nbox */
    }

    /* free up stuff */
    if (pboxNew2)
    {
	DEALLOCATE_LOCAL(pptNew2);
	DEALLOCATE_LOCAL(pboxNew2);
    }
    if (pboxNew1)
    {
	DEALLOCATE_LOCAL(pptNew1);
	DEALLOCATE_LOCAL(pboxNew1);
    }
}

/* 1toN case */
void
cfbDoBitbltCopy8 (pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int *psrcBase, *pdstBase;	/* start of src and dst bitmaps */
    int widthSrc, widthDst;	/* add to get to same position in next line */
    BoxPtr pbox;
    int nbox;
    BoxPtr pboxNew1, pboxNew2; /* temporaries for shuffling rectangles */
    DDXPointPtr pptNew1, pptNew2; /* shuffling boxes entails shuffling the
				     source points too */
    int w, h;
    int xdir;			/* 1 = left right, -1 = right left/ */
    int ydir;			/* 1 = top down, -1 = bottom up */
    unsigned int *psrcLine, *pdstLine;	
				/* pointers to line with current src and dst */
				/* following used for looping through a line */
    register unsigned int *psrc;
    register unsigned int *pdst;
             unsigned int *psrcOrg;
             unsigned int *pdstOrg;
    register unsigned int bits;
    register unsigned int bits1;
    register int tmp;
    int    xoffSrc, xoffDst;
    register int    i;
    int	   leftShift, rightShift;
    int careful;
    int srcPlaneOffset;		/* source drawable plane size */
    int dstPlaneOffset;		/* dstnation drawble plane size */
    OmronFbInfoPtr fbInfo;
    int depth = pDstDrawable->depth;
    int	inx;

    initPmask(pmask, depth);

    if(! pmask) return;

    fbInfo = getOmronFbInfo( pSrcDrawable );

    setBaseAndWidth( pSrcDrawable, psrcBase, widthSrc, srcPlaneOffset );

    if(isCommonPlane(psrcBase,fbInfo))
    {
	psrcBase = (unsigned int *)getPlaneAddres(fbInfo);
    }

    setBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    if(isCommonPlane(pdstBase,fbInfo))
    {
	pdstBase = (unsigned int *)getPlaneAddres(fbInfo);
	sx_set_operation_all( GXcopy, ~0 );
    }

    /* XXX we have to err on the side of safety when both are windows,
     * because we don't know if IncludeInferiors is being used.
     */
    careful = (psrcBase == pdstBase);

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    checkAndChangeRectBox( careful, pptSrc, pbox, nbox,
                           pboxNew1, pptNew1, pboxNew2, pptNew,
                           ydir, xdir);

    {
	while(nbox--)
	{
	    w = pbox->x2 - pbox->x1;
	    h = pbox->y2 - pbox->y1;

	    if (ydir == -1) /* start at last scanline of rectangle */
	    {
	        psrcLine = psrcBase + ((pptSrc->y+h-1) * -widthSrc);
	        pdstLine = pdstBase + ((pbox->y2-1) * -widthDst);
	    }
	    else /* start at first scanline */
	    {
	        psrcLine = psrcBase + (pptSrc->y * widthSrc);
    		pdstLine = pdstBase + (pbox->y1 * widthDst);
	     }

	    xoffDst = pbox->x1 & PIM;

	    /* width < PPW */
#if defined(FASTGETBITS) && defined(FASTPUTBITS)
	    if (w <= PPW)
#else
	    if ((xoffDst + w) <= PPW)
#endif
	    {	
#if	defined(FASTGETBITS) && defined(FASTPUTBITS)
		w &= PIM;
#else
		maskpartialbits((xoffDst), (w), bits1);	/* bits1 == mask */
#endif
		pdstLine += (pbox->x1 >> PWSH);
		psrcLine += (pptSrc->x >> PWSH);
		xoffSrc = pptSrc->x & PIM;

		while (h--)
		{  
		    psrc = psrcLine;
		    pdst = pdstLine;
#if	(defined(FASTGETBITS) && defined(FASTPUTBITS))
	   	    FASTGETBITS(psrc, xoffSrc, w, bits);
		    for(i=0; i<depth; i++)
		    {
			switch(cfbPlaneCopyType[i])
			{
			case CopySrc:
    		    	    FASTPUTBITS(bits, xoffDst, w, pdst);
			    break;
			case CopyInv:
    		    	    FASTPUTBITS(~bits, xoffDst, w, pdst);
			    break;
			case FillSet:
    		    	    FASTPUTBITS(~0, xoffDst, w, pdst);
			    break;
			case FillClear:
    		    	    FASTPUTBITS(0, xoffDst, w, pdst);
			    break;
			case CopyNoop: ;
			}
			pdst += dstPlaneOffset;
		    }
#else
		    getbits(psrc, xoffSrc, w, bits); 
		    for(i=0; i<depth; i++)
		    {
			switch(cfbPlaneCopyType[i])
			{
			case CopySrc:
	    	    	    slow_putbits_port1( bits, xoffDst, bits1, pdst );
			    break;
			case CopyInv:
	    	    	    slow_putbits_port1( ~bits, xoffDst, bits1, pdst );
			    break;
			case FillSet:
			    *pdst |= bits1;
			    break;
			case FillClear:
			    *pdst &= ~bits1;
			    break;
			case CopyNoop: ;
			}
			pdst += dstPlaneOffset;
		    }
#endif
		    psrcLine += widthSrc;
		    pdstLine += widthDst;
		} 
	    }
	    else	/* width > PPW */
	    {
		unsigned int startmask, endmask;
    		int nlMiddle;		/* whole longwords in dst */
    		register int nl;	/* temp copy of nlMiddle */

	    	maskbits(pbox->x1, w, startmask, endmask, nlMiddle);

	    	if (xdir == 1)
	    	{
	    	    xoffSrc = pptSrc->x & PIM;
		    pdstLine += (pbox->x1 >> PWSH);
		    psrcLine += (pptSrc->x >> PWSH);

		    if (xoffSrc == xoffDst)
		    { 
	    	    	while (h--)
			{   
		    	    psrc = psrcLine;
		    	    pdstOrg = pdstLine;
			    if (startmask)
			    {
				pdst = pdstOrg;
				bits = *psrc++;
			        for(i=0; i<depth; i++)
			        {
				    switch(cfbPlaneCopyType[i])
				    {
				    case CopySrc:
					*pdst = ((*pdst) & ~startmask) |
				  		(bits & startmask);
				        break;
				    case CopyInv:
					*pdst = ((*pdst) & ~startmask) |
				  		((~bits) & startmask);
				        break;
				    case FillSet:
					*pdst |= startmask;
				        break;
				    case FillClear:
					*pdst &= ~startmask;
				        break;
				    case CopyNoop: ;
				    }
				    pdst += dstPlaneOffset;
			        }
				pdstOrg++;
			    }

			    psrcOrg = psrc;
			    if(nlMiddle) 
				for(i=0; i<depth; i++)
			    {
			        bits = 0;
			    	pdst = pdstOrg + (i * dstPlaneOffset);
				psrc = psrcOrg;
				switch(cfbPlaneCopyType[i])
				{
				case CopySrc:
			    	    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
			            psrc += nl & (UNROLL-1);
			            pdst += nl & (UNROLL-1);
#define BodyOdd(n) pdst[-n] = psrc[-n];
#define BodyEven(n) pdst[-n] = psrc[-n];
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n)  *pdst++ = *psrc++;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
			    	    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    	    DuffL(nl, label11, *pdst++ = *psrc++;)
#endif
				    break;
				case CopyInv:
			    	    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
			            psrc += nl & (UNROLL-1);
			            pdst += nl & (UNROLL-1);
#define BodyOdd(n) pdst[-n] = ~ psrc[-n];
#define BodyEven(n) pdst[-n] = ~ psrc[-n];
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n)  *pdst++ = ~ *psrc++;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
			    	    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    	    DuffL(nl, label12, *pdst++ = ~ *psrc++;)
#endif
				    break;
				case FillSet:
				    bits = ~0;

				case FillClear:
			    	    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
			            psrc += nl & (UNROLL-1);
			            pdst += nl & (UNROLL-1);
#define BodyOdd(n) pdst[-n] = bits;
#define BodyEven(n) pdst[-n] = bits;
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n)  *pdst++ = bits;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
			    	    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    	    DuffL(nl, label13, *pdst++ = bits;)
#endif
				    break;
				case CopyNoop: ;
				}
			     }

			    if (endmask)
			    {
				psrc = psrcOrg + nlMiddle;
				pdst = pdstOrg + nlMiddle;
				bits = *psrc;
			        for(i=0; i<depth; i++)
			        {
				    switch(cfbPlaneCopyType[i])
				    {
				    case CopySrc:
					*pdst = ((*pdst) & ~endmask) |
				  		(bits & endmask);
				        break;
				    case CopyInv:
					*pdst = ((*pdst) & ~endmask) |
				  		((~bits) & endmask);
				        break;
				    case FillSet:
					*pdst |= endmask;
				        break;
				    case FillClear:
					*pdst &= ~endmask;
				        break;
				    case CopyNoop: ;
				    }
				    pdst += dstPlaneOffset;
			        }
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
			} /* h */
		    }
		    else /* xoffSrc == xoffDst */
		    {
		    	if (xoffSrc > xoffDst)
			{
			    leftShift = (xoffSrc - xoffDst);
			    rightShift = PPW - leftShift;
			}
		    	else
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}

		    	while (h--)
		    	{
			    psrc = psrcLine;
			    pdstOrg = pdstLine;
			    bits = 0;
		 	    if (xoffSrc > xoffDst)
		    		bits = *psrc++;
			    if (startmask)
			    {
			    	bits1 = SCRLEFT(bits,leftShift);
			    	bits = *psrc++;
			    	bits1 |= SCRRIGHT(bits,rightShift);
				pdst = pdstOrg;
			        for(i=0; i<depth; i++)
			        {
				    switch(cfbPlaneCopyType[i])
				    {
				    case CopySrc:
					*pdst = ((*pdst) & ~startmask) |
			  			(bits1 & startmask);
				        break;
				    case CopyInv:
					*pdst = ((*pdst) & ~startmask) |
			  			((~bits1) & startmask);
				        break;
				    case FillSet:
					*pdst |= startmask;
				        break;
				    case FillClear:
					*pdst &= ~startmask;
				        break;
				    case CopyNoop: ;
				    }
				    pdst += dstPlaneOffset;
				}
			    	pdstOrg++;
			    }

		    	    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
			    bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
			    psrc += nl & (UNROLL-1);
			    pdstOrg += nl & (UNROLL-1);
#define BodyOdd(n) \
    bits = psrc[-n]; \
    tmp = BitLeft(bits1, leftShift)|BitRight(bits, rightShift); \
    pdst = pdstOrg; \
    for(i=0; i<depth; i++) \
    { \
	switch(cfbPlaneCopyType[i]) \
	{ \
	case CopySrc: \
	    psdt[-n] = tmp; \
	    break; \
	case CopyInv: \
	    psdt[-n] = ~tmp; \
	    break; \
	case FillSet: \
	    psdt[-n] = ~0; \
	    break; \
	case FillClear: \
	    psdt[-n] = 0; \
	    break; \
	case CopyNoop: ; \
	} \
	pdst += dstPlaneOffset; \
    }
#define BodyEven(n) \
    bits1 = psrc[-n]; \
    tmp = BitLeft(bits, leftShift)|BitRight(bits1, rightShift);
    pdst = pdstOrg; \
    for(i=0; i<depth; i++) \
    { \
	switch(cfbPlaneCopyType[i]) \
	{ \
	case CopySrc: \
	    psdt[-n] = tmp; \
	    break; \
	case CopyInv: \
	    psdt[-n] = ~tmp; \
	    break; \
	case FillSet: \
	    psdt[-n] = ~0; \
	    break; \
	case FillClear: \
	    psdt[-n] = 0; \
	    break; \
	case CopyNoop: ; \
	} \
	pdst += dstPlaneOffset; \
    }
#define LoopReset \
    pdstOrg += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n) \
    bits = *psrc++; \
    tmp = BitLeft(bits1, leftShift)|BitRight(bits, rightShift); \
    pdst = pdstOrg++; \
    for(i=0; i<depth; i++) \
    { \
	switch(cfbPlaneCopyType[i]) \
	{ \
	case CopySrc: \
	    *psdt = tmp; \
	    break; \
	case CopyInv: \
	    *psdt = ~tmp; \
	    break; \
	case FillSet: \
	    *psdt = ~0; \
	    break; \
	case FillClear: \
	    *psdt = 0; \
	    break; \
	case CopyNoop: ; \
	} \
	pdst += dstPlaneOffset; \
    }
#define BodyEven(n) \
    bits1 = *psrc++; \
    tmp = BitLeft(bits, leftShift)|BitRight(bits1, rightShift); \
    pdst = pdstOrg++; \
    for(i=0; i<depth; i++) \
    { \
	switch(cfbPlaneCopyType[i]) \
	{ \
	case CopySrc: \
	    *psdt = tmp; \
	    break; \
	case CopyInv: \
	    *psdt = ~tmp; \
	    break; \
	case FillSet: \
	    *psdt = ~0; \
	    break; \
	case FillClear: \
	    *psdt = 0; \
	    break; \
	case CopyNoop: ; \
	} \
	pdst += dstPlaneOffset; \
    }
#define LoopReset   ;
#endif	/* !FAST_CONSTANT_OFFSET_MODE */
			    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    DuffL (nl,label2,
				bits1 = BitLeft(bits, leftShift);
				bits = *psrc++;
				pdst = pdstOrg++;
				tmp = bits1 | BitRight(bits, rightShift);
			        for(i=0; i<depth; i++)
			        {
				    switch(cfbPlaneCopyType[i])
				    {
				    case CopySrc:
					*pdst = tmp;
				        break;
				    case CopyInv:
					*pdst = ~tmp;
				        break;
				    case FillSet:
					*pdst = ~0;
				        break;
				    case FillClear:
					*pdst = 0;
				        break;
				    case CopyNoop: ;
				    }
				    pdst += dstPlaneOffset;
				}
	    		    )
#endif
			    if (endmask)
			    {
			    	bits1 = SCRLEFT(bits, leftShift);
			    	if (SCRLEFT(endmask, rightShift))
			    	{
				    bits = *psrc++;
				    bits1 |= SCRRIGHT(bits, rightShift);
			    	}
				pdst = pdstOrg;
			        for(i=0; i<depth; i++)
			        {
				    switch(cfbPlaneCopyType[i])
				    {
				    case CopySrc:
			    		*pdst = ((*pdst) & ~endmask) |
				  		(bits1 & endmask);
				        break;
				    case CopyInv:
			    		*pdst = ((*pdst) & ~endmask) |
				  		((~bits1) & endmask);
				        break;
				    case FillSet:
					*pdst |= endmask;
				        break;
				    case FillClear:
					*pdst &= ~endmask;
				        break;
				    case CopyNoop: ;
				    }
				    pdst += dstPlaneOffset;
				}
			    }
			    pdstLine += widthDst;
			    psrcLine += widthSrc;
		    	} /* h */
		    } /* xoffSrc == xoffDst */
	    	}
	    	else	/* xdir == -1 */
	    	{
	    	    xoffSrc = (pptSrc->x + w - 1) & PIM;
	    	    xoffDst = (pbox->x2 - 1) & PIM;
		    pdstLine += ((pbox->x2-1) >> PWSH) + 1;
		    psrcLine += ((pptSrc->x+w - 1) >> PWSH) + 1;
		    if (xoffSrc == xoffDst)
		    {
	    	    	while (h--)
	    	    	{
			    psrc = psrcLine;
			    pdstOrg = pdstLine;
			    if (endmask)
			    {
			    	pdst = --pdstOrg;
				psrc--;
				bits = *psrc;
			        for(i=0; i<depth; i++)
			        {
				    switch(cfbPlaneCopyType[i])
				    {
				    case CopySrc:
			 	   	*pdst = ((*pdst) & ~endmask) |
					  	(bits & endmask);
				        break;
				    case CopyInv:
			 	   	*pdst = ((*pdst) & ~endmask) |
					  	(~bits & endmask);
				        break;
				    case FillSet:
					*pdst |= endmask;
				        break;
				    case FillClear:
					*pdst &= ~endmask;
				        break;
				    case CopyNoop: ;
				    }
				    pdst += dstPlaneOffset;
				}
			    }

			    psrcOrg = psrc;
				
			    if(nlMiddle)
			    	for(i=0; i<depth; i++)
			    {
				pdst = pdstOrg + (i * dstPlaneOffset);
				psrc = psrcOrg;
				bits = 0;
				switch(cfbPlaneCopyType[i])
				{
				case CopySrc:
				    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
				    psrc -= nl & (UNROLL - 1);
				    pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) pdst[n-1] = psrc[n-1];
#define BodyEven(n) pdst[n-1] = psrc[n-1];
#define LoopReset \
    pdst -= UNROLL;\
    psrc -= UNROLL;
#else
#define BodyOdd(n)  *--pdst = *--psrc;
		    
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
				    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				    DuffL(nl,label31, *--pdst = *--psrc;)
#endif
				    break;
				case CopyInv:
				    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
				    psrc -= nl & (UNROLL - 1);
				    pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) pdst[n-1] = ~psrc[n-1];
#define BodyEven(n) pdst[n-1] = ~psrc[n-1];
#define LoopReset \
    pdst -= UNROLL;\
    psrc -= UNROLL;
#else
#define BodyOdd(n)  *--pdst = ~ *--psrc;
		    
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
				    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				    DuffL(nl,label32, *--pdst = ~ *--psrc;)
#endif
				    break;
				case FillSet:
				    bits = ~0;
				case FillClear:
				    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
				    psrc -= nl & (UNROLL - 1);
				    pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) pdst[n-1] = bits;
#define BodyEven(n) pdst[n-1] = bits;
#define LoopReset \
    pdst -= UNROLL;\
    psrc -= UNROLL;
#else
#define BodyOdd(n)  *--pdst = bits;
		    
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
				    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				    DuffL(nl,label33, *--pdst = bits;)
#endif
				    break;
				case CopyNoop: ;
				}
				pdst += dstPlaneOffset;
			    }
			    if (startmask)
			    {
			    	pdst = pdstOrg - nlMiddle - 1;
			    	psrc = psrcOrg - nlMiddle - 1;
				bits = *--psrc;
			        for(i=0; i<depth; i++)
			        {
				    switch(cfbPlaneCopyType[i])
				    {
				    case CopySrc:
			    		*pdst = (*pdst & ~startmask) |
						(bits & startmask);
				        break;
				    case CopyInv:
			    		*pdst = (*pdst & ~startmask) |
						(~bits & startmask);
				        break;
				    case FillSet:
					*pdst |= startmask;
				        break;
				    case FillClear:
					*pdst &= ~startmask;
				        break;
				    case CopyNoop: ;
				    }
				    pdst += dstPlaneOffset;
				}
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    }
		    else /* xoffSrc == xoffDst */
		    {
			if (xoffDst > xoffSrc)
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}
			else
			{
		    	    leftShift = (xoffSrc - xoffDst);
		    	    rightShift = PPW - leftShift;
			}
	    	    	while (h--)
	    	    	{
			    psrc = psrcLine;
			    pdstOrg = pdstLine;
			    bits = 0;
			    if (xoffDst > xoffSrc)
				bits = *--psrc;
			    if (endmask)
			    {
			    	bits1 = SCRRIGHT(bits, rightShift);
			    	bits = *--psrc;
			    	bits1 |= SCRLEFT(bits, leftShift);
			    	pdst = --pdstOrg;
			        for(i=0; i<depth; i++)
			        {
				    switch(cfbPlaneCopyType[i])
				    {
				    case CopySrc:
			    		*pdst = ((*pdst) & ~endmask) |
				  		(bits1 & endmask);
				        break;
				    case CopyInv:
			    		*pdst = ((*pdst) & ~endmask) |
				  		((~bits1) & endmask);
				        break;
				    case FillSet:
					*pdst |= endmask;
				        break;
				    case FillClear:
					*pdst &= ~endmask;
				        break;
				    case CopyNoop: ;
				    }
				    pdst += dstPlaneOffset;
				}
			    }
			    nl = nlMiddle;
	
#ifdef LARGE_INSTRUCTION_CACHE
			    bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
			    psrc -= nl & (UNROLL - 1);
			    pdstOrg -= nl & (UNROLL - 1);
#define BodyOdd(n) \
    bits = psrc[n-1]; \
    tmp = BitRight(bits1, rightShift) | BitLeft(bits, leftShift); \
    pdst = pdstOrg; \
    for(i=0; i<depth; i++) \
    { \
	switch(cfbPlaneCopyType[i]) \
	{ \
	case CopySrc: \
	    pdst[n-1] = tmp; \
	    break; \
	case CopyInv: \
	    pdst[n-1] = ~tmp; \
	    break; \
	case FillSet: \
	    pdst[n-1] = ~0; \
	    break; \
	case FillClear: \
	    pdst[n-1] = 0; \
	    break; \
	case CopyNoop: ; \
	} \
	pdst += dstPlaneOffset; \
    }
#define BodyEven(n) \
    bits1 = psrc[n-1]; \
    tmp = BitRight(bits, rightShift) | BitLeft(bits1, leftShift); \
    pdst = pdstOrg; \
    for(i=0; i<depth; i++) \
    { \
	switch(cfbPlaneCopyType[i]) \
	{ \
	case CopySrc: \
	    pdst[n-1] = tmp; \
	    break; \
	case CopyInv: \
	    pdst[n-1] = ~tmp; \
	    break; \
	case FillSet: \
	    pdst[n-1] = ~0; \
	    break; \
	case FillClear: \
	    pdst[n-1] = 0; \
	    break; \
	case CopyNoop: ; \
	} \
	pdst += dstPlaneOffset; \
    }
#define LoopReset \
    pdstOrg -= UNROLL; \
    psrc -= UNROLL;
#else
#define BodyOdd(n) \
    bits = *--psrc;  pdst = --pdstOrg; \
    tmp = BitRight(bits1, rightShift) | BitLeft(bits, leftShift); \
    for(i=0; i<depth; i++) \
    { \
	switch(cfbPlaneCopyType[i]) \
	{ \
	case CopySrc: \
	    pdst[n-1] = tmp; \
	    break; \
	case CopyInv: \
	    pdst[n-1] = ~tmp; \
	    break; \
	case FillSet: \
	    pdst[n-1] = ~0; \
	    break; \
	case FillClear: \
	    pdst[n-1] = 0; \
	    break; \
	case CopyNoop: ; \
	} \
	pdst += dstPlaneOffset; \
    }
#define BodyEven(n) \
    bits1 = *--psrc;  pdst = --pdstOrg; \
    tmp =  BitRight(bits, rightShift) | BitLeft(bits1, leftShift); \
    for(i=0; i<depth; i++) \
    { \
	switch(cfbPlaneCopyType[i]) \
	{ \
	case CopySrc: \
	    pdst[n-1] = tmp; \
	    break; \
	case CopyInv: \
	    pdst[n-1] = ~tmp; \
	    break; \
	case FillSet: \
	    pdst[n-1] = ~0; \
	    break; \
	case FillClear: \
	    pdst[n-1] = 0; \
	    break; \
	case CopyNoop: ; \
	} \
	pdst += dstPlaneOffset; \
    }
#define LoopReset   ;
#endif
			    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    DuffL (nl, label4,
				bits1 = BitRight(bits, rightShift);
				bits = *--psrc; pdst = --pdstOrg;
				tmp = bits1 | BitLeft(bits, leftShift);
			        for(i=0; i<depth; i++)
			        {
				    switch(cfbPlaneCopyType[i])
				    {
				    case CopySrc:
					*pdst = tmp;
				        break;
				    case CopyInv:
					*pdst = ~tmp;
				        break;
				    case FillSet:
					*pdst = ~0;
				        break;
				    case FillClear:
					*pdst = 0;
				        break;
				    case CopyNoop: ;
				    }
				    pdst += dstPlaneOffset;
				}
			    )
#endif
			    if (startmask)
			    {
			    	bits1 = SCRRIGHT(bits, rightShift);
			    	if (SCRRIGHT (startmask, leftShift))
			    	{
				    bits = *--psrc;
				    bits1 |= SCRLEFT(bits, leftShift);
			    	}
				pdst = --pdstOrg;
			        for(i=0; i<depth; i++)
			        {
				    switch(cfbPlaneCopyType[i])
				    {
				    case CopySrc:
				    	*pdst = ((*pdst) & ~startmask) |
					   	(bits1 & startmask);
				        break;
				    case CopyInv:
				    	*pdst = ((*pdst) & ~startmask) |
					   	((~bits1) & startmask);
				        break;
				    case FillSet:
					*pdst |= startmask;
				        break;
				    case FillClear:
					*pdst &= ~startmask;
				        break;
				    case CopyNoop: ;
				    }
				    pdst += dstPlaneOffset;
				}
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    } /* xoffSrc == xoffDst */
	    	} /* xdir */
	    } /* width */
	    pbox++;
	    pptSrc++;
	} /* nbox */
    }

    /* free up stuff */
    if (pboxNew2)
    {
	DEALLOCATE_LOCAL(pptNew2);
	DEALLOCATE_LOCAL(pboxNew2);
    }
    if (pboxNew1)
    {
	DEALLOCATE_LOCAL(pptNew1);
	DEALLOCATE_LOCAL(pboxNew1);
    }
}

/* 1toN case (alu != GXcopy) */
void
cfbDoBitbltOther8 (pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int *psrcBase, *pdstBase;	/* start of src and dst bitmaps */
    int widthSrc, widthDst;	/* add to get to same position in next line */
    BoxPtr pbox;
    int nbox;
    BoxPtr pboxNew1, pboxNew2; /* temporaries for shuffling rectangles */
    DDXPointPtr pptNew1, pptNew2; /* shuffling boxes entails shuffling the
				     source points too */
    int w, h;
    int xdir;			/* 1 = left right, -1 = right left/ */
    int ydir;			/* 1 = top down, -1 = bottom up */
    unsigned int *psrcLine, *pdstLine;	
				/* pointers to line with current src and dst */
				/* following used for looping through a line */
    register unsigned int *psrc;
    register unsigned int *pdst;
             unsigned int *psrcOrg;
             unsigned int *pdstOrg;
    register unsigned int bits;
    register unsigned int bits1;
    register unsigned int tmp, tmp1;
    int    xoffSrc, xoffDst;
    register int    i;
    int	   leftShift, rightShift;
    int careful;
    int srcPlaneOffset;		/* source drawable plane size */
    int dstPlaneOffset;		/* dstnation drawble plane size */
    OmronFbInfoPtr fbInfo;
    int depth = pDstDrawable->depth;
    int	inx;
    char  roptype[5];
    char  palu[32];
    

    initPmask(pmask, depth);
    if(! pmask) return;

    /* optim */
    roptype[CopyNoop]  = GXnoop;
    roptype[FillSet]   = mfbReduceRop(alu, 1);
    roptype[FillClear] = mfbReduceRop(alu, 0);
    roptype[CopySrc]   = alu;
    roptype[CopyInv]   = InverseAlu[alu];

    for(i = 0; i < depth; i++ )
    {
	palu[i] = roptype[cfbPlaneCopyType[i]];
    }

    fbInfo = getOmronFbInfo( pSrcDrawable );

    setBaseAndWidth( pSrcDrawable, psrcBase, widthSrc, srcPlaneOffset );

    if(isCommonPlane(psrcBase,fbInfo))
    {
	psrcBase = (unsigned int *)getPlaneAddres(fbInfo);
    }

    setBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    if(isCommonPlane(pdstBase,fbInfo))
    {
	pdstBase = (unsigned int *)getPlaneAddres(fbInfo);
	sx_set_operation_all( GXcopy, ~0 );
    }

    /* XXX we have to err on the side of safety when both are windows,
     * because we don't know if IncludeInferiors is being used.
     */
    careful = (psrcBase == pdstBase);

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    checkAndChangeRectBox( careful, pptSrc, pbox, nbox,
                           pboxNew1, pptNew1, pboxNew2, pptNew,
                           ydir, xdir);

    {
	while(nbox--)
	{
	    w = pbox->x2 - pbox->x1;
	    h = pbox->y2 - pbox->y1;

	    if (ydir == -1) /* start at last scanline of rectangle */
	    {
	        psrcLine = psrcBase + ((pptSrc->y+h-1) * -widthSrc);
	        pdstLine = pdstBase + ((pbox->y2-1) * -widthDst);
	    }
	    else /* start at first scanline */
	    {
	        psrcLine = psrcBase + (pptSrc->y * widthSrc);
    		pdstLine = pdstBase + (pbox->y1 * widthDst);
	     }

	    xoffDst = pbox->x1 & PIM;

	    /* width < PPW */
	    if ((xoffDst + w) <= PPW)
	    {	
		maskpartialbits((xoffDst), (w), bits1);	/* bits1 == mask */
		pdstLine += (pbox->x1 >> PWSH);
		psrcLine += (pptSrc->x >> PWSH);
		xoffSrc = pptSrc->x & PIM;

		while (h--)
		{  
		    psrc = psrcLine;
		    pdst = pdstLine;
		    getbits(psrc, xoffSrc, w, bits); 
		    bits = SCRRIGHT(bits,xoffDst);
		    for(i=0; i<depth; i++)
		    {
			alu = palu[i];
			if(alu != GXnoop)
			{
			    tmp = *pdst;
			    *pdst = (tmp & ~bits1) |
				    (DoRop(alu,bits,tmp) & bits1);
			}
			pdst += dstPlaneOffset;
		    }
		    psrcLine += widthSrc;
		    pdstLine += widthDst;
		} 
	    }
	    else	/* width > PPW */
	    {
		unsigned int startmask, endmask;
    		int nlMiddle;		/* whole longwords in dst */
    		register int nl;	/* temp copy of nlMiddle */

	    	maskbits(pbox->x1, w, startmask, endmask, nlMiddle);

	    	if (xdir == 1)
	    	{
	    	    xoffSrc = pptSrc->x & PIM;
		    pdstLine += (pbox->x1 >> PWSH);
		    psrcLine += (pptSrc->x >> PWSH);

		    if (xoffSrc == xoffDst)
		    { 
	    	    	while (h--)
			{   
		    	    psrc = psrcLine;
		    	    pdstOrg = pdstLine;
			    if (startmask)
			    {
				pdst = pdstOrg;
				bits = *psrc++;
			        for(i=0; i<depth; i++)
			        {
				    alu = palu[i];
				    if(alu != GXnoop)
				    {
					tmp = *pdst;
					*pdst = (tmp & ~startmask) |
				  	     (DoRop(alu,bits,tmp) & startmask);
				    }
				    pdst += dstPlaneOffset;
			        }
				pdstOrg++;
			    }

			    psrcOrg = psrc;
			    if(nlMiddle) 
				for(i=0; i<depth; i++)
			    {
			    	pdst = pdstOrg + (i * dstPlaneOffset);
				psrc = psrcOrg;
				alu = palu[i];
				if(alu != GXnoop)
				{
			    	    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
			            psrc += nl & (UNROLL-1);
			            pdst += nl & (UNROLL-1);
#define BodyOdd(n)  pdst[-n] = DoRop(alu, psrc[-n], pdst[-n]);
#define BodyEven(n) pdst[-n] = DoRop(alu, psrc[-n], pdst[-n]);
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n)  { *pdst = DoRop(alu, *psrc++, *pdst); pdst++; }
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
			    	    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    	    DuffL(nl, label11, 
					  { *pdst = DoRop(alu, *psrc++, *pdst);
					    pdst++; }
				    )
#endif
				} /* alu != GXnoop */
			     }

			    if (endmask)
			    {
				psrc = psrcOrg + nlMiddle;
				pdst = pdstOrg + nlMiddle;
				bits = *psrc;
			        for(i=0; i<depth; i++)
			        {
				    alu = palu[i];
				    if(alu != GXnoop)
				    {
					tmp = *pdst;
					*pdst = (tmp & ~endmask) |
				  	     (DoRop(alu,bits,tmp) & endmask);
				    }
				    pdst += dstPlaneOffset;
			        }
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
			} /* h */
		    }
		    else /* xoffSrc == xoffDst */
		    {
		    	if (xoffSrc > xoffDst)
			{
			    leftShift = (xoffSrc - xoffDst);
			    rightShift = PPW - leftShift;
			}
		    	else
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}

		    	while (h--)
		    	{
			    psrc = psrcLine;
			    pdstOrg = pdstLine;
			    bits = 0;
		 	    if (xoffSrc > xoffDst)
		    		bits = *psrc++;
			    if (startmask)
			    {
			    	bits1 = SCRLEFT(bits,leftShift);
			    	bits = *psrc++;
			    	bits1 |= SCRRIGHT(bits,rightShift);
				pdst = pdstOrg;
			        for(i=0; i<depth; i++)
			        {
				    alu = palu[i];
				    if(alu != GXnoop)
				    {
					tmp = *pdst;
					*pdst = (tmp & ~startmask) |
				  	     (DoRop(alu,bits1,tmp) & startmask);
				    }
				    pdst += dstPlaneOffset;
			        }
			    	pdstOrg++;
			    }

		    	    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
			    bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
			    psrc += nl & (UNROLL-1);
			    pdstOrg += nl & (UNROLL-1);
#define BodyOdd(n) \
    bits = psrc[-n]; \
    tmp = BitLeft(bits1, leftShift)|BitRight(bits, rightShift); \
    pdst = pdstOrg; \
    for(i=0; i<depth; i++) \
    { \
	alu = palu[i]; \
	if(alu != GXnoop) \
	{ \
	    pdst[-n] = DoRop(alu,tmp,pdst[-n]); \
	} \
        pdst += dstPlaneOffset; \
    } 

#define BodyEven(n) \
    bits1 = psrc[-n]; \
    tmp = BitLeft(bits, leftShift)|BitRight(bits1, rightShift);
    pdst = pdstOrg; \
    for(i=0; i<depth; i++) \
    { \
	alu = palu[i]; \
	if(alu != GXnoop) \
	{ \
	    pdst[-n] = DoRop(alu,tmp,pdst[-n]); \
	} \
        pdst += dstPlaneOffset; \
    } 

#define LoopReset \
    pdstOrg += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n) \
    bits = *psrc++; \
    tmp = BitLeft(bits1, leftShift)|BitRight(bits, rightShift); \
    pdst = pdstOrg++; \
    for(i=0; i<depth; i++) \
    { \
	alu = palu[i]; \
	if(alu != GXnoop) \
	{ \
	    *pdst = DoRop(alu,tmp,*pdst); \
	} \
        pdst += dstPlaneOffset; \
    } 

#define BodyEven(n) \
    bits1 = *psrc++; \
    tmp = BitLeft(bits, leftShift)|BitRight(bits1, rightShift); \
    pdst = pdstOrg++; \
    for(i=0; i<depth; i++) \
    { \
	alu = palu[i]; \
	if(alu != GXnoop) \
	{ \
	    *pdst = DoRop(alu,tmp,*pdst); \
	} \
        pdst += dstPlaneOffset; \
    } 

#define LoopReset   ;
#endif	/* !FAST_CONSTANT_OFFSET_MODE */
			    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    DuffL (nl,label2,
				bits1 = BitLeft(bits, leftShift);
				bits = *psrc++;
				pdst = pdstOrg++;
				tmp = bits1 | BitRight(bits, rightShift);
    				for(i=0; i<depth; i++)
    				{
				    alu = palu[i];
				    if(alu != GXnoop)
				    {
	    				*pdst = DoRop(alu,tmp,*pdst);
				    }
				    pdst += dstPlaneOffset;
				}
	    		    )
#endif
			    if (endmask)
			    {
			    	bits1 = SCRLEFT(bits, leftShift);
			    	if (SCRLEFT(endmask, rightShift))
			    	{
				    bits = *psrc++;
				    bits1 |= SCRRIGHT(bits, rightShift);
			    	}
				pdst = pdstOrg;
    				for(i=0; i<depth; i++)
    				{
				    alu = palu[i];
				    if(alu != GXnoop)
				    {
					tmp = *pdst;
					*pdst = (tmp & ~endmask) |
				  	     (DoRop(alu,bits1,tmp) & endmask);
				    }
				    pdst += dstPlaneOffset;
				}
			    }
			    pdstLine += widthDst;
			    psrcLine += widthSrc;
		    	} /* h */
		    } /* xoffSrc == xoffDst */
	    	}
	    	else	/* xdir == -1 */
	    	{
	    	    xoffSrc = (pptSrc->x + w - 1) & PIM;
	    	    xoffDst = (pbox->x2 - 1) & PIM;
		    pdstLine += ((pbox->x2-1) >> PWSH) + 1;
		    psrcLine += ((pptSrc->x+w - 1) >> PWSH) + 1;
		    if (xoffSrc == xoffDst)
		    {
	    	    	while (h--)
	    	    	{
			    psrc = psrcLine;
			    pdstOrg = pdstLine;
			    if (endmask)
			    {
			    	pdst = --pdstOrg;
				psrc--;
				bits = *psrc;
    				for(i=0; i<depth; i++)
    				{
				    alu = palu[i];
				    if(alu != GXnoop)
				    {
					tmp = *pdst;
					*pdst = (tmp & ~endmask) |
				  	     (DoRop(alu,bits,tmp) & endmask);
				    }
				    pdst += dstPlaneOffset;
				}
			    }

			    psrcOrg = psrc;
				
			    if(nlMiddle)
			    	for(i=0; i<depth; i++)
			    {
				pdst = pdstOrg + (i * dstPlaneOffset);
				psrc = psrcOrg;
				bits = 0;
				alu = palu[i];
				if(alu != GXnoop)
				{
				    nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
				    psrc -= nl & (UNROLL - 1);
				    pdst -= nl & (UNROLL - 1);
#define BodyOdd(n)  pdst[n-1] = DoRop(alu, psrc[n-1], pdst[n-1]);
#define BodyEven(n) pdst[n-1] = DoRop(alu, psrc[n-1], pdst[n-1]);
#define LoopReset \
    pdst -= UNROLL;\
    psrc -= UNROLL;
#else
#define BodyOdd(n)  --pdst; *pdst = DoRop(alu, *--psrc, *pdst);
		    
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
				    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				    DuffL(nl,label31, 
					  --pdst; *pdst = DoRop(alu, *--psrc, *pdst);)
#endif
				}
				
			    }
			    if (startmask)
			    {
			    	pdst = pdstOrg - nlMiddle - 1;
			    	psrc = psrcOrg - nlMiddle - 1;
				bits = *--psrc;
    				for(i=0; i<depth; i++)
    				{
				    alu = palu[i];
				    if(alu != GXnoop)
				    {
					tmp = *pdst;
					*pdst = (tmp & ~startmask) |
				  	     (DoRop(alu,bits,tmp) & startmask);
				    }
				    pdst += dstPlaneOffset;
				}
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    }
		    else /* xoffSrc == xoffDst */
		    {
			if (xoffDst > xoffSrc)
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}
			else
			{
		    	    leftShift = (xoffSrc - xoffDst);
		    	    rightShift = PPW - leftShift;
			}
	    	    	while (h--)
	    	    	{
			    psrc = psrcLine;
			    pdstOrg = pdstLine;
			    bits = 0;
			    if (xoffDst > xoffSrc)
				bits = *--psrc;
			    if (endmask)
			    {
			    	bits1 = SCRRIGHT(bits, rightShift);
			    	bits = *--psrc;
			    	bits1 |= SCRLEFT(bits, leftShift);
			    	pdst = --pdstOrg;
    				for(i=0; i<depth; i++)
    				{
				    alu = palu[i];
				    if(alu != GXnoop)
				    {
					tmp = *pdst;
					*pdst = (tmp & ~endmask) |
				  	     (DoRop(alu,bits1,tmp) & endmask);
				    }
				    pdst += dstPlaneOffset;
				}
			    }
			    nl = nlMiddle;
	
#ifdef LARGE_INSTRUCTION_CACHE
			    bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
			    psrc -= nl & (UNROLL - 1);
			    pdstOrg -= nl & (UNROLL - 1);
#define BodyOdd(n) \
    bits = psrc[n-1]; \
    tmp = BitRight(bits1, rightShift) | BitLeft(bits, leftShift); \
    pdst = pdstOrg; \
    for(i=0; i<depth; i++) \
    { \
	alu = palu[i]; \
	if(alu != GXnoop) \
	{ \
	    pdst[n-1] = DoRop(alu,tmp,pdst[n-1]); \
	} \
        pdst += dstPlaneOffset; \
    } 

#define BodyEven(n) \
    bits1 = psrc[n-1]; \
    tmp = BitRight(bits, rightShift) | BitLeft(bits1, leftShift); \
    pdst = pdstOrg; \
    for(i=0; i<depth; i++) \
    { \
	alu = palu[i]; \
	if(alu != GXnoop) \
	{ \
	    pdst[n-1] = DoRop(alu,tmp,pdst[n-1]); \
	} \
        pdst += dstPlaneOffset; \
    } 

#define LoopReset \
    pdstOrg -= UNROLL; \
    psrc -= UNROLL;
#else
#define BodyOdd(n) \
    bits = *--psrc;  pdst = --pdstOrg; \
    tmp = BitRight(bits1, rightShift) | BitLeft(bits, leftShift); \
    for(i=0; i<depth; i++) \
    { \
	alu = palu[i]; \
	if(alu != GXnoop) \
	{ \
	    *pdst = DoRop(alu,tmp,*pdst); \
	} \
        pdst += dstPlaneOffset; \
    } 

#define BodyEven(n) \
    bits1 = *--psrc;  pdst = --pdstOrg; \
    tmp =  BitRight(bits, rightShift) | BitLeft(bits1, leftShift); \
    for(i=0; i<depth; i++) \
    { \
	alu = palu[i]; \
	if(alu != GXnoop) \
	{ \
	    *pdst = DoRop(alu,tmp,*pdst); \
	} \
        pdst += dstPlaneOffset; \
    } 

#define LoopReset   ;
#endif
			    PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    DuffL (nl, label4,
				bits1 = BitRight(bits, rightShift);
				bits = *--psrc; pdst = --pdstOrg;
				tmp = bits1 | BitLeft(bits, leftShift);
    				for(i=0; i<depth; i++)
    				{
				    alu = palu[i];
				    if(alu != GXnoop)
				    {
	    				*pdst = DoRop(alu,tmp,*pdst);
				    }
				    pdst += dstPlaneOffset;
				}
			    )
#endif
			    if (startmask)
			    {
			    	bits1 = SCRRIGHT(bits, rightShift);
			    	if (SCRRIGHT (startmask, leftShift))
			    	{
				    bits = *--psrc;
				    bits1 |= SCRLEFT(bits, leftShift);
			    	}
				pdst = --pdstOrg;
    				for(i=0; i<depth; i++)
    				{
				    alu = palu[i];
				    if(alu != GXnoop)
				    {
					tmp = *pdst;
					*pdst = (tmp & ~startmask) |
				  	     (DoRop(alu,bits1,tmp) & startmask);
				    }
				    pdst += dstPlaneOffset;
				}
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    } /* xoffSrc == xoffDst */
	    	} /* xdir */
	    } /* width */
	    pbox++;
	    pptSrc++;
	} /* nbox */
    }

    /* free up stuff */
    if (pboxNew2)
    {
	DEALLOCATE_LOCAL(pptNew2);
	DEALLOCATE_LOCAL(pboxNew2);
    }
    if (pboxNew1)
    {
	DEALLOCATE_LOCAL(pptNew1);
	DEALLOCATE_LOCAL(pboxNew1);
    }
}


/*
** frame memory BitBlt functions
*/

/*
** not use src (NotN, Nto1, 1toN)
*/
void
cfbDoBitbltNotUseSrc(pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int  *pdstBase;	/* start of src and dst bitmaps */
    int widthDst;		/* add to get to same position in next line */
    register BoxPtr pbox;
    int nbox;
    int dstPlaneOffset;		/* dstnation drawble plane size */
    int depth = pDstDrawable->depth;

   if(depth == 1)
   {
	pmask &= 1;
   }

   if(! pmask) return;

    setBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    {
	FB_bitblt_fill_area(nbox, pbox, pdstBase, widthDst, alu, pmask,
			    label4);
    }
}


/*
** use src data (NtoN)
*/
void
cfbDoBitbltUseSrc(pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int *psrcBase, *pdstBase;	
				/* start of src and dst bitmaps */
    int widthSrc, widthDst;	/* add to get to same position in next line */

    register BoxPtr pbox;
    int nbox;

    BoxPtr pboxNew1, pboxNew2; /* temporaries for shuffling rectangles */
    DDXPointPtr pptNew1, pptNew2; /* shuffling boxes entails shuffling the
				     source points too */
    int w, h;
    int xdir;			/* 1 = left right, -1 = right left/ */
    int ydir;			/* 1 = top down, -1 = bottom up */
    unsigned int *psrcLine, *pdstLine;	
				/* pointers to line with current src and dst */
				/* following used for looping through a line */
    register unsigned int bits;
    register unsigned int bits1;
    int careful;
    int srcPlaneOffset;		/* src drawable plane size */
    int dstPlaneOffset;		/* dst drawable plane size */
    int depth = pDstDrawable->depth;
    OmronFbInfoPtr fbInfo;
    register int  pm = pmask;

    initPmask(pmask,depth);

    if(! pmask) return;

    fbInfo = getOmronFbInfo( pSrcDrawable );

    setBaseAndWidth( pSrcDrawable, psrcBase, widthSrc, srcPlaneOffset );

    if(isCommonPlane(psrcBase,fbInfo))
    {
	psrcBase = (unsigned int *)getPlaneAddres(fbInfo);
    }

    setWindowBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    if(isCommonPlane(pdstBase,fbInfo))
    {
	pdstBase = (unsigned int *)getPlaneAddres(fbInfo);
    }

    /* XXX we have to err on the side of safety when both are windows,
     * because we don't know if IncludeInferiors is being used.
     */
    careful = (psrcBase == pdstBase);

    checkRevSrc(psrcBase, srcPlaneOffset, pSrcDrawable->depth);

    if(depth == 1) 
    {
	if(copyPlaneBitPlane)
	{
	    psrcBase += (copyPlaneBitPlane - 1) * srcPlaneOffset;
	    pmask = (pm & 1) ? bitmapPmask() : 0;
	}
    }
    else if(pSrcDrawable->depth == 1)
    {
	srcPlaneOffset = 0;
    }

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    checkAndChangeRectBox( careful, pptSrc, pbox, nbox,
                           pboxNew1, pptNew1, pboxNew2, pptNew,
                           ydir, xdir);

    {
	register int    xoffSrc, xoffDst;
	int	leftShift, rightShift;

	while(nbox--)
	{
	    w = pbox->x2 - pbox->x1;
	    h = pbox->y2 - pbox->y1;

	    if (ydir == -1) /* start at last scanline of rectangle */
	    {
	        psrcLine = psrcBase + ((pptSrc->y+h-1) * -widthSrc);
	        pdstLine = pdstBase + ((pbox->y2-1) * -widthDst);
	    }
	    else /* start at first scanline */
	    {
	        psrcLine = psrcBase + (pptSrc->y * widthSrc);
    		pdstLine = pdstBase + (pbox->y1 * widthDst);
	     }

	    xoffDst = pbox->x1 & PIM;

	    /* width < PPW */
	    if ((xoffDst + w) <= PPW)
	    {
		pdstLine += (pbox->x1 >> PWSH);
		psrcLine += (pptSrc->x >> PWSH);
		xoffSrc = pptSrc->x & PIM;

		FB_bitblt_small_copy( psrcLine, pdstLine, xoffSrc, xoffDst, 
				      srcPlaneOffset, dstPlaneOffset,
				      h, w, alu, pmask);
	    }
	    else	/* width > PPW */
	    {   register unsigned int *psrc;
		register unsigned int *pdst;
		unsigned int *psrcOrg, *pdstOrg;
		unsigned int *psrcEnd, *pdstEnd;
    		unsigned int startmask, endmask;
    		int nlMiddle;		/* whole longwords in dst */
    		register int nl;	/* temp copy of nlMiddle */

	    	maskbits(pbox->x1, w, startmask, endmask, nlMiddle);
	    	if (xdir == 1)
	    	{
	    	    xoffSrc = pptSrc->x & PIM;
		    pdstLine += (pbox->x1 >> PWSH);
		    psrcLine += (pptSrc->x >> PWSH);
		    if (xoffSrc == xoffDst)
		    {   
			while (h--)
			{
			    psrcOrg = psrcLine;
			    pdstOrg = pdstLine;
			    if (startmask)
			    {
				sx_set_operation_all(alu,startmask);
				psrc = psrcOrg;
				pdst = pdstOrg;
				pm = pmask;
				planeLoop(pm) 
				{
				    if(checkPmask(pm))
				    {
					*pdst = *psrc;
				    }
				    updatePmask(pm);
				    pdst += dstPlaneOffset;
				    psrc += srcPlaneOffset;
				}
			        psrcOrg++;
			        pdstOrg++;
			    }

			    psrcEnd = psrcOrg + nlMiddle;
			    pdstEnd = pdstOrg + nlMiddle;

			    if(nlMiddle) {
				sx_set_operation_all(alu,~0);
			        pm = pmask;
				planeLoop(pm) 
				{
				    if(checkPmask(pm))
				    {
					psrc = psrcOrg;
					pdst = pdstOrg;
			  		nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
					psrc += nl & (UNROLL-1);
					pdst += nl & (UNROLL-1);
#define BodyOdd(n) pdst[-n] = psrc[-n];
#define BodyEven(n) pdst[-n] = psrc[-n];
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n)  *pdst++ = *psrc++;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
				    	PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
			    		DuffL(nl, label1, *pdst++ = *psrc++;)
#endif
				    } /* checkPmask */
				    updatePmask(pm);
				    pdstOrg += dstPlaneOffset;
				    psrcOrg += srcPlaneOffset;
				} /* endLoop */
			    } /* nlMiddle */

			    if (endmask)
			    {
				sx_set_operation_all(alu,endmask);
				pm = pmask;
				psrc = psrcEnd;
				pdst = pdstEnd;
				planeLoop(pm) 
				{
				    if(checkPmask(pm))
				    {
					*pdst = *psrc;
				    }
				    updatePmask(pm);
				    pdst += dstPlaneOffset;
				    psrc += srcPlaneOffset;
				}
			    } /* endmask */
			    psrcLine += widthSrc;
			    pdstLine += widthDst;
			} /* h */
		    }
		    else
		    {
			if (xoffSrc > xoffDst)
			{
			    leftShift = (xoffSrc - xoffDst);
			    rightShift = PPW - leftShift;
			}
			else
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}

			while (h--)
			{
		    	    psrcOrg = psrcLine;
		    	    pdstOrg = pdstLine;
			    if (startmask)
			    {
			        sx_set_operation_all(alu,startmask);
				pm = pmask;
				psrc = psrcOrg;
				pdst = pdstOrg;
				if (xoffSrc > xoffDst)
				{
				    planeLoop(pm)
				    {
					if(checkPmask(pm))
					{
				    	    *pdst = SCRLEFT(psrc[0],leftShift) |
				    		    SCRRIGHT(psrc[1],rightShift);
					}
					updatePmask(pm);
					pdst += dstPlaneOffset;
					psrc += srcPlaneOffset;
				    }
				    psrcOrg++;
				    pdstOrg++;
				}
				else 
				{
				    planeLoop(pm)
				    {
					if(checkPmask(pm))
					{
				    	    *pdst = SCRRIGHT(psrc[0],rightShift);
					}
					updatePmask(pm);
					pdst += dstPlaneOffset;
					psrc += srcPlaneOffset;
				    }
				    pdstOrg++;
				}
			    } /* startmask */

			    psrcEnd = psrcOrg + nlMiddle;
			    pdstEnd = pdstOrg + nlMiddle;
			
			    if( nlMiddle )
			    {
			        sx_set_operation_all(alu,~0);
				pm = pmask;
				planeLoop(pm)
				{
				    if(checkPmask(pm))
				    {
				        psrc = psrcOrg;
				        pdst = pdstOrg;
					bits = *psrc++;
					nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
					bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
					psrc += nl & (UNROLL-1);
					pdst += nl & (UNROLL-1);
#define BodyOdd(n) \
    bits = psrc[-n]; \
    pdst[-n] = BitLeft(bits1, leftShift) | BitRight(bits, rightShift);
#define BodyEven(n) \
    bits1 = psrc[-n]; \
    pdst[-n] = BitLeft(bits, leftShift) | BitRight(bits1, rightShift);
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n) \
    bits = *psrc++; \
    *pdst++ = BitLeft(bits1, leftShift) | BitRight(bits, rightShift);
#define BodyEven(n) \
    bits1 = *psrc++; \
    *pdst++ = BitLeft(bits, leftShift) | BitRight(bits1, rightShift);
#define LoopReset   ;
#endif	/* !FAST_CONSTANT_OFFSET_MODE */
					PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
					DuffL (nl,label2,
					    bits1 = BitLeft(bits, leftShift);
					    bits = *psrc++;
					    *pdst++ = bits1 |
					    BitRight(bits, rightShift);
					)
#endif
				    } /* checkPmask() */
				    updatePmask(pm);
				    pdstOrg += dstPlaneOffset;
				    psrcOrg += srcPlaneOffset;
				} /* endLoop */
			    } /* nlMiddle */

			    if (endmask)
			    {
				sx_set_operation_all(alu,endmask);
				psrc = psrcEnd;
		    	        pdst = pdstEnd;
				pm = pmask;
				if (SCRLEFT(endmask, rightShift))
				{	
				    planeLoop(pm) 
				    {
					if(checkPmask(pm))
					{
				    	    *pdst = SCRLEFT(psrc[0], leftShift) |
					    	    SCRRIGHT(psrc[1], rightShift);
				    	}
				        updatePmask(pm);
				        pdst += dstPlaneOffset;
				        psrc += srcPlaneOffset;
				    }
				}
				else /* SCRLEFT */
				{	
				    planeLoop(pm) 
				    {
					if(checkPmask(pm))
					{
				    	    *pdst = SCRLEFT(psrc[0], leftShift);
				    	}
				        updatePmask(pm);
				        pdst += dstPlaneOffset;
				        psrc += srcPlaneOffset;
				    }
				}
			    } /* endmask */
			    psrcLine += widthSrc;
			    pdstLine += widthDst;
			} /* h */
		    }
	    	}
	    	else	/* xdir == -1 */
	    	{
	    	    xoffSrc = (pptSrc->x + w - 1) & PIM;
	    	    xoffDst = (pbox->x2 - 1) & PIM;
		    pdstLine += ((pbox->x2-1) >> PWSH) + 1;
		    psrcLine += ((pptSrc->x+w - 1) >> PWSH) + 1;

		    if (xoffSrc == xoffDst)
		    {
	    	    	while (h--)
	    	    	{
			    psrcOrg = psrcLine;
			    pdstOrg = pdstLine;
			    if (endmask)
			    {
				sx_set_operation_all(alu,endmask);
				pm = pmask;
				psrc = --psrcOrg;
				pdst = --pdstOrg;
				planeLoop(pm)
				{
				    if(checkPmask(pm))
				    {
				    	*pdst = *psrc;
				    }
				    updatePmask(pm);
				    pdst += dstPlaneOffset;
				    psrc += srcPlaneOffset;
				}
			    }

			    psrcEnd = psrcOrg - nlMiddle;
			    pdstEnd = pdstOrg - nlMiddle;

			    if(nlMiddle)
			    {
				sx_set_operation_all(alu,~0);
				pm = pmask;
				planeLoop(pm)
				{
				    if(checkPmask(pm))
				    {
				        psrc = psrcOrg;
				        pdst = pdstOrg;
					nl = nlMiddle;

#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
					psrc -= nl & (UNROLL - 1);
					pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) pdst[n-1] = psrc[n-1];
#define BodyEven(n) pdst[n-1] = psrc[n-1];
#define LoopReset \
    pdst -= UNROLL;\
    psrc -= UNROLL;
#else
#define BodyOdd(n)  *--pdst = *--psrc;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
					PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
					DuffL(nl,label4, *--pdst = *--psrc;)
#endif
				    }
				    updatePmask(pm);
				    pdstOrg += dstPlaneOffset;
				    psrcOrg += srcPlaneOffset;
				}
			    }

			    if (startmask)
			    {
				sx_set_operation_all(alu,startmask);
				pm = pmask;
				psrc = psrcEnd - 1;
				pdst = pdstEnd - 1;
				planeLoop(pm)
				{
				    if(checkPmask(pm))
				    {
				    	*pdst = *psrc;
				    }
				    updatePmask(pm);
				    pdst += dstPlaneOffset;
				    psrc += srcPlaneOffset;
				}
			    } /* startmask */
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    }
		    else  /* xoffDst > xoffSrc */
		    {
			if (xoffDst > xoffSrc)
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}
			else
			{
			    leftShift = (xoffSrc - xoffDst);
			    rightShift = PPW - leftShift;
			}

	    	    	while (h--)
	    	    	{
			    psrcOrg = psrcLine;
			    pdstOrg = pdstLine;
			    if (endmask)
			    {
				pm = pmask;
				psrc = psrcOrg - 1;
				pdst = --pdstOrg;
				sx_set_operation_all(alu,endmask);
				if (xoffDst > xoffSrc)
				{
				    planeLoop( pm )
				    {
					if(checkPmask(pm))
					{
					    *pdst = SCRRIGHT(psrc[0], rightShift) |
						    SCRLEFT(psrc[-1], leftShift);
					}
					updatePmask(pm);
					pdst += dstPlaneOffset;
					psrc += srcPlaneOffset;
				    }
				    psrcOrg--;
				}
				else
				{
				    planeLoop( pm )
				    {
					if(checkPmask(pm))
					{
					    *pdst = SCRLEFT(psrc[0], leftShift);
					}
					updatePmask(pm);
					pdst += dstPlaneOffset;
					psrc += srcPlaneOffset;
				    }
				}
			    }

			    psrcEnd = psrcOrg - nlMiddle;
			    pdstEnd = pdstOrg - nlMiddle;

			    if( nlMiddle )
			    {
				sx_set_operation_all(alu,~0);
				pm = pmask;
				planeLoop( pm )
				{
				    if(checkPmask(pm))
				    {
					psrc = psrcOrg;
					pdst = pdstOrg;
					nl = nlMiddle;
					bits = *--psrc;
	
#ifdef LARGE_INSTRUCTION_CACHE
					bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
					psrc -= nl & (UNROLL - 1);
					pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) \
    bits = psrc[n-1]; \
    pdst[n-1] = BitRight(bits1, rightShift) | BitLeft(bits, leftShift);
#define BodyEven(n) \
    bits1 = psrc[n-1]; \
    pdst[n-1] = BitRight(bits, rightShift) | BitLeft(bits1, leftShift);
#define LoopReset \
    pdst -= UNROLL; \
    psrc -= UNROLL;
#else
#define BodyOdd(n) \
    bits = *--psrc; \
    *--pdst = BitRight(bits1, rightShift) | BitLeft(bits, leftShift);
#define BodyEven(n) \
    bits1 = *--psrc; \
    *--pdst = BitRight(bits, rightShift) | BitLeft(bits1, leftShift);
#define LoopReset   ;
#endif
					PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
					DuffL (nl, label5,
					    bits1 = BitRight(bits, rightShift);
					    bits = *--psrc;
					    *--pdst = bits1 |
						  BitLeft(bits, leftShift);
					)
#endif
				    }
				    updatePmask(pm);
				    pdstOrg += dstPlaneOffset;
				    psrcOrg += srcPlaneOffset;
				}
			    }

			    if (startmask)
			    {
				sx_set_operation_all(alu,startmask);
				pm = pmask;
				psrc = psrcEnd - 1;
				pdst = pdstEnd - 1;
			    	if (SCRRIGHT (startmask, leftShift))
			    	{
				    planeLoop( pm )
				    {
				        if(checkPmask(pm))
				        {
				    	    *pdst = SCRRIGHT(psrc[0], rightShift) |
				    	            SCRLEFT(psrc[-1], leftShift);
					}
				        updatePmask(pm);
				        psrc += srcPlaneOffset;
				        pdst += dstPlaneOffset;
				    }
			    	}
				else
			    	{
				    planeLoop( pm )
				    {
				        if(checkPmask(pm))
				        {
				    	    *pdst = SCRRIGHT(psrc[0], rightShift);
					}
				        updatePmask(pm);
				        psrc += srcPlaneOffset;
				        pdst += dstPlaneOffset;
				    }
			    	}
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	}
		    }
	    	} /* xdir */
	    } /* width */
	    pbox++;
	    pptSrc++;
	} /* nbox */
    }

    /* free up stuff */
    if (pboxNew2)
    {
	DEALLOCATE_LOCAL(pptNew2);
	DEALLOCATE_LOCAL(pboxNew2);
    }
    if (pboxNew1)
    {
	DEALLOCATE_LOCAL(pptNew1);
	DEALLOCATE_LOCAL(pboxNew1);
    }
}


/*
** use src data (Nto1)
*/
void
cfbDoBitbltUseSrc1(pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int *psrcBase, *pdstBase;	
				/* start of src and dst bitmaps */
    int widthSrc, widthDst;	/* add to get to same position in next line */

    BoxPtr pbox;
    int nbox;

    BoxPtr pboxNew1, pboxNew2; /* temporaries for shuffling rectangles */
    DDXPointPtr pptNew1, pptNew2; /* shuffling boxes entails shuffling the
				     source points too */
    int w, h;
    int xdir;			/* 1 = left right, -1 = right left/ */
    int ydir;			/* 1 = top down, -1 = bottom up */
    unsigned int *psrcLine, *pdstLine;	
				/* pointers to line with current src and dst */
				/* following used for looping through a line */
    register unsigned int bits;
    register unsigned int bits1;
    int careful;
    int srcPlaneOffset;		/* src drawable plane size */
    int dstPlaneOffset;		/* dst drawable plane size */
    int depth = pDstDrawable->depth;
    OmronFbInfoPtr fbInfo;

    if ( !(pmask & 1) ) return;

    fbInfo = getOmronFbInfo( pSrcDrawable );

    setBaseAndWidth( pSrcDrawable, psrcBase, widthSrc, srcPlaneOffset );
    setBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    if(isCommonPlane(psrcBase,fbInfo))
    {
	psrcBase = (unsigned int *)getPlaneAddres(fbInfo);
    }

    if(isCommonPlane(pdstBase,fbInfo))
    {
	pdstBase = (unsigned int *)getPlaneAddres(fbInfo);
    }

    /* XXX we have to err on the side of safety when both are windows,
     * because we don't know if IncludeInferiors is being used.
     */
    careful = (psrcBase == pdstBase);

    checkRevSrc(psrcBase, srcPlaneOffset, pSrcDrawable->depth);

    psrcBase += (copyPlaneBitPlane - 1) * srcPlaneOffset;

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    checkAndChangeRectBox( careful, pptSrc, pbox, nbox,
                           pboxNew1, pptNew1, pboxNew2, pptNew,
                           ydir, xdir);

    { 
	register int    xoffSrc, xoffDst;
	int	leftShift, rightShift;

	while(nbox--)
	{
	    w = pbox->x2 - pbox->x1;
	    h = pbox->y2 - pbox->y1;

	    if (ydir == -1) /* start at last scanline of rectangle */
	    {
	        psrcLine = psrcBase + ((pptSrc->y+h-1) * -widthSrc);
	        pdstLine = pdstBase + ((pbox->y2-1) * -widthDst);
	    }
	    else /* start at first scanline */
	    {
	        psrcLine = psrcBase + (pptSrc->y * widthSrc);
    		pdstLine = pdstBase + (pbox->y1 * widthDst);
	     }

	    xoffDst = pbox->x1 & PIM;

	    /* width < PPW */
	    if ((xoffDst + w) <= PPW)
	    {
		pdstLine += (pbox->x1 >> PWSH);
		psrcLine += (pptSrc->x >> PWSH);
		xoffSrc = pptSrc->x & PIM;
		FB_bitblt_small_copy1( psrcLine, pdstLine, xoffSrc, xoffDst, 
				       h, w, alu);
	    }
	    else	/* width > PPW */
	    {   register unsigned int *psrc, *pdst;
		unsigned int *psrcOrg, *pdstOrg;
		unsigned int *psrcEnd, *pdstEnd;
    		unsigned int startmask, endmask;
    		int nlMiddle;		/* whole longwords in dst */
    		register int nl;	/* temp copy of nlMiddle */

	    	maskbits(pbox->x1, w, startmask, endmask, nlMiddle);
	    	if (xdir == 1)
	    	{
	    	    xoffSrc = pptSrc->x & PIM;
		    pdstLine += (pbox->x1 >> PWSH);
		    psrcLine += (pptSrc->x >> PWSH);
		    if (xoffSrc == xoffDst)
		    {   
			while (h--)
			{
			    psrc = psrcLine;
			    pdst = pdstLine;
			    if (startmask)
			    {
				sx_set_operation_all(alu,startmask);
				*pdst++ = *psrc++;
			    }

			    if(nlMiddle) {
				sx_set_operation_all(alu,~0);
		  		nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
				psrc += nl & (UNROLL-1);
				pdst += nl & (UNROLL-1);
#define BodyOdd(n) pdst[-n] = psrc[-n];
#define BodyEven(n) pdst[-n] = psrc[-n];
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n)  *pdst++ = *psrc++;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
			    	PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
		    		DuffL(nl, label1, *pdst++ = *psrc++;)
#endif
			    } /* nlMiddle */

			    if (endmask)
			    {
				sx_set_operation_all(alu,endmask);
				*pdst = *psrc;
			    } /* endmask */
			    psrcLine += widthSrc;
			    pdstLine += widthDst;
			} /* h */
		    }
		    else
		    {
			if (xoffSrc > xoffDst)
			{
			    leftShift = (xoffSrc - xoffDst);
			    rightShift = PPW - leftShift;
			}
			else
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}

			while (h--)
			{
		    	    psrc = psrcLine;
		    	    pdst = pdstLine;
			    bits = 0;
			    if (xoffSrc > xoffDst)
			        bits = *psrc++;
			    if (startmask)
			    {
			        sx_set_operation_all(alu,startmask);
				bits1 = SCRLEFT(bits,leftShift);
			    	bits = *psrc++;
			    	*pdst++ = SCRRIGHT(bits,rightShift) | bits1;
			    } /* startmask */

			    psrcEnd = psrcOrg + nlMiddle;
			    pdstEnd = pdstOrg + nlMiddle;
			
			    if( nlMiddle )
			    {
			        sx_set_operation_all(alu,~0);
				nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
				bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
				psrc += nl & (UNROLL-1);
				pdst += nl & (UNROLL-1);
#define BodyOdd(n) \
    bits = psrc[-n]; \
    pdst[-n] = BitLeft(bits1, leftShift) | BitRight(bits, rightShift);
#define BodyEven(n) \
    bits1 = psrc[-n]; \
    pdst[-n] = BitLeft(bits, leftShift) | BitRight(bits1, rightShift);
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n) \
    bits = *psrc++; \
    *pdst++ = BitLeft(bits1, leftShift) | BitRight(bits, rightShift);
#define BodyEven(n) \
    bits1 = *psrc++; \
    *pdst++ = BitLeft(bits, leftShift) | BitRight(bits1, rightShift);
#define LoopReset   ;
#endif	/* !FAST_CONSTANT_OFFSET_MODE */
				PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				DuffL (nl,label2,
				    bits1 = BitLeft(bits, leftShift);
				    bits = *psrc++;
				    *pdst++ = bits1 |
				    BitRight(bits, rightShift);
				)
#endif
			    } /* nlMiddle */

			    if (endmask)
			    {
			        sx_set_operation_all(alu,endmask);
			    	if (SCRLEFT(endmask, rightShift))
			    	{
			    	    bits1 = SCRLEFT(bits, leftShift);
				    bits = *psrc++;
				    *pdst = SCRRIGHT(bits, rightShift) | bits1;
			    	}
				else
				{
			    	    *pdst = SCRLEFT(bits, leftShift);
				}
			    } /* endmask */
			    psrcLine += widthSrc;
			    pdstLine += widthDst;
			} /* h */
		    }
	    	}
	    	else	/* xdir == -1 */
	    	{
	    	    xoffSrc = (pptSrc->x + w - 1) & PIM;
	    	    xoffDst = (pbox->x2 - 1) & PIM;
		    pdstLine += ((pbox->x2-1) >> PWSH) + 1;
		    psrcLine += ((pptSrc->x+w - 1) >> PWSH) + 1;

		    if (xoffSrc == xoffDst)
		    {
	    	    	while (h--)
	    	    	{
			    psrc = psrcLine;
			    pdst = pdstLine;
			    if (endmask)
			    {
				sx_set_operation_all(alu,endmask);
			   	*--pdst = *--psrc;
			    }

			    if(nlMiddle)
			    {
				sx_set_operation_all(alu,~0);
				nl = nlMiddle;

#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
				psrc -= nl & (UNROLL - 1);
				pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) pdst[n-1] = psrc[n-1];
#define BodyEven(n) pdst[n-1] = psrc[n-1];
#define LoopReset \
    pdst -= UNROLL;\
    psrc -= UNROLL;
#else
#define BodyOdd(n)  *--pdst = *--psrc;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
				PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				DuffL(nl,label4, *--pdst = *--psrc;)
#endif
			    }

			    if (startmask)
			    {
				sx_set_operation_all(alu,startmask);
				*--pdst = *--psrc;
			    } /* startmask */
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    }
		    else  /* xoffDst > xoffSrc */
		    {
			if (xoffSrc > xoffDst)
			{
			    leftShift = (xoffSrc - xoffDst);
			    rightShift = PPW - leftShift;
			}
			else
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}

	    	    	while (h--)
	    	    	{
			    psrc = psrcLine;
			    pdst = pdstLine;
			    bits = 0;
			    if (xoffDst > xoffSrc)
				bits = *--psrc;
			    if (endmask)
			    {
				sx_set_operation_all(alu,endmask);
			    	bits1 = SCRRIGHT(bits, rightShift);
			    	bits = *--psrc;
			    	*--pdst = SCRLEFT(bits, leftShift) | bits1;
			    }

			    if( nlMiddle )
			    {
				sx_set_operation_all(alu,~0);
				nl = nlMiddle;
	
#ifdef LARGE_INSTRUCTION_CACHE
				bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
				psrc -= nl & (UNROLL - 1);
				pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) \
    bits = psrc[n-1]; \
    pdst[n-1] = BitRight(bits1, rightShift) | BitLeft(bits, leftShift);
#define BodyEven(n) \
    bits1 = psrc[n-1]; \
    pdst[n-1] = BitRight(bits, rightShift) | BitLeft(bits1, leftShift);
#define LoopReset \
    pdst -= UNROLL; \
    psrc -= UNROLL;
#else
#define BodyOdd(n) \
    bits = *--psrc; \
    *--pdst = BitRight(bits1, rightShift) | BitLeft(bits, leftShift);
#define BodyEven(n) \
    bits1 = *--psrc; \
    *--pdst = BitRight(bits, rightShift) | BitLeft(bits1, leftShift);
#define LoopReset   ;
#endif
				PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				DuffL (nl, label5,
				    bits1 = BitRight(bits, rightShift);
				    bits = *--psrc;
				    *--pdst = bits1 |
					  BitLeft(bits, leftShift);
				)
#endif
			    }

			    if (startmask)
			    {
				sx_set_operation_all(alu,startmask);
				bits1 = SCRRIGHT(bits, rightShift);
				if(SCRRIGHT(startmask, leftShift))
				{
				    bits = *--psrc;
				    bits1 |= SCRLEFT(bits, leftShift);
				}
				--pdst;
				*pdst = (*pdst & ~startmask) |
				    	(bits1 & startmask);
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    } /* xoffSrc == xoffDst */
	    	} /* xdir */
	    } /* width */
	    pbox++;
	    pptSrc++;
	} /* nbox */
    } /* GCxopy */

    /* free up stuff */
    if (pboxNew2)
    {
	DEALLOCATE_LOCAL(pptNew2);
	DEALLOCATE_LOCAL(pboxNew2);
    }
    if (pboxNew1)
    {
	DEALLOCATE_LOCAL(pptNew1);
	DEALLOCATE_LOCAL(pboxNew1);
    }
}


/*
** use src data (1toN)
*/
void
cfbDoBitbltUseSrc8(pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int *psrcBase, *pdstBase;	
				/* start of src and dst bitmaps */
    int widthSrc, widthDst;	/* add to get to same position in next line */

    BoxPtr pbox;
    int nbox;

    BoxPtr pboxNew1, pboxNew2; /* temporaries for shuffling rectangles */
    DDXPointPtr pptNew1, pptNew2; /* shuffling boxes entails shuffling the
				     source points too */
    int w, h;
    int xdir;			/* 1 = left right, -1 = right left/ */
    int ydir;			/* 1 = top down, -1 = bottom up */
    unsigned int *psrcLine, *pdstLine;	
				/* pointers to line with current src and dst */
				/* following used for looping through a line */
    register unsigned int bits;
    register unsigned int bits1;
    int careful;
    int srcPlaneOffset;		/* src drawable plane size */
    int dstPlaneOffset;		/* dst drawable plane size */
    int depth = pDstDrawable->depth;
    unsigned long  fg    =    cfbPlanePixelsfg;
    unsigned long  bg    =    cfbPlanePixelsbg;
    unsigned long pm1, pm2, pm3, pm4;
    OmronFbInfoPtr fbInfo;

    fbInfo = getOmronFbInfo( pSrcDrawable );

    setBaseAndWidth( pSrcDrawable, psrcBase, widthSrc, srcPlaneOffset );
    setBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    /* XXX we have to err on the side of safety when both are windows,
     * because we don't know if IncludeInferiors is being used.
     */
    careful = (psrcBase == pdstBase);

    if(isCommonPlane(psrcBase,fbInfo))
    {
	psrcBase = (unsigned int *)getPlaneAddres(fbInfo);
    }

    compute_1toN_operation( alu, fg, bg, pm1, pm2, pm3, pm4, pmask );

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    checkAndChangeRectBox( careful, pptSrc, pbox, nbox,
                           pboxNew1, pptNew1, pboxNew2, pptNew,
                           ydir, xdir);

    { 
	register int    xoffSrc, xoffDst;
	int	leftShift, rightShift;

	while(nbox--)
	{
	    w = pbox->x2 - pbox->x1;
	    h = pbox->y2 - pbox->y1;

	    if (ydir == -1) /* start at last scanline of rectangle */
	    {
	        psrcLine = psrcBase + ((pptSrc->y+h-1) * -widthSrc);
	        pdstLine = pdstBase + ((pbox->y2-1) * -widthDst);
	    }
	    else /* start at first scanline */
	    {
	        psrcLine = psrcBase + (pptSrc->y * widthSrc);
    		pdstLine = pdstBase + (pbox->y1 * widthDst);
	     }

	    xoffDst = pbox->x1 & PIM;

	    /* width < PPW */
	    if ((xoffDst + w) <= PPW)
	    {
		pdstLine += (pbox->x1 >> PWSH);
		psrcLine += (pptSrc->x >> PWSH);
		xoffSrc = pptSrc->x & PIM;
		FB_bitblt_small_copy8( psrcLine, pdstLine, xoffSrc, xoffDst, 
				       h, w, alu, pm1, pm2, pm3, pm4, pmask);
	    }
	    else	/* width > PPW */
	    {   register unsigned int *psrc, *pdst;
		unsigned int *psrcOrg, *pdstOrg;
		unsigned int *psrcEnd, *pdstEnd;
    		unsigned int startmask, endmask;
    		int nlMiddle;		/* whole longwords in dst */
    		register int nl;	/* temp copy of nlMiddle */

	    	maskbits(pbox->x1, w, startmask, endmask, nlMiddle);
	    	if (xdir == 1)
	    	{
	    	    xoffSrc = pptSrc->x & PIM;
		    pdstLine += (pbox->x1 >> PWSH);
		    psrcLine += (pptSrc->x >> PWSH);
		    if (xoffSrc == xoffDst)
		    {   
			while (h--)
			{
			    psrc = psrcLine;
			    pdst = pdstLine;
			    if (startmask)
			    {
				sx_set_1toN_operation(alu, startmask, pm1, pm2, pm3, pm4, pmask);
				*pdst++ = *psrc++;
			    }

			    if(nlMiddle) {
				sx_set_1toN_operation(alu, ~0, pm1, pm2, pm3, pm4, pmask);
		  		nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
				psrc += nl & (UNROLL-1);
				pdst += nl & (UNROLL-1);
#define BodyOdd(n) pdst[-n] = psrc[-n];
#define BodyEven(n) pdst[-n] = psrc[-n];
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n)  *pdst++ = *psrc++;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
			    	PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
		    		DuffL(nl, label1, *pdst++ = *psrc++;)
#endif
			    } /* nlMiddle */

			    if (endmask)
			    {
				sx_set_1toN_operation(alu, endmask, pm1, pm2, pm3, pm4, pmask);
				*pdst = *psrc;
			    } /* endmask */
			    psrcLine += widthSrc;
			    pdstLine += widthDst;
			} /* h */
		    }
		    else
		    {
			if (xoffSrc > xoffDst)
			{
			    leftShift = (xoffSrc - xoffDst);
			    rightShift = PPW - leftShift;
			}
			else
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}

			while (h--)
			{
		    	    psrc = psrcLine;
		    	    pdst = pdstLine;
			    bits = 0;
			    if (xoffSrc > xoffDst)
			        bits = *psrc++;
			    if (startmask)
			    {
				sx_set_1toN_operation(alu, startmask, pm1, pm2, pm3, pm4, pmask);
				bits1 = SCRLEFT(bits,leftShift);
			    	bits = *psrc++;
			    	*pdst++ = SCRRIGHT(bits,rightShift) | bits1;
			    } /* startmask */

			    psrcEnd = psrcOrg + nlMiddle;
			    pdstEnd = pdstOrg + nlMiddle;
			
			    if( nlMiddle )
			    {
				sx_set_1toN_operation(alu, ~0, pm1, pm2, pm3, pm4, pmask);
				nl = nlMiddle;
#ifdef LARGE_INSTRUCTION_CACHE
				bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
				psrc += nl & (UNROLL-1);
				pdst += nl & (UNROLL-1);
#define BodyOdd(n) \
    bits = psrc[-n]; \
    pdst[-n] = BitLeft(bits1, leftShift) | BitRight(bits, rightShift);
#define BodyEven(n) \
    bits1 = psrc[-n]; \
    pdst[-n] = BitLeft(bits, leftShift) | BitRight(bits1, rightShift);
#define LoopReset \
    pdst += UNROLL; \
    psrc += UNROLL;
#else
#define BodyOdd(n) \
    bits = *psrc++; \
    *pdst++ = BitLeft(bits1, leftShift) | BitRight(bits, rightShift);
#define BodyEven(n) \
    bits1 = *psrc++; \
    *pdst++ = BitLeft(bits, leftShift) | BitRight(bits1, rightShift);
#define LoopReset   ;
#endif	/* !FAST_CONSTANT_OFFSET_MODE */
				PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				DuffL (nl,label2,
				    bits1 = BitLeft(bits, leftShift);
				    bits = *psrc++;
				    *pdst++ = bits1 |
				    BitRight(bits, rightShift);
				)
#endif
			    } /* nlMiddle */

			    if (endmask)
			    {
				sx_set_1toN_operation(alu, endmask, pm1, pm2, pm3, pm4, pmask);
			    	if (SCRLEFT(endmask, rightShift))
			    	{
			    	    bits1 = SCRLEFT(bits, leftShift);
				    bits = *psrc++;
				    *pdst = SCRRIGHT(bits, rightShift) | bits1;
			    	}
				else
				{
			    	    *pdst = SCRLEFT(bits, leftShift);
				}
			    } /* endmask */
			    psrcLine += widthSrc;
			    pdstLine += widthDst;
			} /* h */
		    }
	    	}
	    	else	/* xdir == -1 */
	    	{
	    	    xoffSrc = (pptSrc->x + w - 1) & PIM;
	    	    xoffDst = (pbox->x2 - 1) & PIM;
		    pdstLine += ((pbox->x2-1) >> PWSH) + 1;
		    psrcLine += ((pptSrc->x+w - 1) >> PWSH) + 1;

		    if (xoffSrc == xoffDst)
		    {
	    	    	while (h--)
	    	    	{
			    psrc = psrcLine;
			    pdst = pdstLine;
			    if (endmask)
			    {
				sx_set_1toN_operation(alu, endmask, pm1, pm2, pm3, pm4, pmask);
			   	*--pdst = *--psrc;
			    }

			    if(nlMiddle)
			    {
				sx_set_1toN_operation(alu, ~0, pm1, pm2, pm3, pm4, pmask);
				nl = nlMiddle;

#ifdef LARGE_INSTRUCTION_CACHE
#ifdef FAST_CONSTANT_OFFSET_MODE
				psrc -= nl & (UNROLL - 1);
				pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) pdst[n-1] = psrc[n-1];
#define BodyEven(n) pdst[n-1] = psrc[n-1];
#define LoopReset \
    pdst -= UNROLL;\
    psrc -= UNROLL;
#else
#define BodyOdd(n)  *--pdst = *--psrc;
#define BodyEven(n) BodyOdd(n)
#define LoopReset   ;
#endif
				PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				DuffL(nl,label4, *--pdst = *--psrc;)
#endif
			    }

			    if (startmask)
			    {
				sx_set_1toN_operation(alu, startmask, pm1, pm2, pm3, pm4, pmask);
				*--pdst = *--psrc;
			    } /* startmask */
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    }
		    else  /* xoffDst > xoffSrc */
		    {
			if (xoffSrc > xoffDst)
			{
			    leftShift = (xoffSrc - xoffDst);
			    rightShift = PPW - leftShift;
			}
			else
			{
			    rightShift = (xoffDst - xoffSrc);
			    leftShift = PPW - rightShift;
			}

	    	    	while (h--)
	    	    	{
			    psrc = psrcLine;
			    pdst = pdstLine;
			    bits = 0;
			    if (xoffDst > xoffSrc)
				bits = *--psrc;
			    if (endmask)
			    {
				sx_set_1toN_operation(alu, endmask, pm1, pm2, pm3, pm4, pmask);
			    	bits1 = SCRRIGHT(bits, rightShift);
			    	bits = *--psrc;
			    	*--pdst = SCRLEFT(bits, leftShift) | bits1;
			    }

			    if( nlMiddle )
			    {
				sx_set_1toN_operation(alu, ~0, pm1, pm2, pm3, pm4, pmask);
				nl = nlMiddle;
	
#ifdef LARGE_INSTRUCTION_CACHE
				bits1 = bits;
#ifdef FAST_CONSTANT_OFFSET_MODE
				psrc -= nl & (UNROLL - 1);
				pdst -= nl & (UNROLL - 1);
#define BodyOdd(n) \
    bits = psrc[n-1]; \
    pdst[n-1] = BitRight(bits1, rightShift) | BitLeft(bits, leftShift);
#define BodyEven(n) \
    bits1 = psrc[n-1]; \
    pdst[n-1] = BitRight(bits, rightShift) | BitLeft(bits1, leftShift);
#define LoopReset \
    pdst -= UNROLL; \
    psrc -= UNROLL;
#else
#define BodyOdd(n) \
    bits = *--psrc; \
    *--pdst = BitRight(bits1, rightShift) | BitLeft(bits, leftShift);
#define BodyEven(n) \
    bits1 = *--psrc; \
    *--pdst = BitRight(bits, rightShift) | BitLeft(bits1, leftShift);
#define LoopReset   ;
#endif
				PackedLoop
#undef BodyOdd
#undef BodyEven
#undef LoopReset
#else
				DuffL (nl, label5,
				    bits1 = BitRight(bits, rightShift);
				    bits = *--psrc;
				    *--pdst = bits1 |
					  BitLeft(bits, leftShift);
				)
#endif
			    }

			    if (startmask)
			    {
				sx_set_1toN_operation(alu, startmask, pm1, pm2, pm3, pm4, pmask);
				bits1 = SCRRIGHT(bits, rightShift);
				if(SCRRIGHT(startmask, leftShift))
				{
				    bits = *--psrc;
				    bits1 |= SCRLEFT(bits, leftShift);
				}
				--pdst;
				*pdst = (*pdst & ~startmask) |
				    	(bits1 & startmask);
			    }
		    	    pdstLine += widthDst;
		    	    psrcLine += widthSrc;
		    	} /* h */
		    } /* xoffSrc == xoffDst */
	    	} /* xdir */
	    } /* width */
	    pbox++;
	    pptSrc++;
	} /* nbox */
    } /* GCxopy */

    /* free up stuff */
    if (pboxNew2)
    {
	DEALLOCATE_LOCAL(pptNew2);
	DEALLOCATE_LOCAL(pboxNew2);
    }
    if (pboxNew1)
    {
	DEALLOCATE_LOCAL(pptNew1);
	DEALLOCATE_LOCAL(pboxNew1);
    }
}


/*
** main memory BitBlt functions
*/

void
cfbDoBitbltMemClear(pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int  *pdstBase;	/* start of src and dst bitmaps */
    int widthDst;		/* add to get to same position in next line */
    register BoxPtr pbox;
    int nbox;
    int dstPlaneOffset;		/* dstnation drawble plane size */
    int depth = pDstDrawable->depth;
    OmronFbInfoPtr fbInfo;

    initPmask(pmask,depth);

    if(! pmask) return;

    fbInfo = getOmronFbInfo( pSrcDrawable );

    setBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    if(isCommonPlane(pdstBase,fbInfo))
    {
	pdstBase = (unsigned int *)getPlaneAddres(fbInfo);
	sx_set_operation_all( GXcopy, ~0 );
    }

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    bitblt_fill_area(nbox, pbox, pdstBase, widthDst, dstPlaneOffset,
		     pmask, &=~ , label3 );
}

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnAND
#define	BITBLTFUNC	cfbDoBitbltMemAnd
#include "cfbbitblt.h"

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnANDREVERSE
#define	BITBLTFUNC	cfbDoBitbltMemAndReverse
#include "cfbbitblt.h"

/* copy --> use cfbDoBitbltCopy */

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnANDINVERTED
#define	BITBLTFUNC	cfbDoBitbltMemAndInverted
#include "cfbbitblt.h"

/* noop --> use cfbDoNoop */

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnXOR
#define	BITBLTFUNC	cfbDoBitbltMemXor
#include "cfbbitblt.h"

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnOR
#define	BITBLTFUNC	cfbDoBitbltMemOr
#include "cfbbitblt.h"

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnNOR
#define	BITBLTFUNC	cfbDoBitbltMemNor
#include "cfbbitblt.h"

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnEQUIV
#define	BITBLTFUNC	cfbDoBitbltMemEquiv
#include "cfbbitblt.h"

void
cfbDoBitbltMemInvert(pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int  *pdstBase;	/* start of src and dst bitmaps */
    int widthDst;		/* add to get to same position in next line */
    register BoxPtr pbox;
    int nbox;
    int dstPlaneOffset;		/* dstnation drawble plane size */
    int depth = pDstDrawable->depth;
    OmronFbInfoPtr fbInfo;

    initPmask(pmask,depth);

    if(! pmask) return;

    fbInfo = getOmronFbInfo( pSrcDrawable );

    setBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    if(isCommonPlane(pdstBase,fbInfo))
    {
	pdstBase = (unsigned int *)getPlaneAddres(fbInfo);
	sx_set_operation_all( GXcopy, ~0 );
    }

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    bitblt_fill_area(nbox, pbox, pdstBase, widthDst, dstPlaneOffset,
		     pmask, ^= , label1 );
}

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnORREVERSE
#define	BITBLTFUNC	cfbDoBitbltMemOrReverse
#include "cfbbitblt.h"

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnCOPYINVERTED
#define	BITBLTFUNC	cfbDoBitbltMemCopyInverted
#include "cfbbitblt.h"

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnORINVERTED
#define	BITBLTFUNC	cfbDoBitbltMemOrInverted
#include "cfbbitblt.h"

#undef	ROPFUNC
#undef	BITBLTFUNC
#define	ROPFUNC		fnNAND
#define	BITBLTFUNC	cfbDoBitbltMemNand
#include "cfbbitblt.h"

void
cfbDoBitbltMemSet(pSrcDrawable, pDstDrawable, alu, prgnDst, pptSrc, pmask)
DrawablePtr pSrcDrawable;
DrawablePtr pDstDrawable;
unsigned char alu;
RegionPtr prgnDst;
DDXPointPtr pptSrc;
int	  pmask;
{
    unsigned int  *pdstBase;	/* start of src and dst bitmaps */
    int widthDst;		/* add to get to same position in next line */
    register BoxPtr pbox;
    int nbox;
    int dstPlaneOffset;		/* dstnation drawble plane size */
    int depth = pDstDrawable->depth;
    OmronFbInfoPtr fbInfo;

    initPmask(pmask,depth);

    if(! pmask) return;

    fbInfo = getOmronFbInfo( pSrcDrawable );

    setBaseAndWidth( pDstDrawable, pdstBase, widthDst, dstPlaneOffset );

    if(isCommonPlane(pdstBase,fbInfo))
    {
	pdstBase = (unsigned int *)getPlaneAddres(fbInfo);
	sx_set_operation_all( GXcopy, ~0 );
    }

    pbox = REGION_RECTS(prgnDst);
    nbox = REGION_NUM_RECTS(prgnDst);

    bitblt_fill_area(nbox, pbox, pdstBase, widthDst, dstPlaneOffset,
		     pmask,|= , label1 );
}




FUNCTBL cfbPlaneCopyNtoNTbl[16] =
{
/* GXclear		0x0		 0 */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemClear		},
/* GXand		0x1		 src AND dst */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemAnd 		},
/* GXandReverse		0x2		 src AND NOT dst */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemAndReverse	},
/* GXcopy		0x3		 src */
	{ cfbDoBitbltCopy,	cfbDoBitbltCopy			},
/* GXandInverted	0x4		 NOT src AND dst */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemAndInverted	},
/* GXnoop		0x5		 dst */
	{ cfbDoNoop,		cfbDoNoop			},
/* GXxor		0x6		 src XOR dst */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemXor		},
/* GXor			0x7		 src OR dst */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemOr		},
/* GXnor		0x8		 NOT src AND NOT dst */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemNor		},
/* GXequiv		0x9		 NOT src XOR dst */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemEquiv		},
/* GXinvert		0xa		 NOT dst */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemInvert		},
/* GXorReverse		0xb		 src OR NOT dst */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemOrReverse		},
/* GXcopyInverted	0xc		 NOT src */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemCopyInverted	},
/* GXorInverted		0xd		 NOT src OR dst */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemOrInverted	},
/* GXnand		0xe		 NOT src OR NOT dst */
	{ cfbDoBitbltUseSrc,	cfbDoBitbltMemNand		},
/* GXset		0xf		 1 */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemSet		}
};

FUNCTBL cfbPlaneCopyNto1Tbl[16] =
{
/* GXclear		0x0		 0 */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemClear		},
/* GXand		0x1		 src AND dst */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemAnd 		},
/* GXandReverse		0x2		 src AND NOT dst */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemAndReverse	},
/* GXcopy		0x3		 src */
	{ cfbDoBitbltCopy1,	cfbDoBitbltCopy1		},
/* GXandInverted	0x4		 NOT src AND dst */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemAndInverted	},
/* GXnoop		0x5		 dst */
	{ cfbDoNoop,		cfbDoNoop			},
/* GXxor		0x6		 src XOR dst */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemXor		},
/* GXor			0x7		 src OR dst */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemOr		},
/* GXnor		0x8		 NOT src AND NOT dst */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemNor		},
/* GXequiv		0x9		 NOT src XOR dst */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemEquiv		},
/* GXinvert		0xa		 NOT dst */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemInvert		},
/* GXorReverse		0xb		 src OR NOT dst */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemOrReverse		},
/* GXcopyInverted	0xc		 NOT src */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemCopyInverted	},
/* GXorInverted		0xd		 NOT src OR dst */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemOrInverted	},
/* GXnand		0xe		 NOT src OR NOT dst */
	{ cfbDoBitbltUseSrc1,	cfbDoBitbltMemNand		},
/* GXset		0xf		 1 */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemSet		}
};

#if 0
FUNCTBL cfbPlaneCopy1toNTbl[16] =
{
/* GXclear		0x0		 0 */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemClear		},
/* GXand		0x1		 src AND dst */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8 		},
/* GXandReverse		0x2		 src AND NOT dst */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8		},
/* GXcopy		0x3		 src */
	{ cfbDoBitbltCopy8,	cfbDoBitbltCopy8		},
/* GXandInverted	0x4		 NOT src AND dst */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8		},
/* GXnoop		0x5		 dst */
	{ cfbDoNoop,		cfbDoNoop			},
/* GXxor		0x6		 src XOR dst */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8		},
/* GXor			0x7		 src OR dst */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8		},
/* GXnor		0x8		 NOT src AND NOT dst */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8		},
/* GXequiv		0x9		 NOT src XOR dst */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8		},
/* GXinvert		0xa		 NOT dst */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemInvert		},
/* GXorReverse		0xb		 src OR NOT dst */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8		},
/* GXcopyInverted	0xc		 NOT src */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8		},
/* GXorInverted		0xd		 NOT src OR dst */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8		},
/* GXnand		0xe		 NOT src OR NOT dst */
	{ cfbDoBitbltOther8,	cfbDoBitbltOther8		},
/* GXset		0xf		 1 */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemSet		}
};
#else
FUNCTBL cfbPlaneCopy1toNTbl[16] =
{
/* GXclear		0x0		 0 */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemClear		},
/* GXand		0x1		 src AND dst */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8 		},
/* GXandReverse		0x2		 src AND NOT dst */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8		},
/* GXcopy		0x3		 src */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltCopy8		},
/* GXandInverted	0x4		 NOT src AND dst */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8		},
/* GXnoop		0x5		 dst */
	{ cfbDoNoop,		cfbDoNoop			},
/* GXxor		0x6		 src XOR dst */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8		},
/* GXor			0x7		 src OR dst */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8		},
/* GXnor		0x8		 NOT src AND NOT dst */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8		},
/* GXequiv		0x9		 NOT src XOR dst */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8		},
/* GXinvert		0xa		 NOT dst */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemInvert		},
/* GXorReverse		0xb		 src OR NOT dst */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8		},
/* GXcopyInverted	0xc		 NOT src */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8		},
/* GXorInverted		0xd		 NOT src OR dst */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8		},
/* GXnand		0xe		 NOT src OR NOT dst */
	{ cfbDoBitbltUseSrc8,	cfbDoBitbltOther8		},
/* GXset		0xf		 1 */
	{ cfbDoBitbltNotUseSrc,	cfbDoBitbltMemSet		}
};
#endif


FUNCTBL *cfbPlaneDoBitBlt = cfbPlaneCopyNtoNTbl;

RegionPtr
cfbCopyArea(pSrcDrawable, pDstDrawable,
            pGC, srcx, srcy, width, height, dstx, dsty)
    register DrawablePtr pSrcDrawable;
    register DrawablePtr pDstDrawable;
    GC *pGC;
    int srcx, srcy;
    int width, height;
    int dstx, dsty;
{
    RegionPtr prgnSrcClip;	/* may be a new region, or just a copy */
    Bool freeSrcClip = FALSE;

    RegionPtr prgnExposed;
    RegionRec rgnDst;
    DDXPointPtr pptSrc;
    register DDXPointPtr ppt;
    register BoxPtr pbox;
    int i;
    register int dx;
    register int dy;
    xRectangle origSource;
    DDXPointRec origDest;
    int numRects;
    BoxRec fastBox;
    int fastClip = 0;		/* for fast clipping with pixmap source */
    int fastExpose = 0;		/* for fast exposures with pixmap source */

    origSource.x = srcx;
    origSource.y = srcy;
    origSource.width = width;
    origSource.height = height;
    origDest.x = dstx;
    origDest.y = dsty;

    if ((pSrcDrawable != pDstDrawable) &&
	pSrcDrawable->pScreen->SourceValidate)
    {
	FUNCTBL  *SaveDoBitBlt;

	SaveDoBitBlt = cfbPlaneDoBitBlt;
	cfbPlaneDoBitBlt = cfbPlaneCopyNtoNTbl;
	(*pSrcDrawable->pScreen->SourceValidate) (pSrcDrawable, srcx, srcy, width, height);
	cfbPlaneDoBitBlt = SaveDoBitBlt;
    }

    srcx += pSrcDrawable->x;
    srcy += pSrcDrawable->y;

    /* clip the source */

    if (pSrcDrawable->type == DRAWABLE_PIXMAP)
    {
	if ((pSrcDrawable == pDstDrawable) &&
	    (pGC->clientClipType == CT_NONE))
	{
	    prgnSrcClip = ((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip;
	}
	else
	{
	    fastClip = 1;
	}
    }
    else
    {
	if (pGC->subWindowMode == IncludeInferiors)
	{
	    if (!((WindowPtr) pSrcDrawable)->parent)
	    {
		/*
		 * special case bitblt from root window in
		 * IncludeInferiors mode; just like from a pixmap
		 */
		fastClip = 1;
	    }
	    else if ((pSrcDrawable == pDstDrawable) &&
		(pGC->clientClipType == CT_NONE))
	    {
		prgnSrcClip = ((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip;
	    }
	    else
	    {
		prgnSrcClip = NotClippedByChildren((WindowPtr)pSrcDrawable);
		freeSrcClip = TRUE;
	    }
	}
	else
	{
	    prgnSrcClip = &((WindowPtr)pSrcDrawable)->clipList;
	}
    }

    fastBox.x1 = srcx;
    fastBox.y1 = srcy;
    fastBox.x2 = srcx + width;
    fastBox.y2 = srcy + height;

    /* Don't create a source region if we are doing a fast clip */
    if (fastClip)
    {
	fastExpose = 1;
	/*
	 * clip the source; if regions extend beyond the source size,
 	 * make sure exposure events get sent
	 */
	if (fastBox.x1 < pSrcDrawable->x)
	{
	    fastBox.x1 = pSrcDrawable->x;
	    fastExpose = 0;
	}
	if (fastBox.y1 < pSrcDrawable->y)
	{
	    fastBox.y1 = pSrcDrawable->y;
	    fastExpose = 0;
	}
	if (fastBox.x2 > pSrcDrawable->x + (int) pSrcDrawable->width)
	{
	    fastBox.x2 = pSrcDrawable->x + (int) pSrcDrawable->width;
	    fastExpose = 0;
	}
	if (fastBox.y2 > pSrcDrawable->y + (int) pSrcDrawable->height)
	{
	    fastBox.y2 = pSrcDrawable->y + (int) pSrcDrawable->height;
	    fastExpose = 0;
	}
    }
    else
    {
	(*pGC->pScreen->RegionInit)(&rgnDst, &fastBox, 1);
	(*pGC->pScreen->Intersect)(&rgnDst, &rgnDst, prgnSrcClip);
    }

    dstx += pDstDrawable->x;
    dsty += pDstDrawable->y;

    if (pDstDrawable->type == DRAWABLE_WINDOW)
    {
	if (!((WindowPtr)pDstDrawable)->realized)
	{
	    if (!fastClip)
		(*pGC->pScreen->RegionUninit)(&rgnDst);
	    if (freeSrcClip)
		(*pGC->pScreen->RegionDestroy)(prgnSrcClip);
	    return NULL;
	}
    }

    dx = srcx - dstx;
    dy = srcy - dsty;

    /* Translate and clip the dst to the destination composite clip */
    if (fastClip)
    {
	RegionPtr cclip;

        /* Translate the region directly */
        fastBox.x1 -= dx;
        fastBox.x2 -= dx;
        fastBox.y1 -= dy;
        fastBox.y2 -= dy;

	/* If the destination composite clip is one rectangle we can
	   do the clip directly.  Otherwise we have to create a full
	   blown region and call intersect */

	/* XXX because CopyPlane uses this routine for 8-to-1 bit
	 * copies, this next line *must* also correctly fetch the
	 * composite clip from an mfb gc
	 */

	cclip = ((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip;
        if (REGION_NUM_RECTS(cclip) == 1)
        {
	    BoxPtr pBox = REGION_RECTS(cclip);

	    if (fastBox.x1 < pBox->x1) fastBox.x1 = pBox->x1;
	    if (fastBox.x2 > pBox->x2) fastBox.x2 = pBox->x2;
	    if (fastBox.y1 < pBox->y1) fastBox.y1 = pBox->y1;
	    if (fastBox.y2 > pBox->y2) fastBox.y2 = pBox->y2;

	    /* Check to see if the region is empty */
	    if (fastBox.x1 >= fastBox.x2 || fastBox.y1 >= fastBox.y2)
		(*pGC->pScreen->RegionInit)(&rgnDst, NullBox, 0);
	    else
		(*pGC->pScreen->RegionInit)(&rgnDst, &fastBox, 1);
	}
        else
	{
	    /* We must turn off fastClip now, since we must create
	       a full blown region.  It is intersected with the
	       composite clip below. */
	    fastClip = 0;
	    (*pGC->pScreen->RegionInit)(&rgnDst, &fastBox,1);
	}
    }
    else
    {
        (*pGC->pScreen->TranslateRegion)(&rgnDst, -dx, -dy);
    }

    if (!fastClip)
    {
	(*pGC->pScreen->Intersect)(&rgnDst,
				   &rgnDst,
				 ((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip);
    }

    /* Do bit blitting */
    numRects = REGION_NUM_RECTS(&rgnDst);
    if (numRects && width && height)
    {
	if(!(pptSrc = (DDXPointPtr)ALLOCATE_LOCAL(numRects *
						  sizeof(DDXPointRec))))
	{
	    (*pGC->pScreen->RegionUninit)(&rgnDst);
	    if (freeSrcClip)
		(*pGC->pScreen->RegionDestroy)(prgnSrcClip);
	    return NULL;
	}
	pbox = REGION_RECTS(&rgnDst);
	ppt = pptSrc;
	for (i = numRects; --i >= 0; pbox++, ppt++)
	{
	    ppt->x = pbox->x1 + dx;
	    ppt->y = pbox->y1 + dy;
	}

	if (pDstDrawable->type == DRAWABLE_WINDOW)
	{
	    (*(cfbPlaneDoBitBlt[pGC->alu].fbmem)) (
			pSrcDrawable, pDstDrawable, pGC->alu, &rgnDst, pptSrc,
			pGC->planemask);
	}
	else
	{	
	    (*(cfbPlaneDoBitBlt[pGC->alu].mem)) (
			pSrcDrawable, pDstDrawable, pGC->alu, &rgnDst, pptSrc,
			pGC->planemask);
	}
	DEALLOCATE_LOCAL(pptSrc);
    }

    prgnExposed = NULL;
    if (((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->fExpose)
    {
	extern RegionPtr    miHandleExposures();

        /* Pixmap sources generate a NoExposed (we return NULL to do this) */
        if (!fastExpose)
	    prgnExposed =
		miHandleExposures(pSrcDrawable, pDstDrawable, pGC,
				  origSource.x, origSource.y,
				  (int)origSource.width,
				  (int)origSource.height,
				  origDest.x, origDest.y, (unsigned long)0);
    }
    (*pGC->pScreen->RegionUninit)(&rgnDst);
    if (freeSrcClip)
	(*pGC->pScreen->RegionDestroy)(prgnSrcClip);
    return prgnExposed;
}

RegionPtr cfbCopyPlane(pSrcDrawable, pDstDrawable,
	    pGC, srcx, srcy, width, height, dstx, dsty, bitPlane)
    DrawablePtr 	pSrcDrawable;
    DrawablePtr		pDstDrawable;
    GCPtr		pGC;
    int 		srcx, srcy;
    int 		width, height;
    int 		dstx, dsty;
    unsigned long	bitPlane;
{
    RegionPtr	ret;
    int		diff;
    int 	depth = pDstDrawable->depth;
    int 	srcDepth = pSrcDrawable->depth;
    extern RegionPtr    miHandleExposures();

    diff = (srcDepth != depth);

    if (diff && srcDepth == 1)
    {
	if (bitPlane == 1) 
	{
	    unsigned long fg, bg;

	    cfbPlaneDoBitBlt = cfbPlaneCopy1toNTbl;
	    fg = pGC->fgPixel;
	    bg = pGC->bgPixel;
	    if (checkFgAndBg(fg, bg, pGC->planemask))
	        cfbPlaneSetPixels (fg, bg, depth, pGC->planemask);
    	    ret = cfbCopyArea (pSrcDrawable, pDstDrawable,
	    	    pGC, srcx, srcy, width, height, dstx, dsty);
	    cfbPlaneDoBitBlt = cfbPlaneCopyNtoNTbl;
	}
	else
	    ret = miHandleExposures (pSrcDrawable, pDstDrawable,
	    	pGC, srcx, srcy, width, height, dstx, dsty, bitPlane);
    }
    else if (diff && depth == 1)
    {
	int 	oldalu = pGC->alu;
	int	fg = pGC->fgPixel & 1;
	int	bg = pGC->bgPixel & 1;

    	if (fg == 0 && bg == 1)
	    pGC->alu = InverseAlu[pGC->alu];
    	else if (fg == bg)
	    pGC->alu = mfbReduceRop(pGC->alu, fg);
	copyPlaneBitPlane = findBits(bitPlane,srcDepth);
	cfbPlaneDoBitBlt = cfbPlaneCopyNto1Tbl;
	ret = cfbCopyArea (pSrcDrawable, pDstDrawable,
		    pGC, srcx, srcy, width, height, dstx, dsty);
	cfbPlaneDoBitBlt = cfbPlaneCopyNtoNTbl;
        copyPlaneBitPlane = 0;
	pGC->alu = oldalu;
    }
    else
    {
	PixmapPtr	pBitmap;
	ScreenPtr	pScreen = pSrcDrawable->pScreen;
	GCPtr		pGC1;
	unsigned long	fg, bg;

	pBitmap = (*pScreen->CreatePixmap) (pScreen, width, height, 1);
	if (!pBitmap)
	    return NULL;
	pGC1 = GetScratchGC (1, pScreen);
	if (!pGC1)
	{
	    (*pScreen->DestroyPixmap) (pBitmap);
	    return NULL;
	}
	/*
	 * don't need to set pGC->fgPixel,bgPixel as copyPlane8to1
	 * ignores pixel values, expecting the rop to "do the
	 * right thing", which GXcopy will.
	 */
	ValidateGC ((DrawablePtr) pBitmap, pGC1);
	copyPlaneBitPlane = findBits(bitPlane,srcDepth);
	cfbPlaneDoBitBlt = cfbPlaneCopyNto1Tbl;
	/* no exposures here, scratch GC's don't get graphics expose */
	(void) cfbCopyArea (pSrcDrawable, (DrawablePtr) pBitmap,
			    pGC1, srcx, srcy, width, height, 0, 0);
        copyPlaneBitPlane = 0;
	cfbPlaneDoBitBlt = cfbPlaneCopy1toNTbl;
	fg = pGC->fgPixel;
	bg = pGC->bgPixel;
	if (checkFgAndBg(fg, bg, pGC->planemask))
	    cfbPlaneSetPixels (fg, bg, depth, pGC->planemask);
	/* no exposures here, copy bits from inside a pixmap */
	(void) cfbCopyArea ((DrawablePtr) pBitmap, pDstDrawable, pGC,
			    0, 0, width, height, dstx, dsty);
	cfbPlaneDoBitBlt = cfbPlaneCopyNtoNTbl;
	FreeScratchGC (pGC1);
	(*pScreen->DestroyPixmap) (pBitmap);
	/* compute resultant exposures */
	ret = miHandleExposures (pSrcDrawable, pDstDrawable, pGC,
				 srcx, srcy, width, height,
				 dstx, dsty, bitPlane);
    }
    return ret;
}


void
cfbCopyImagePlane (pSrcDrawable, pDstDrawable, rop, prgnDst, pptSrc, planemask)
    DrawablePtr pSrcDrawable;
    DrawablePtr pDstDrawable;
    int	rop;
    unsigned long planemask;
    RegionPtr prgnDst;
    DDXPointPtr pptSrc;
{
    void (* bitblt)();

    copyPlaneBitPlane = findBits(planemask,pSrcDrawable->depth);
    if (pDstDrawable->type == DRAWABLE_WINDOW)
    {
	bitblt = cfbPlaneCopyNto1Tbl[rop].fbmem;
    }
    else
    {
	bitblt = cfbPlaneCopyNto1Tbl[rop].mem;
    }
    (* (bitblt)) (pSrcDrawable, pDstDrawable, rop, prgnDst,
		  pptSrc, 0xffffffff);
    copyPlaneBitPlane = 0;
}

void
cfbDoBitblt(pSrcDrawable, pDstDrawable, rop, prgnDst, pptSrc, planemask)
    DrawablePtr pSrcDrawable;
    DrawablePtr pDstDrawable;
    int	rop;
    unsigned long planemask;
    RegionPtr prgnDst;
    DDXPointPtr pptSrc;
{
    void (* bitblt)();

    if (pDstDrawable->type == DRAWABLE_WINDOW)
    {
	bitblt = cfbPlaneCopyNtoNTbl[rop].fbmem;
    }
    else
    {
	bitblt = cfbPlaneCopyNtoNTbl[rop].mem;
    }
    (* (bitblt)) (pSrcDrawable, pDstDrawable, rop, prgnDst, pptSrc, planemask);
}
