/************************************************************
Copyright 1987 by Sun Microsystems, Inc. Mountain View, CA.

                    All Rights Reserved

Permission  to  use,  copy,  modify,  and  distribute   this
software  and  its documentation for any purpose and without
fee is hereby granted, provided that the above copyright no-
tice  appear  in all copies and that both that copyright no-
tice and this permission notice appear in  supporting  docu-
mentation,  and  that the names of Sun or MIT not be used in
advertising or publicity pertaining to distribution  of  the
software  without specific prior written permission. Sun and
M.I.T. make no representations about the suitability of this
software for any purpose. It is provided "as is" without any
express or implied warranty.

SUN DISCLAIMS ALL WARRANTIES WITH REGARD TO  THIS  SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
NESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SUN BE  LI-
ABLE  FOR  ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,  DATA  OR
PROFITS,  WHETHER  IN  AN  ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/

/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

#include "X.h"
#include "Xmd.h"
#include "servermd.h"
#include "gcstruct.h"
#include "window.h"
#include "pixmapstr.h"
#include "scrnintstr.h"
#include "windowstr.h"

#include "cfb.h"
#include "cfbmskbits.h"

extern void mfbInvertSolidFS(), mfbBlackSolidFS(), mfbWhiteSolidFS();

/* scanline filling for color frame buffer
   written by drewry, oct 1986 modified by smarks
   changes for compatibility with Little-endian systems Jul 1987; MIT:yba.

   these routines all clip.  they assume that anything that has called
them has already translated the points (i.e. pGC->miTranslate is
non-zero, which is howit gets set in cfbCreateGC().)

   the number of new scnalines created by clipping ==
MaxRectsPerBand * nSpans.

    FillSolid is overloaded to be used for OpaqueStipple as well,
if fgPixel == bgPixel.  
Note that for solids, PrivGC.rop == PrivGC.ropOpStip


    FillTiled is overloaded to be used for OpaqueStipple, if
fgPixel != bgPixel.  based on the fill style, it uses
{RotatedTile, gc.alu} or {RotatedStipple, PrivGC.ropOpStip}
*/

#ifdef	notdef
#include	<stdio.h>
static
dumpspans(n, ppt, pwidth)
    int	n;
    DDXPointPtr ppt;
    int *pwidth;
{
    fprintf(stderr,"%d spans\n", n);
    while (n--) {
	fprintf(stderr, "[%d,%d] %d\n", ppt->x, ppt->y, *pwidth);
	ppt++;
	pwidth++;
    }
    fprintf(stderr, "\n");
}
#endif

#define SOLIDFS(ROP) \
    while (n--) \
    { \
	x = ppt->x; \
	addrl = (int *)(addrlBase + (ppt->y * nlwidth) + (x >> PWSH)); \
	++ppt; \
	width = *pwidth++; \
	if (width) \
	{ \
	    if ( ((x & PIM) + width) < PPW) \
	    { \
		maskpartialbits(x, width, startmask); \
		for (plane = 0; plane < depth; plane++) \
		{ \
		    if (cfbplmk[plane]) \
		    { \
			*addrl = ((ROP(cfbfgmk[plane], *addrl) & startmask) | \
					(*addrl & ~startmask)); \
		    } \
		    addrl += pljmp; \
		} \
	    } \
	    else \
	    { \
		maskbits(x, width, startmask, endmask, nlmiddle); \
		pdst = addrl; \
		for (plane = 0; plane < depth; plane++) \
		{ \
		    if (cfbplmk[plane]) \
		    { \
			addrl = pdst; \
			if ( startmask ) { \
			    *addrl = ((ROP(cfbfgmk[plane], *addrl) & startmask) | \
					(*addrl & ~startmask)); \
			    ++addrl; \
			} \
			nlw = nlmiddle; \
			while ( nlw-- ) { \
			    *addrl = ROP(cfbfgmk[plane], *addrl); \
			    ++addrl; \
			} \
			if ( endmask ) { \
			    *addrl = ((ROP(cfbfgmk[plane], *addrl) & endmask) | \
					(*addrl & ~endmask)); \
			} \
		    } \
		    pdst += pljmp; \
	    	} \
	    } \
    	} \
    }

void
cfbSolidFS(pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted)
    DrawablePtr pDrawable;
    GCPtr	pGC;
    int		nInit;			/* number of spans to fill */
    DDXPointPtr pptInit;		/* pointer to list of start points */
    int		*pwidthInit;		/* pointer to list of n widths */
    int 	fSorted;
{
				/* next three parameters are post-clip */
    int n;			/* number of spans to fill */
    DDXPointPtr ppt;		/* pointer to list of start points */
    register int *pwidth;	/* pointer to list of n widths */
    int *addrlBase;		/* pointer to start of bitmap */
    int nlwidth;		/* width in longwords of bitmap */
    register int *addrl;	/* pointer to current longword in bitmap */
    register int width;		/* current span width */
    register int nlw;
    register int x;
    register unsigned int startmask, endmask;
    int rop;			/* rasterop */
    int depth;
    int nlmiddle;
    unsigned long planemask;
    unsigned long pixel;
    int *pwidthFree;		/* copies of the pointers to free */
    DDXPointPtr pptFree;
    OmronFbInfoPtr  fbInfo;
    int drawFbFlag;

    if (!(pGC->planemask) || pGC->alu == GXnoop)
	return;

    n = nInit * miFindMaxBand(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip);
    pwidthFree = (int *)ALLOCATE_LOCAL(n * sizeof(int));
    pptFree = (DDXPointRec *)ALLOCATE_LOCAL(n * sizeof(DDXPointRec));
    if(!pptFree || !pwidthFree)
    {
	if (pptFree) DEALLOCATE_LOCAL(pptFree);
	if (pwidthFree) DEALLOCATE_LOCAL(pwidthFree);
	return;
    }
#ifdef	notdef
    dumpspans(n, pptInit, pwidthInit);
#endif
    pwidth = pwidthFree;
    ppt = pptFree;
    n = miClipSpans(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip,
		     pptInit, pwidthInit, nInit,
		     ppt, pwidth, fSorted);

#ifdef	notdef
    dumpspans(n, ppt, pwidth);
#endif

    rop = pGC->alu;
    depth = pDrawable->depth;
    planemask = pGC->planemask;
    pixel = pGC->fgPixel;

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	addrlBase = (int *)
		(((PixmapPtr)(pDrawable->pScreen->devPrivate))->devPrivate.ptr);
	nlwidth = (int)
		  (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;
	drawFbFlag = TRUE;
    }
    else
    {
	addrlBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	nlwidth = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;
	fbInfo = getOmronFbInfo(pDrawable);
	if(isCommonPlane( addrlBase, fbInfo ))
	    drawFbFlag = TRUE;
	else
	    drawFbFlag = FALSE;
    }

    if (drawFbFlag == TRUE)
    {
	sx_check_transparent(rop, ALL_ONE, pixel, planemask);
	sx_plane_select(planemask);

	while (n--)
	{
	    x = ppt->x;
	    addrl = (int *)(addrlBase + (ppt->y * nlwidth) + (x >> PWSH));
	    ++ppt;
	    width = *pwidth++;
	    if (!width)
		continue;
	    if ( ((x & PIM) + width) < PPW)
	    {
		maskpartialbits(x, width, startmask);
		*addrl = startmask;
	    }
	    else
	    {
		maskbits(x, width, startmask, endmask, nlw);
		if ( startmask ) {
		    *addrl = startmask;
		    ++addrl;
		}
		Duff(nlw, *addrl++ = ALL_ONE);
		if ( endmask )
		    *addrl = endmask;
	    }
	}
    }
    else
    {
	int pljmp;
	register int plane;
	register int *pdst;

	pljmp = nlwidth * pDrawable->height;

	for (plane = 0; plane < depth; plane++)
	{
	    cfbplmk[plane] = (planemask >> plane) & 0x1;
	    cfbfgmk[plane] = ((pixel >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	}

	switch(rop)
	{
	    case GXcopy:
		SOLIDFS(fnCOPY);
		break;
	    case GXxor:
		SOLIDFS(fnXOR);
		break;
	    case GXand:
		SOLIDFS(fnAND);
		break;
	    case GXor:
		SOLIDFS(fnOR);
		break;
	    case GXset:
		SOLIDFS(fnSET);
		break;
	    case GXclear:
		SOLIDFS(fnCLEAR);
		break;
	    case GXandReverse:
		SOLIDFS(fnANDREVERSE);
		break;
	    case GXandInverted:
		SOLIDFS(fnANDINVERTED);
		break;
	    case GXnor:
		SOLIDFS(fnNOR);
		break;
	    case GXequiv:
		SOLIDFS(fnEQUIV);
		break;
	    case GXinvert:
		SOLIDFS(fnINVERT);
		break;
	    case GXorReverse:
		SOLIDFS(fnORREVERSE);
		break;
	    case GXcopyInverted:
		SOLIDFS(fnCOPYINVERTED);
		break;
	    case GXorInverted:
		SOLIDFS(fnORINVERTED);
		break;
	    case GXnand:
		SOLIDFS(fnNAND);
		break;
	    default:
		break;
	}
    }
    DEALLOCATE_LOCAL(pptFree);
    DEALLOCATE_LOCAL(pwidthFree);
}

#define TILEFS(ROP) \
    while (n--) \
    { \
	iline = (ppt->y - ySrc) % tileHeight; \
	pdst = addrlBase + (ppt->y * nlwidth) + (ppt->x >> PWSH); \
	psrcT = (int *) pTile->devPrivate.ptr + (iline * tlwidth); \
	x = ppt->x; \
	if (*pwidth) \
	{ \
	    width = *pwidth; \
	    while(width > 0) \
	    { \
		psrc = psrcT; \
		w = min(tileWidth, width); \
		if((rem = (x - xSrc)  % tileWidth) != 0) \
		{ \
		    /* if we're in the middle of the tile, get \
		       as many bits as will finish the span, or \
		       as many as will get to the left edge of the tile, \
		       or a longword worth, starting at the appropriate \
		       offset in the tile. \
		    */ \
		    w = min(min(tileWidth - rem, width), BITMAP_SCANLINE_UNIT); \
		    endinc = rem / BITMAP_SCANLINE_UNIT; \
		    tilegandpbitsrop((psrc+endinc), (rem & PIM), (x & PIM), \
				w, pdst, dstpljmp, srcpljmp, depth, ROP); \
		    if((x & PIM) + w >= PPW) \
			pdst++; \
		} \
		else if(((x & PIM) + w) < PPW) \
		{ \
		    /* doing < PPW bits is easy, and worth special-casing */ \
		    tilepbitsrop(psrc, x & PIM, w, pdst, \
				dstpljmp, srcpljmp, depth, ROP); \
		} \
		else \
		{ \
		    /* start at the left edge of the tile, \
   		       and put down as much as we can \
		    */ \
		    maskbits(x, w, startmask, endmask, nlMiddle); \
		    if (startmask) \
			nstart = PPW - (x & PIM); \
		    else \
			nstart = 0; \
		    if (endmask) \
			nend = (x + w)  & PIM; \
		    else \
			nend = 0; \
		    srcStartOver = nstart > PLST; \
		    if(startmask) \
		    { \
			tilepbitsrop(psrc, (x & PIM), nstart, pdst, \
					dstpljmp, srcpljmp, depth, ROP); \
			pdst++; \
			if(srcStartOver) \
			    psrc++; \
		    } \
		    while(nlMiddle--) \
		    { \
			tilegandpbitsrop0(psrc, nstart, PPW, pdst, \
					dstpljmp, srcpljmp, depth, ROP); \
			pdst++; \
			psrc++; \
		    } \
		    if(endmask) \
		    { \
			tilegandpbitsrop0(psrc, nstart, nend, pdst, \
					dstpljmp, srcpljmp, depth, ROP); \
		    } \
		} \
		x += w; \
		width -= w; \
	    } \
	} \
	ppt++; \
	pwidth++; \
    }

/* Fill spans with tiles that aren't 32 bits wide */
void
cfbUnnaturalTileFS(pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted)
DrawablePtr pDrawable;
GC		*pGC;
int		nInit;		/* number of spans to fill */
DDXPointPtr pptInit;		/* pointer to list of start points */
int *pwidthInit;		/* pointer to list of n widths */
int fSorted;
{
    int	iline;			/* first line of tile to use */
				/* next three parameters are post-clip */
    int n;			/* number of spans to fill */
    DDXPointPtr ppt;		/* pointer to list of start points */
    int *pwidth;		/* pointer to list of n widths */
    int *addrlBase;		/* pointer to start of bitmap */
    int nlwidth;		/* width in longwords of bitmap */
    register int *pdst;		/* pointer to current word in bitmap */
    register int *psrc;		/* pointer to current word in tile */
    register int nlMiddle;
    register int nstart;
    unsigned int startmask, endmask;
    unsigned long planemask;
    PixmapPtr pTile;		/* pointer to tile we want to fill with */
    int	w, width, x, xSrc, ySrc, srcStartOver, nend;
    int	tlwidth, rem, tileWidth, tileHeight, endinc;
    int	depth, plane, rop;
    int	srcpljmp, dstpljmp;
    int *psrcT;
    int *pwidthFree;		/* copies of the pointers to free */
    DDXPointPtr pptFree;
    OmronFbInfoPtr fbInfo;


    if (!(pGC->planemask) || pGC->alu == GXnoop)
	return;

    n = nInit * miFindMaxBand(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip);
    pwidthFree = (int *)ALLOCATE_LOCAL(n * sizeof(int));
    pptFree = (DDXPointRec *)ALLOCATE_LOCAL(n * sizeof(DDXPointRec));
    if(!pptFree || !pwidthFree)
    {
	if (pptFree) DEALLOCATE_LOCAL(pptFree);
	if (pwidthFree) DEALLOCATE_LOCAL(pwidthFree);
	return;
    }
    pwidth = pwidthFree;
    ppt = pptFree;
    n = miClipSpans(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip,
		     pptInit, pwidthInit, nInit, 
		     ppt, pwidth, fSorted);

    pTile = pGC->tile.pixmap;
    tlwidth = pTile->devKind >> 2;

    xSrc = pDrawable->x;
    ySrc = pDrawable->y;

    tileWidth = pTile->drawable.width;
    tileHeight = pTile->drawable.height;
    srcpljmp = tlwidth * tileHeight;
    rop = pGC->alu;
    depth = pDrawable->depth;
    planemask = pGC->planemask;

    for (plane = 0; plane < depth; plane++)
    {
	cfbplmk[plane] = (planemask >> plane) & 0x1;
    }

    /* this replaces rotating the tile. Instead we just adjust the offset
     * at which we start grabbing bits from the tile.
     * Ensure that ppt->x - xSrc >= 0 and ppt->y - ySrc >= 0,
     * so that iline and xrem always stay within the tile bounds.
     */
    xSrc += (pGC->patOrg.x % tileWidth) - tileWidth;
    ySrc += (pGC->patOrg.y % tileHeight) - tileHeight;

    fbInfo = getOmronFbInfo(pDrawable);

    if (pDrawable->type == DRAWABLE_WINDOW)
    {

	sx_plane_select(planemask);

	addrlBase = (int *)getPlaneAddres(fbInfo);
	nlwidth = (int)
		  (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;
	dstpljmp = getFbPlaneLongSize();

	while (n--)
	{
	    iline = (ppt->y - ySrc) % tileHeight;
	    pdst = addrlBase + (ppt->y * nlwidth) + (ppt->x >> PWSH);
	    psrcT = (int *) pTile->devPrivate.ptr + (iline * tlwidth);
	    x = ppt->x;

	    if (*pwidth)
	    {
		width = *pwidth;
		while(width > 0)
		{
		    psrc = psrcT;
		    w = min(tileWidth, width);
		    if((rem = (x - xSrc)  % tileWidth) != 0)
		    {
			/* if we're in the middle of the tile, get
			   as many bits as will finish the span, or
			   as many as will get to the left edge of the tile,
			   or a longword worth, starting at the appropriate
			   offset in the tile.
			*/
			w = min(min(tileWidth - rem, width), BITMAP_SCANLINE_UNIT);
			endinc = rem / BITMAP_SCANLINE_UNIT;
			_tilegandpbits((psrc+endinc), (rem & PIM), (x & PIM),
					w, pdst, dstpljmp, srcpljmp, depth, rop);
			if((x & PIM) + w >= PPW)
			    pdst++;
		    }
		    else if(((x & PIM) + w) < PPW)
		    {
			/* doing < PPW bits is easy, and worth special-casing */
			_tilepbits(psrc, x & PIM, w, pdst,
					dstpljmp, srcpljmp, depth, rop);
		    }
		    else
		    {
			/* start at the left edge of the tile,
			   and put down as much as we can
			*/
			maskbits(x, w, startmask, endmask, nlMiddle);

			if (startmask)
			    nstart = PPW - (x & PIM);
			else
			    nstart = 0;
			if (endmask)
			    nend = (x + w)  & PIM;
			else
			    nend = 0;

			srcStartOver = nstart > PLST;

			if(startmask)
			{
			    _tilepbits(psrc, (x & PIM), nstart, pdst,
					dstpljmp, srcpljmp, depth, rop);
			    pdst++;
			    if(srcStartOver)
				psrc++;
			}
 
			while(nlMiddle--)
			{
			    _tilegandpbits0(psrc, nstart, PPW, pdst,
					dstpljmp, srcpljmp, depth, rop);
			    pdst++;
			    psrc++;
			}
			if(endmask)
			{
			    _tilegandpbits0(psrc, nstart, nend, pdst,
					dstpljmp, srcpljmp, depth, rop);
			}
		    }
		    x += w;
		    width -= w;
		}
	    }
	    ppt++;
	    pwidth++;
	}
    }
    else
    {
	addrlBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	nlwidth = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;
	dstpljmp = nlwidth * pDrawable->height;
	if(isCommonPlane( addrlBase, fbInfo ))
	{
	    addrlBase = (int *)getPlaneAddres(fbInfo);
	    sx_set_operation_all( GXcopy, ~0 );
	}

	switch(rop)
	{
	    case GXxor:
		TILEFS(fnXOR);
		break;
	    case GXand:
		TILEFS(fnAND);
		break;
	    case GXor:
		TILEFS(fnOR);
		break;
	    case GXset:
		TILEFS(fnSET);
		break;
	    case GXclear:
		TILEFS(fnCLEAR);
		break;
	    case GXandReverse:
		TILEFS(fnANDREVERSE);
		break;
	    case GXandInverted:
		TILEFS(fnANDINVERTED);
		break;
	    case GXnor:
		TILEFS(fnNOR);
		break;
	    case GXequiv:
		TILEFS(fnEQUIV);
		break;
	    case GXinvert:
		TILEFS(fnINVERT);
		break;
	    case GXorReverse:
		TILEFS(fnORREVERSE);
		break;
	    case GXcopyInverted:
		TILEFS(fnCOPYINVERTED);
		break;
	    case GXorInverted:
		TILEFS(fnORINVERTED);
		break;
	    case GXnand:
		TILEFS(fnNAND);
		break;
	    default:
		break;
	}
    }
    DEALLOCATE_LOCAL(pptFree);
    DEALLOCATE_LOCAL(pwidthFree);
}

#define STIPFS(ROP) \
    STIPFILLSP(ROP, stipgandpbitsrop, stippbitsrop, stipgandpbitsrop0)

#define OPSTIPFS(ROP) \
    STIPFILLSP(ROP, opaqgandpbitsrop, opaqpbitsrop, opaqgandpbitsrop0)

#define STIPFILLSP(ROP, PATERN1, PATERN2, PATERN3) \
    while (n--) \
    { \
	iline = (ppt->y - ySrc) % stippleHeight; \
	pdst = addrlBase + (ppt->y * nlwidth) + (ppt->x >> PWSH); \
	psrcT = (int *) pStipple->devPrivate.ptr + (iline * stwidth); \
	x = ppt->x; \
	if (*pwidth) \
	{ \
	    width = *pwidth; \
	    while(width > 0) \
	    { \
		psrc = psrcT; \
		w = min(stippleWidth, width); \
		if((rem = (x - xSrc) % stippleWidth) != 0) \
		{ \
		    /* if we're in the middle of the tile, get \
		       as many bits as will finish the span, or \
		       as many as will get to the left edge of the tile, \
		       or a longword worth, starting at the appropriate \
		       offset in the tile. \
		    */ \
		    w = min(min(stippleWidth - rem, width), BITMAP_SCANLINE_UNIT); \
		    endinc = rem / BITMAP_SCANLINE_UNIT; \
		    PATERN1((psrc + endinc), (rem & PIM), \
				(x & PIM), w, pdst, dstpljmp, depth, ROP); \
		    if((x & PIM) + w >= PPW) \
			pdst++; \
		} \
		else if(((x & PIM) + w) < PPW) \
		{ \
		    /* doing < PPW bits is easy, and worth special-casing */ \
		    PATERN2(*psrc, x & PIM, w, pdst, dstpljmp, depth, ROP); \
		} \
		else \
		{ \
		    /* start at the left edge of the tile, \
		       and put down as much as we can \
		    */ \
		    maskbits(x, w, startmask, endmask, nlMiddle); \
		    if (startmask) \
			nstart = PPW - (x & PIM); \
		    else \
			nstart = 0; \
		    if (endmask) \
			nend = (x + w)  & PIM; \
		    else \
			nend = 0; \
		    srcStartOver = nstart > PLST; \
		    if(startmask) \
		    { \
			PATERN2(*psrc, (x & PIM), nstart, pdst, \
						dstpljmp, depth, ROP); \
			pdst++; \
			if(srcStartOver) \
			    psrc++; \
		    } \
		    while(nlMiddle--) \
		    { \
			PATERN3(psrc, nstart, PPW, pdst, \
						dstpljmp, depth, ROP); \
			pdst++; \
			psrc++; \
		    } \
		    if(endmask) \
		    { \
			PATERN3(psrc, nstart, nend, pdst, \
						dstpljmp, depth, ROP); \
		    } \
		} \
		x += w; \
		width -= w; \
	    } \
	} \
	ppt++; \
	pwidth++; \
    }

/* Fill spans with stipples that aren't 32 bits wide */
void
cfbUnnaturalStipFS(pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted)
DrawablePtr pDrawable;
GC		*pGC;
int		nInit;		/* number of spans to fill */
DDXPointPtr pptInit;		/* pointer to list of start points */
int *pwidthInit;		/* pointer to list of n widths */
int fSorted;
{
    int	iline;			/* first line of tile to use */
				/* next three parameters are post-clip */
    int n;			/* number of spans to fill */
    DDXPointPtr ppt;		/* pointer to list of start points */
    int *pwidth;		/* pointer to list of n widths */
    int	*addrlBase;		/* pointer to start of bitmap */
    int	 nlwidth;		/* width in longwords of bitmap */
    register int *pdst;
    register int *psrc;		/* pointer to current word in tile */
    register int nlMiddle;
    register int nstart;
    unsigned int startmask, endmask;
    PixmapPtr	pStipple;	/* pointer to stipple we want to fill with */
    int	w, width, x, xSrc, ySrc, srcStartOver, nend;
    int	stwidth, rem, stippleWidth, stippleHeight, endinc;
    int	*psrcT;
    unsigned long planemask, pixel;
    int rop;
    int depth;
    int *pwidthFree;		/* copies of the pointers to free */
    DDXPointPtr pptFree;

    if (!(pGC->planemask) || pGC->alu == GXnoop)
	return;

    n = nInit * miFindMaxBand(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip);
    pwidthFree = (int *)ALLOCATE_LOCAL(n * sizeof(int));
    pptFree = (DDXPointRec *)ALLOCATE_LOCAL(n * sizeof(DDXPointRec));
    if(!pptFree || !pwidthFree)
    {
	if (pptFree) DEALLOCATE_LOCAL(pptFree);
	if (pwidthFree) DEALLOCATE_LOCAL(pwidthFree);
	return;
    }
    pwidth = pwidthFree;
    ppt = pptFree;
    n = miClipSpans(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip,
		     pptInit, pwidthInit, nInit, 
		     ppt, pwidth, fSorted);

    rop = pGC->alu;
    depth = pDrawable->depth;
    planemask = pGC->planemask;
    pixel = pGC->fgPixel;
    pStipple = pGC->stipple;

    if (pStipple->drawable.depth != 1) {
	FatalError( "Stipple depth not equal to 1!\n" );
    }

    stwidth = pStipple->devKind >> 2;
    stippleWidth = pStipple->drawable.width;
    stippleHeight = pStipple->drawable.height;

    xSrc = pDrawable->x;
    ySrc = pDrawable->y;

    /* this replaces rotating the stipple. Instead we just adjust the offset
     * at which we start grabbing bits from the stipple.
     * Ensure that ppt->x - xSrc >= 0 and ppt->y - ySrc >= 0,
     * so that iline and xrem always stay within the stipple bounds.
     */
    xSrc += (pGC->patOrg.x % stippleWidth) - stippleWidth;
    ySrc += (pGC->patOrg.y % stippleHeight) - stippleHeight;

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	addrlBase = (int *)
		(((PixmapPtr)(pDrawable->pScreen->devPrivate))->devPrivate.ptr);
	nlwidth = (int)
		(((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;

	sx_check_transparent(rop, ALL_ONE, pixel, planemask);
	sx_plane_select(planemask);

	while (n--)
	{
	    iline = (ppt->y - ySrc) % stippleHeight;
	    pdst = addrlBase + (ppt->y * nlwidth) + (ppt->x >> PWSH);
	    psrcT = (int *) pStipple->devPrivate.ptr + (iline * stwidth);
	    x = ppt->x;

	    if (*pwidth)
	    {
		width = *pwidth;
		while(width > 0)
		{
		    psrc = psrcT;
		    w = min(stippleWidth, width);
		    if((rem = (x - xSrc) % stippleWidth) != 0)
		    {
			/* if we're in the middle of the tile, get
   			   as many bits as will finish the span, or
			   as many as will get to the left edge of the tile,
			   or a longword worth, starting at the appropriate
			   offset in the tile.
			*/
			w = min(min(stippleWidth - rem, width), BITMAP_SCANLINE_UNIT);
			endinc = rem / BITMAP_SCANLINE_UNIT;
			_getandputbits((psrc + endinc), (rem & PIM), (x & PIM),
				w, pdst, BitOn);
			if((x & PIM) + w >= PPW)
			    pdst++;
		    }
		    else if(((x & PIM) + w) < PPW)
		    {
			/* doing < PPW bits is easy, and worth special-casing */
			_putbits(*psrc, x & PIM, w, pdst);
		    }
		    else
		    {
			/* start at the left edge of the tile,
			   and put down as much as we can
			*/
			maskbits(x, w, startmask, endmask, nlMiddle);

			if (startmask)
			    nstart = PPW - (x & PIM);
			else
			    nstart = 0;
			if (endmask)
			    nend = (x + w)  & PIM;
			else
			    nend = 0;

			srcStartOver = nstart > PLST;

			if(startmask)
			{
			    _putbits(*psrc, (x & PIM), nstart, pdst);
			    pdst++;
			    if(srcStartOver)
				psrc++;
			}
 
			while(nlMiddle--)
			{
			    _getandputbits0(psrc, nstart, PPW, pdst, BitOn);
			    pdst++;
			    psrc++;
			}
			if(endmask)
			{
			    _getandputbits0(psrc, nstart, nend, pdst, BitOn);
			}
 		    }
		    x += w;
		    width -= w;
		}
	    }
	    ppt++;
	    pwidth++;
	}
    }
    else
    {
	int plane, dstpljmp;
	OmronFbInfoPtr fbInfo;

	addrlBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	nlwidth = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;
	dstpljmp = nlwidth * pDrawable->height;
	fbInfo = getOmronFbInfo(pDrawable);
	if(isCommonPlane( addrlBase, fbInfo ))
	{
	    addrlBase = (int *)getPlaneAddres(fbInfo);
	    sx_set_operation_all( GXcopy, ~0 );
	}

	for (plane = 0; plane < depth; plane++)
	{
	    cfbfgmk[plane] = ((pixel >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	    cfbplmk[plane] = (planemask >> plane) & 0x1;
	}

	switch(rop)
	{
	    case GXcopy:
		STIPFS(fnCOPY);
		break;
	    case GXxor:
		STIPFS(fnXOR);
		break;
	    case GXand:
		STIPFS(fnAND);
		break;
	    case GXor:
		STIPFS(fnOR);
		break;
	    case GXset:
		STIPFS(fnSET);
		break;
	    case GXclear:
		STIPFS(fnCLEAR);
		break;
	    case GXandReverse:
		STIPFS(fnANDREVERSE);
		break;
	    case GXandInverted:
		STIPFS(fnANDINVERTED);
		break;
	    case GXnor:
		STIPFS(fnNOR);
		break;
	    case GXequiv:
		STIPFS(fnEQUIV);
		break;
	    case GXinvert:
		STIPFS(fnINVERT);
		break;
	    case GXorReverse:
		STIPFS(fnORREVERSE);
		break;
	    case GXcopyInverted:
		STIPFS(fnCOPYINVERTED);
		break;
	    case GXorInverted:
		STIPFS(fnORINVERTED);
		break;
	    case GXnand:
		STIPFS(fnNAND);
		break;
	    default:
		break;
	}
    }
    DEALLOCATE_LOCAL(pptFree);
    DEALLOCATE_LOCAL(pwidthFree);
}

#define compute_1toN_operation( alu, fg, bg, pm1, pm2, pm3, pm4, pmask) \
{ \
	unsigned long diff = fg ^ bg; \
	unsigned long same = ~diff; \
	pm1 = same & (~fg); \
	pm2 = diff & bg; \
	pm3 = diff & fg; \
	pm4 = same & fg; \
	if(conv_op_table[alu] == GXnoop) \
		pmask &= ~pm1; \
	if(invert_op_table[alu] == GXnoop) \
		pmask &= ~pm2; \
	if(const_src_op_table[alu] == GXnoop) \
		pmask &= ~pm4; \
}

#define	sx_set_1toN_operation(alu, mask, pm1, pm2, pm3, pm4, pmask) \
{ \
	sx_plane_select(pm1); \
	*sx_conv_op_cont0[alu] = mask; \
	sx_plane_select(pm2); \
	*sx_invert_op_cont0[alu] = mask; \
	sx_plane_select(pm3); \
	*sx_op_cont0[alu] = mask; \
	sx_plane_select(pm4); \
	*sx_const_src_op_cont0[alu] = mask; \
	sx_plane_select(pmask); \
}


/* Fill spans with stipples that aren't 32 bits wide */
void
cfbUnnaturalOpStipFS(pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted)
DrawablePtr pDrawable;
GC		*pGC;
int		nInit;		/* number of spans to fill */
DDXPointPtr pptInit;		/* pointer to list of start points */
int *pwidthInit;		/* pointer to list of n widths */
int fSorted;
{
    int	iline;			/* first line of tile to use */
				/* next three parameters are post-clip */
    int n;			/* number of spans to fill */
    DDXPointPtr ppt;		/* pointer to list of start points */
    int *pwidth;		/* pointer to list of n widths */
    int	*addrlBase;		/* pointer to start of bitmap */
    int	 nlwidth;		/* width in longwords of bitmap */
    register int *pdst;
    register int *psrc;		/* pointer to current word in tile */
    register int nlMiddle;
    register int nstart;
    unsigned int startmask, endmask;
    unsigned int tmp;
    PixmapPtr pStipple;	/* pointer to stipple we want to fill with */
    int	w, width, x, xSrc, ySrc, srcStartOver, nend;
    int	stwidth, rem, stippleWidth, stippleHeight, endinc;
    int *psrcT;
    unsigned long pmask;
    unsigned long pm1, pm2, pm3, pm4;
    unsigned long fg, bg;
    int rop;
    int depth;
    int nlw;
    int *pwidthFree;		/* copies of the pointers to free */
    DDXPointPtr pptFree;

    if (!(pGC->planemask) || pGC->alu == GXnoop)
	return;

    n = nInit * miFindMaxBand(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip);
    pwidthFree = (int *)ALLOCATE_LOCAL(n * sizeof(int));
    pptFree = (DDXPointRec *)ALLOCATE_LOCAL(n * sizeof(DDXPointRec));
    if(!pptFree || !pwidthFree)
    {
	if (pptFree) DEALLOCATE_LOCAL(pptFree);
	if (pwidthFree) DEALLOCATE_LOCAL(pwidthFree);
	return;
    }
    pwidth = pwidthFree;
    ppt = pptFree;
    n = miClipSpans(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip,
		     pptInit, pwidthInit, nInit, 
		     ppt, pwidth, fSorted);

    rop = pGC->alu;
    depth = pDrawable->depth;
    pmask = pGC->planemask;
    fg = pGC->fgPixel;
    bg = pGC->bgPixel;
    pStipple = pGC->stipple;

    if (pStipple->drawable.depth != 1) {
	FatalError( "OpaqueStipple depth not equal to 1!\n" );
    }

    stwidth = pStipple->devKind >> 2;
    stippleWidth = pStipple->drawable.width;
    stippleHeight = pStipple->drawable.height;

    xSrc = pDrawable->x;
    ySrc = pDrawable->y;

    /* this replaces rotating the stipple. Instead we just adjust the offset
     * at which we start grabbing bits from the stipple.
     * Ensure that ppt->x - xSrc >= 0 and ppt->y - ySrc >= 0,
     * so that iline and xrem always stay within the stipple bounds.
     */
    xSrc += (pGC->patOrg.x % stippleWidth) - stippleWidth;
    ySrc += (pGC->patOrg.y % stippleHeight) - stippleHeight;

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	addrlBase = (int *)
		(((PixmapPtr)(pDrawable->pScreen->devPrivate))->devPrivate.ptr);
	nlwidth = (int)
		(((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;

	compute_1toN_operation( rop, fg, bg, pm1, pm2, pm3, pm4, pmask);

	while (n--)
	{
	    iline = (ppt->y - ySrc) % stippleHeight;
	    pdst = addrlBase + (ppt->y * nlwidth) + (ppt->x >> PWSH);
	    psrcT = (int *) pStipple->devPrivate.ptr + (iline * stwidth);
	    x = ppt->x;

	    if (*pwidth)
	    {
		width = *pwidth;
		while(width > 0)
		{
		    psrc = psrcT;
		    w = min(stippleWidth, width);
		    if((rem = (x - xSrc) % stippleWidth) != 0)
		    {
			/* if we're in the middle of the tile, get
   			   as many bits as will finish the span, or
			   as many as will get to the left edge of the tile,
			   or a longword worth, starting at the appropriate
			   offset in the tile.
			*/
			w = min(min(stippleWidth - rem, width), BITMAP_SCANLINE_UNIT);
			endinc = rem / BITMAP_SCANLINE_UNIT;

			if(((x & PIM) + w) < PPW)
			{
			    maskpartialbits(x, w, startmask);
		            sx_set_1toN_operation(rop, startmask, pm1, pm2, pm3, pm4,
						       pmask);
			    sx_plane_select(pmask);
    			    getbits((psrc + endinc), (rem & PIM), w, tmp); 
			    *pdst = SCRRIGHT(tmp, (x & PIM));
			}
			else
			{
			    mask32bits(x, w, startmask, endmask);
		            sx_set_1toN_operation(rop, startmask, pm1, pm2, pm3, pm4,
						       pmask);
			    sx_plane_select(pmask);
    			    getbits((psrc + endinc), (rem & PIM), w, tmp); 
			    *pdst++ = SCRRIGHT(tmp, (x & PIM));
		            sx_set_1toN_operation(rop, endmask, pm1, pm2, pm3, pm4,
						       pmask);
			    sx_plane_select(pmask);
    			    getbits((psrc + endinc), (rem & PIM), w, tmp); 
			    *pdst = SCRLEFT(tmp, (32 - (x & PIM)));
			}
		    }
		    else if(((x & PIM) + w) < PPW)
		    {
			/* doing < PPW bits is easy, and worth special-casing */
			maskpartialbits(x, w, startmask);
		        sx_set_1toN_operation(rop, startmask, pm1, pm2, pm3, pm4,
						   pmask);
			sx_plane_select(pmask);
			_putbits(*psrc, x & PIM, w, pdst);
		    }
		    else
		    {
			int *tmpsrc, *tmpdst;

			/* start at the left edge of the tile,
			   and put down as much as we can
			*/
			maskbits(x, w, startmask, endmask, nlw);

			if (startmask)
			    nstart = PPW - (x & PIM);
			else
			    nstart = 0;
			if (endmask)
			    nend = (x + w)  & PIM;
			else
			    nend = 0;

			srcStartOver = nstart > PLST;

			tmpsrc = psrc;
			tmpdst = pdst;

			if(startmask)
			{
		             sx_set_1toN_operation(rop, startmask, pm1, pm2, 
							   pm3, pm4, pmask);
			     sx_plane_select(pmask);
			     _putbits(*psrc, (x & PIM), nstart, pdst);
			     pdst++;
			     if(srcStartOver)
				psrc++;
			}
			nlMiddle = nlw;
			while(nlMiddle--)
			{
		             sx_set_1toN_operation(rop, ~0, pm1, pm2, 
							   pm3, pm4, pmask);
			     sx_plane_select(pmask);
			     _getandputbits0(psrc, nstart, PPW, pdst, BitOn);
			     pdst++;
			     psrc++;
			}
			if(endmask)
			{
		             sx_set_1toN_operation(rop, endmask, pm1, pm2, 
							   pm3, pm4, pmask);
			     sx_plane_select(pmask);
			     _getandputbits0(psrc, nstart, nend, pdst, BitOn);
			}
 		    }
		    x += w;
		    width -= w;
		}
	    }
	    ppt++;
	    pwidth++;
	}
    }
    else
    {
	int plane, dstpljmp;
	OmronFbInfoPtr fbInfo;

	addrlBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	nlwidth = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;
	dstpljmp = nlwidth * pDrawable->height;
	fbInfo = getOmronFbInfo(pDrawable);
	if(isCommonPlane( addrlBase, fbInfo ))
	{
	    addrlBase = (int *)getPlaneAddres(fbInfo);
	    sx_set_operation_all( GXcopy, ~0 );
	}

	for (plane = 0; plane < depth; plane++)
	{
	    cfbfgmk[plane] = ((fg >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	    cfbbgmk[plane] = ((bg >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	    cfbplmk[plane] = (pmask >> plane) & 0x1;
	}

	switch(rop)
	{
	    case GXxor:
		OPSTIPFS(fnXOR);
		break;
	    case GXand:
		OPSTIPFS(fnAND);
		break;
	    case GXor:
		OPSTIPFS(fnOR);
		break;
	    case GXset:
		OPSTIPFS(fnSET);
		break;
	    case GXclear:
		OPSTIPFS(fnCLEAR);
		break;
	    case GXandReverse:
		OPSTIPFS(fnANDREVERSE);
		break;
	    case GXandInverted:
		OPSTIPFS(fnANDINVERTED);
		break;
	    case GXnor:
		OPSTIPFS(fnNOR);
		break;
	    case GXequiv:
		OPSTIPFS(fnEQUIV);
		break;
	    case GXinvert:
		OPSTIPFS(fnINVERT);
		break;
	    case GXorReverse:
		OPSTIPFS(fnORREVERSE);
		break;
	    case GXcopyInverted:
		OPSTIPFS(fnCOPYINVERTED);
		break;
	    case GXorInverted:
		OPSTIPFS(fnORINVERTED);
		break;
	    case GXnand:
		OPSTIPFS(fnNAND);
		break;
	    default:
		break;
	}
    }
    DEALLOCATE_LOCAL(pptFree);
    DEALLOCATE_LOCAL(pwidthFree);
}
