/************************************************************
Copyright 1987 by Sun Microsystems, Inc. Mountain View, CA.

                    All Rights Reserved

Permission  to  use,  copy,  modify,  and  distribute   this
software  and  its documentation for any purpose and without
fee is hereby granted, provided that the above copyright no-
tice  appear  in all copies and that both that copyright no-
tice and this permission notice appear in  supporting  docu-
mentation,  and  that the names of Sun or MIT not be used in
advertising or publicity pertaining to distribution  of  the
software  without specific prior written permission. Sun and
M.I.T. make no representations about the suitability of this
software for any purpose. It is provided "as is" without any
express or implied warranty.

SUN DISCLAIMS ALL WARRANTIES WITH REGARD TO  THIS  SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
NESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SUN BE  LI-
ABLE  FOR  ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,  DATA  OR
PROFITS,  WHETHER  IN  AN  ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/

/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

#include "X.h"
#include "Xmd.h"
#include "servermd.h"
#include "gcstruct.h"
#include "window.h"
#include "pixmapstr.h"
#include "scrnintstr.h"
#include "windowstr.h"

#include "cfb.h"
#include "cfbmskbits.h"

/* Fill spans with tiles that aren't 32 bits wide */
void
cfbUnnaturalTileCopyFS(pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted)
DrawablePtr pDrawable;
GC		*pGC;
int		nInit;		/* number of spans to fill */
DDXPointPtr pptInit;		/* pointer to list of start points */
int *pwidthInit;		/* pointer to list of n widths */
int fSorted;
{
    int	iline;			/* first line of tile to use */
				/* next three parameters are post-clip */
    int n;			/* number of spans to fill */
    DDXPointPtr ppt;		/* pointer to list of start points */
    int *pwidth;		/* pointer to list of n widths */
    int *addrlBase;		/* pointer to start of bitmap */
    int nlwidth;		/* width in longwords of bitmap */
    register int *pdst;		/* pointer to current word in bitmap */
    register int *psrc;		/* pointer to current word in tile */
    int srcpljmp, dstpljmp;
    register int nlMiddle;
    register int nstart;
    unsigned int startmask, endmask;
    unsigned long planemask;
    PixmapPtr	pTile;		/* pointer to tile we want to fill with */
    int	w, width, x, xSrc, ySrc, srcStartOver, nend;
    int	tlwidth, rem, tileWidth, tileHeight, endinc;
    int	depth, plane;
    int	*psrcT;
    int *pwidthFree;		/* copies of the pointers to free */
    DDXPointPtr pptFree;
    OmronFbInfoPtr fbInfo;

    if (!(pGC->planemask))
	return;

    n = nInit * miFindMaxBand(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip);
    pwidthFree = (int *)ALLOCATE_LOCAL(n * sizeof(int));
    pptFree = (DDXPointRec *)ALLOCATE_LOCAL(n * sizeof(DDXPointRec));
    if(!pptFree || !pwidthFree)
    {
	if (pptFree) DEALLOCATE_LOCAL(pptFree);
	if (pwidthFree) DEALLOCATE_LOCAL(pwidthFree);
	return;
    }
    pwidth = pwidthFree;
    ppt = pptFree;
    n = miClipSpans(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip,
		     pptInit, pwidthInit, nInit, 
		     ppt, pwidth, fSorted);

    pTile = pGC->tile.pixmap;
    tlwidth = pTile->devKind >> 2;

    xSrc = pDrawable->x;
    ySrc = pDrawable->y;

    tileWidth = pTile->drawable.width;
    tileHeight = pTile->drawable.height;
    srcpljmp = tlwidth * tileHeight;
    depth = pDrawable->depth;
    planemask = pGC->planemask;

    for (plane = 0; plane < depth; plane++)
    {
	cfbplmk[plane] = (planemask >> plane) & 0x1;
    }

    /* this replaces rotating the tile. Instead we just adjust the offset
     * at which we start grabbing bits from the tile.
     * Ensure that ppt->x - xSrc >= 0 and ppt->y - ySrc >= 0,
     * so that iline and xrem always stay within the tile bounds.
     */
    xSrc += (pGC->patOrg.x % tileWidth) - tileWidth;
    ySrc += (pGC->patOrg.y % tileHeight) - tileHeight;

    fbInfo = getOmronFbInfo(pDrawable);

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	addrlBase = (int *)getPlaneAddres(fbInfo);
	nlwidth = (int)
		  (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;
	dstpljmp = getFbPlaneLongSize();

#ifdef FASTPUTBITS
	sx_set_operation_all(GXcopy, ALL_ONE);
#else
	sx_plane_select(planemask);
#endif

	while (n--)
	{
	    iline = (ppt->y - ySrc) % tileHeight;
	    pdst = addrlBase + (ppt->y * nlwidth) + (ppt->x >> PWSH);
	    psrcT = (int *) pTile->devPrivate.ptr + (iline * tlwidth);
	    x = ppt->x;

	    if (*pwidth)
	    {
		width = *pwidth;
		while(width > 0)
		{
		    psrc = psrcT;
		    w = min(tileWidth, width);
		    if((rem = (x - xSrc)  % tileWidth) != 0)
		    {
			/* if we're in the middle of the tile, get
			   as many bits as will finish the span, or
			   as many as will get to the left edge of the tile,
			   or a longword worth, starting at the appropriate
			   offset in the tile.
			*/
			w = min(min(tileWidth - rem, width), BITMAP_SCANLINE_UNIT);
			endinc = rem / BITMAP_SCANLINE_UNIT;
			_TILEGANDPBITS((psrc+endinc), (rem & PIM), (x & PIM),
				w, pdst, dstpljmp, srcpljmp, depth);
			if((x & PIM) + w >= 0x20)
			    pdst++;
		    }
		    else if(((x & PIM) + w) < PPW)
		    {
			/* doing < PPW bits is easy, and worth special-casing */
			_TILEPBITS(psrc, x & PIM, w, pdst, dstpljmp, srcpljmp, depth);
		    }
		    else
		    {
			/* start at the left edge of the tile,
			   and put down as much as we can
			*/
			maskbits(x, w, startmask, endmask, nlMiddle);

			if (startmask)
			    nstart = PPW - (x & PIM);
			else
			    nstart = 0;
			if (endmask)
			    nend = (x + w)  & PIM;
			else
			    nend = 0;

			srcStartOver = nstart > PLST;

			if(startmask)
			{
			    _TILEPBITS(psrc, (x & PIM), nstart, pdst,
						dstpljmp, srcpljmp, depth);
			    pdst++;
			    if(srcStartOver)
				psrc++;
			}
 
			while(nlMiddle--)
			{
			    _TILEGANDPBITS0(psrc, nstart, PPW, pdst,
						dstpljmp, srcpljmp, depth);
			    pdst++;
			    psrc++;
			}
			if(endmask)
			{
			    _TILEGANDPBITS0(psrc, nstart, nend, pdst,
						dstpljmp, srcpljmp, depth);
			}
		    }
		    x += w;
		    width -= w;
		}
	    }
	    ppt++;
	    pwidth++;
	}
    }
    else
    {
	addrlBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	nlwidth = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;
	dstpljmp = nlwidth * pDrawable->height;
	if(isCommonPlane( addrlBase, fbInfo ))
	{
	    addrlBase = (int *)getPlaneAddres(fbInfo);
	    sx_set_operation_all( GXcopy, ~0 );
	}

	while (n--)
	{
	    iline = (ppt->y - ySrc) % tileHeight;
	    pdst = addrlBase + (ppt->y * nlwidth) + (ppt->x >> PWSH);
	    psrcT = (int *) pTile->devPrivate.ptr + (iline * tlwidth);
	    x = ppt->x;

	    if (*pwidth)
	    {
		width = *pwidth;
		while(width > 0)
		{
		    psrc = psrcT;
		    w = min(tileWidth, width);
		    if((rem = (x - xSrc)  % tileWidth) != 0)
		    {
			/* if we're in the middle of the tile, get
			   as many bits as will finish the span, or
			   as many as will get to the left edge of the tile,
			   or a longword worth, starting at the appropriate
			   offset in the tile.
			*/
			w = min(min(tileWidth - rem, width), BITMAP_SCANLINE_UNIT);
			endinc = rem / BITMAP_SCANLINE_UNIT;
			tilegandpbits((psrc+endinc), (rem & PIM), (x & PIM),
			    w, pdst, dstpljmp, srcpljmp, depth);
			if((x & PIM) + w >= 0x20)
			    pdst++;
		    }
		    else if(((x & PIM) + w) < PPW)
		    {
			/* doing < PPW bits is easy, and worth special-casing */
			tilepbits(psrc, x & PIM, w, pdst,
					dstpljmp, srcpljmp, depth);
		    }
		    else
		    {
			/* start at the left edge of the tile,
			   and put down as much as we can
			*/
			maskbits(x, w, startmask, endmask, nlMiddle);

			if (startmask)
			    nstart = PPW - (x & PIM);
			else
			    nstart = 0;
			if (endmask)
			    nend = (x + w) & PIM;
			else
			    nend = 0;

			srcStartOver = nstart > PLST;

			if(startmask)
			{
			    tilepbits(psrc, (x & PIM), nstart, pdst,
						dstpljmp, srcpljmp, depth);
			    pdst++;
			    if(srcStartOver)
				psrc++;
			}
 
			while(nlMiddle--)
			{
			    tilegandpbits0(psrc, nstart, PPW, pdst,
						dstpljmp, srcpljmp, depth);
			    pdst++;
			    psrc++;
			}
			if(endmask)
			{
			    tilegandpbits0(psrc, nstart, nend, pdst,
						dstpljmp, srcpljmp, depth);
			}
		    }
		    x += w;
		    width -= w;
		}
	    }
	    ppt++;
	    pwidth++;
	}
    }
    DEALLOCATE_LOCAL(pptFree);
    DEALLOCATE_LOCAL(pwidthFree);
}

/* Fill spans with opaquestipples that aren't 32 bits wide */
void
cfbUnnaturalOpStipCopyFS(pDrawable, pGC, nInit, pptInit, pwidthInit, fSorted)
DrawablePtr pDrawable;
GC		*pGC;
int		nInit;		/* number of spans to fill */
DDXPointPtr pptInit;		/* pointer to list of start points */
int *pwidthInit;		/* pointer to list of n widths */
int fSorted;
{
    int	iline;			/* first line of tile to use */
				/* next three parameters are post-clip */
    int n;			/* number of spans to fill */
    DDXPointPtr ppt;		/* pointer to list of start points */
    int *pwidth;		/* pointer to list of n widths */
    int	*addrlBase;		/* pointer to start of bitmap */
    int nlwidth;		/* width in longwords of bitmap */
    register int *pdst;
    register int *psrc;		/* pointer to current word in tile */
    register int nlMiddle;
    register int nstart;
    unsigned int startmask,endmask;
    PixmapPtr	pStipple;	/* pointer to stipple we want to fill with */
    int	w, width, x, xSrc, ySrc, srcStartOver, nend;
    int	stwidth, rem, stippleWidth, stippleHeight, endinc;
    int nlw;
    int depth;
    int	*psrcT;
    unsigned long planemask;
    unsigned long fg, bg;
    int *pwidthFree;		/* copies of the pointers to free */
    DDXPointPtr pptFree;

    if (!pGC->planemask)
	return;

    n = nInit * miFindMaxBand(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip);
    pwidthFree = (int *)ALLOCATE_LOCAL(n * sizeof(int));
    pptFree = (DDXPointRec *)ALLOCATE_LOCAL(n * sizeof(DDXPointRec));
    if(!pptFree || !pwidthFree)
    {
	if (pptFree) DEALLOCATE_LOCAL(pptFree);
	if (pwidthFree) DEALLOCATE_LOCAL(pwidthFree);
	return;
    }
    pwidth = pwidthFree;
    ppt = pptFree;
    n = miClipSpans(((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip,
		     pptInit, pwidthInit, nInit, 
		     ppt, pwidth, fSorted);

    depth = pDrawable->depth;
    planemask = pGC->planemask;
    fg = pGC->fgPixel;
    bg = pGC->bgPixel;
    pStipple = pGC->stipple;

    if (pStipple->drawable.depth != 1) {
	FatalError( "OpaqueStipple depth not equal to 1!\n" );
    }

    stwidth = pStipple->devKind >> 2;
    stippleWidth = pStipple->drawable.width;
    stippleHeight = pStipple->drawable.height;

    xSrc = pDrawable->x;
    ySrc = pDrawable->y;

    /* this replaces rotating the stipple. Instead we just adjust the offset
     * at which we start grabbing bits from the stipple.
     * Ensure that ppt->x - xSrc >= 0 and ppt->y - ySrc >= 0,
     * so that iline and xrem always stay within the stipple bounds.
     */
    xSrc += (pGC->patOrg.x % stippleWidth) - stippleWidth;
    ySrc += (pGC->patOrg.y % stippleHeight) - stippleHeight;

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	addrlBase = (int *)
		(((PixmapPtr)(pDrawable->pScreen->devPrivate))->devPrivate.ptr);
	nlwidth = (int)
		(((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;

	while (n--)
	{
	    iline = (ppt->y - ySrc) % stippleHeight;
	    pdst = addrlBase + (ppt->y * nlwidth) + (ppt->x >> PWSH);
	    psrcT = (int *) pStipple->devPrivate.ptr + (iline * stwidth);
	    x = ppt->x;

	    if (*pwidth)
	    {
		width = *pwidth;
		while(width > 0)
		{
		    psrc = psrcT;
		    w = min(stippleWidth, width);
		    if((rem = (x - xSrc) % stippleWidth) != 0)
		    {
			/* if we're in the middle of the tile, get
   			   as many bits as will finish the span, or
			   as many as will get to the left edge of the tile,
			   or a longword worth, starting at the appropriate
			   offset in the tile.
			*/
			w = min(min(stippleWidth - rem, width), BITMAP_SCANLINE_UNIT);
			endinc = rem / BITMAP_SCANLINE_UNIT;

			/* drawing fgpixel */
			sx_set_transparent(GXcopy, ALL_ONE, fg);
			sx_plane_select(planemask);
			_getandputbits((psrc + endinc), (rem & PIM),
					(x & PIM), w, pdst, BitOn);

			/* drawing bgpixel */
			sx_set_transparent(GXcopy, ALL_ONE, bg);
			sx_plane_select(planemask);
			_getandputbits((psrc + endinc), (rem & PIM),
					(x & PIM), w, pdst, BitOff);

			if((x & PIM) + w >= 0x20)
			    pdst++;
		    }
		    else if(((x & PIM) + w) < PPW)
		    {
			/* doing < PPW bits is easy, and worth special-casing */

			/* draw fgpixel */
			sx_set_transparent(GXcopy, ALL_ONE, fg);
			sx_plane_select(planemask);
			_putbits(*psrc, x & PIM, w, pdst);

			/* draw bgpixel */
			sx_set_transparent(GXcopy, ALL_ONE, bg);
			sx_plane_select(planemask);
			_putbits(~(*psrc), x & PIM, w, pdst);
		    }
		    else
		    {
			int *tmpsrc, *tmpdst;

			/* start at the left edge of the tile,
			   and put down as much as we can
			*/
			maskbits(x, w, startmask, endmask, nlw);

			if (startmask)
			    nstart = PPW - (x & PIM);
			else
			    nstart = 0;
			if (endmask)
			    nend = (x + w)  & PIM;
			else
			    nend = 0;

			srcStartOver = nstart > PLST;

			tmpsrc = psrc;
			tmpdst = pdst;

			/* draw fgpixel */
			sx_set_transparent(GXcopy, ALL_ONE, fg);
			sx_plane_select(planemask);
			if(startmask)
			{
			    _putbits(*psrc, (x & PIM), nstart, pdst);
			    pdst++;
			    if(srcStartOver)
				psrc++;
			}
 			nlMiddle = nlw;
			while(nlMiddle--)
			{
			    _getandputbits0(psrc, nstart, PPW, pdst, BitOn);
			    pdst++;
			    psrc++;
			}
			if(endmask)
			{
			    _getandputbits0(psrc, nstart, nend, pdst, BitOn);
			}

			/* draw bgpixel */
			psrc = tmpsrc;
			pdst = tmpdst;
			sx_set_transparent(GXcopy, ALL_ONE, bg);
			sx_plane_select(planemask);
			if(startmask)
			{
			    _putbits(~(*psrc), (x & PIM), nstart, pdst);
			    pdst++;
			    if(srcStartOver)
				psrc++;
			}
			nlMiddle = nlw;
			while(nlMiddle--)
			{
			    _getandputbits0(psrc, nstart, PPW, pdst, BitOff);
			    pdst++;
			    psrc++;
			}
			if(endmask)
			{
			    _getandputbits0(psrc, nstart, nend, pdst, BitOff);
			}
 		    }
		    x += w;
		    width -= w;
		}
	    }
	    ppt++;
	    pwidth++;
	}
    }
    else
    {
	int plane, dstpljmp;
	OmronFbInfoPtr fbInfo;

	addrlBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	nlwidth = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;
	dstpljmp = nlwidth * pDrawable->height;
	fbInfo = getOmronFbInfo(pDrawable);
	if(isCommonPlane( addrlBase, fbInfo ))
	{
	    addrlBase = (int *)getPlaneAddres(fbInfo);
	    sx_set_operation_all( GXcopy, ~0 );
	}

	for (plane = 0; plane < depth; plane++)
	{
	    cfbfgmk[plane] = ((fg >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	    cfbbgmk[plane] = ((bg >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	}

	while (n--)
	{
	    iline = (ppt->y - ySrc) % stippleHeight;
	    pdst = addrlBase + (ppt->y * nlwidth) + (ppt->x >> PWSH);
	    psrcT = (int *) pStipple->devPrivate.ptr + (iline * stwidth);
	    x = ppt->x;

	    if (*pwidth)
	    {
		width = *pwidth;
		while(width > 0)
		{
		    psrc = psrcT;
		    w = min(stippleWidth, width);
		    if((rem = (x - xSrc) % stippleWidth) != 0)
		    {
			/* if we're in the middle of the tile, get
   			   as many bits as will finish the span, or
			   as many as will get to the left edge of the tile,
			   or a longword worth, starting at the appropriate
			   offset in the tile.
			*/
			w = min(min(stippleWidth - rem, width), BITMAP_SCANLINE_UNIT);
			endinc = rem / BITMAP_SCANLINE_UNIT;
			opaqgandpbits((psrc + endinc), (rem & PIM), (x & PIM),
							w, pdst, dstpljmp, depth);
			if((x & PIM) + w >= 0x20)
			    pdst++;
		    }
		    else if(((x & PIM) + w) < PPW)
		    {
			/* doing < PPW bits is easy, and worth special-casing */
			opaqpbits(*psrc, x & PIM, w, pdst, dstpljmp, depth);
		    }
		    else
		    {
			/* start at the left edge of the tile,
			   and put down as much as we can
			*/
			maskbits(x, w, startmask, endmask, nlMiddle);

			if (startmask)
			    nstart = PPW - (x & PIM);
			else
			    nstart = 0;
			if (endmask)
			    nend = (x + w)  & PIM;
			else
			    nend = 0;

			srcStartOver = nstart > PLST;

			if(startmask)
			{
			    opaqpbits(*psrc, (x & PIM), nstart, pdst,
							dstpljmp, depth);
			    pdst++;
			    if(srcStartOver)
				psrc++;
			}
 
			while(nlMiddle--)
			{
			    opaqgandpbits0(psrc, nstart, PPW, pdst,
							dstpljmp, depth);
			    pdst++;
			    psrc++;
			}
			if(endmask)
			{
			    opaqgandpbits0(psrc, nstart, nend, pdst,
							dstpljmp, depth);
			}
 		    }
		    x += w;
		    width -= w;
		}
	    }
	    ppt++;
	    pwidth++;
	}
    }
    DEALLOCATE_LOCAL(pptFree);
    DEALLOCATE_LOCAL(pwidthFree);
}
