/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

/*
 * Fill odd tiled and stippled(opaquestippled) rectangles.
 */

#include "X.h"
#include "Xmd.h"
#include "servermd.h"
#include "gcstruct.h"
#include "window.h"
#include "pixmapstr.h"
#include "scrnintstr.h"
#include "windowstr.h"

#include "cfb.h"
#include "cfbmskbits.h"

#define NextPatBits(pattern) \
{ \
    srcy++; \
    if (srcy == minheight) \
    { \
	srcy = 0; \
	psrc = (pattern); \
    } \
    else \
    { \
	psrc++; \
    } \
}

#define BOXTILEODD(ROP) \
    while (nBox--) \
    { \
	width = pBox->x2 - pBox->x1; \
	height = pBox->y2 - pBox->y1; \
	modulus((pBox->y1 - ySrc), tileHeight, iline); \
	psrcT = (int *) pTile->devPrivate.ptr; \
	pdstT = addrlBase + (pBox->y1 * nlwidth); \
	x = pBox->x1; \
	minheight = min(height, tileHeight); \
	\
	while(width > 0) \
	{ \
	    Pdst = pdstT + (x >> PWSH); \
	    Psrc = psrcT; \
	    modulus((x - xSrc), tileWidth, rem); \
	    w = min((tileWidth - rem), width); \
	    w = min(w, PPW - (x & PIM)); \
	    srcx = (rem & PIM); \
	    srcinc = rem >> PWSH; \
	    maskpartialbits((x & PIM), w, mask); \
	    for (plane = 0; plane < depth; plane++) \
	    { \
		if (cfbplmk[plane]) \
		{ \
		    srcy = iline; \
		    tilesrcp = Tilesrc; \
		    h = minheight; \
		    while(h--) \
		    { \
			psrc = Psrc + (srcy * tlwidth); \
			srcy = ++srcy < tileHeight ? srcy : 0; \
			getbits((psrc + srcinc), srcx, w, (*tilesrcp)); \
			*tilesrcp = SCRRIGHT(*tilesrcp, (x & PIM));  \
			tilesrcp++; \
		    } \
		    srcy = 0; h = height;  \
		    pdst = Pdst; psrc = Tilesrc; \
		    while(h--) \
		    { \
			*pdst = (*pdst & ~mask) | \
				(ROP(*psrc, *pdst) & mask); \
			NextPatBits(Tilesrc); \
			pdst += nlwidth; \
		    } \
		} \
		Pdst += dstpljmp; \
		Psrc += srcpljmp; \
	    } \
	    x += w; \
	    width -= w; \
	} \
	pBox++; \
    }

void
cfbFillBoxTileOdd (pDrawable, pGC, nBox, pBox, pTile, isCopy)
    DrawablePtr	    pDrawable;
    GCPtr	    pGC;
    register int    nBox;	/* number of boxes to fill */
    register BoxPtr pBox;	/* pointer to list of boxes to fill */
    PixmapPtr	    pTile;	/* pointer to tile we want to fill with */
    Bool	    isCopy;
{
    register int *psrc;		/* pointer to current word in tile */
    register int *pdst;		/* pointer to current word in bitmap */
    int *Pdst, *Psrc;
    int *psrcT, *pdstT;
    int *Tilesrc;		/* pointer to list of tile data */
    register int *tilesrcp;
    int *addrlBase;		/* pointer to start of bitmap */
    int tileWidth;		/* width of tile */
    int tileHeight;		/* height of tile */
    register int w, h, width;
    int height, minheight;
    int tlwidth;		/* width in longwords of tile */
    int nlwidth;		/* width in longwords of bitmap */
    int iline;			/* first line of tile to use */
    int xSrc, ySrc;
    int srcx;			/* current tile x position */
    register int srcy;		/* current tile y position */
    int x, srcinc, rem;
    int srcpljmp, dstpljmp;
    register int plane;
    int depth, rop;
    int patOrigx, patOrigy; 
    unsigned long planemask;
    unsigned int mask;
    OmronFbInfoPtr fbInfo;

    if (isCopy)
    {
	patOrigx = patOrigy = 0;
	planemask = ALL_ONE;
	rop = GXcopy;
    }
    else
    {
	if (pGC->alu == GXnoop || pGC->planemask == 0)
	    return;

	patOrigx = pGC->patOrg.x;
	patOrigy = pGC->patOrg.y;
	planemask = pGC->planemask;
	rop = pGC->alu;
    }

    depth = pDrawable->depth;

    tlwidth = pTile->devKind >> 2;
    tileWidth = pTile->drawable.width;
    tileHeight = pTile->drawable.height;
    srcpljmp = tlwidth * tileHeight;

    if (!(Tilesrc = (int *)ALLOCATE_LOCAL(tileHeight * sizeof(int))))
	return;

    fbInfo = getOmronFbInfo(pDrawable);

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	addrlBase = (int *)getPlaneAddres(fbInfo);
	nlwidth = (int)
	    (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;
	dstpljmp = getFbPlaneLongSize();

	xSrc = pDrawable->x + patOrigx;
	ySrc = pDrawable->y + patOrigy;

	sx_plane_select(planemask);

	while (nBox--)
	{
	    width = pBox->x2 - pBox->x1;
	    height = pBox->y2 - pBox->y1;
	    modulus((pBox->y1 - ySrc), tileHeight, iline);
	    psrcT = (int *) pTile->devPrivate.ptr;
	    pdstT = addrlBase + (pBox->y1 * nlwidth);
	    x = pBox->x1;
	    minheight = min(height, tileHeight);

	    while(width > 0)
	    {
		Pdst = pdstT + (x >> PWSH);
		Psrc = psrcT;
		modulus((x - xSrc), tileWidth, rem);
		w = min((tileWidth - rem), width);
		w = min(w, PPW - (x & PIM));
		srcx = (rem & PIM);
		srcinc = rem >> PWSH;
		maskpartialbits((x & PIM), w, mask);
		sx_set_op(rop, mask);
		for (plane = 0; plane < depth; plane++)
		{
		    if (cfbplmk[plane])
		    {
			srcy = iline;
			tilesrcp = Tilesrc;
			h = minheight;
			while(h--)
			{
			    psrc = Psrc + (srcy * tlwidth);
			    srcy = ++srcy < tileHeight ? srcy : 0;
			    getbits((psrc + srcinc), srcx, w, (*tilesrcp));
			    *tilesrcp = SCRRIGHT(*tilesrcp, (x & PIM)); 
			    tilesrcp++;
			}
			srcy = 0; h = height; 
			pdst = Pdst; psrc = Tilesrc;
			while(h--)
			{
			    *pdst = *psrc;
			    NextPatBits(Tilesrc);
			    pdst += nlwidth;
			}
		    }
		    Pdst += dstpljmp;
		    Psrc += srcpljmp;
		}
		x += w;
		width -= w;
	    }
	    pBox++;
	}
    }
    else
    {
	addrlBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	nlwidth = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;

	if(isCommonPlane( addrlBase, fbInfo ))
	    addrlBase = (int *)getPlaneAddres(fbInfo);

	dstpljmp = nlwidth * pDrawable->height;

	xSrc = patOrigx;
	ySrc = patOrigy;

	for (plane = 0; plane < depth; plane++)
	{
	    cfbplmk[plane] = (planemask >> plane) & 0x1;
	}

	switch(rop)
	{
	    case GXcopy:
		BOXTILEODD(fnCOPY);
		break;
	    case GXxor:
		BOXTILEODD(fnXOR);
		break;
	    case GXand:
		BOXTILEODD(fnAND);
		break;
	    case GXor:
		BOXTILEODD(fnOR);
		break;
	    case GXset:
		BOXTILEODD(fnSET);
		break;
	    case GXclear:
		BOXTILEODD(fnCLEAR);
		break;
	    case GXandReverse:
		BOXTILEODD(fnANDREVERSE);
		break;
	    case GXandInverted:
		BOXTILEODD(fnANDINVERTED);
		break;
	    case GXnor:
		BOXTILEODD(fnNOR);
		break;
	    case GXequiv:
		BOXTILEODD(fnEQUIV);
		break;
	    case GXinvert:
		BOXTILEODD(fnINVERT);
		break;
	    case GXorReverse:
		BOXTILEODD(fnORREVERSE);
		break;
	    case GXcopyInverted:
		BOXTILEODD(fnCOPYINVERTED);
		break;
	    case GXorInverted:
		BOXTILEODD(fnORINVERTED);
		break;
	    case GXnand:
		BOXTILEODD(fnNAND);
		break;
	    default:
		break;
	}
    }
    DEALLOCATE_LOCAL(Tilesrc);
}

#define BOXSTIPODD(ROP) \
    while (nBox--) \
    { \
	width = pBox->x2 - pBox->x1; \
	height = pBox->y2 - pBox->y1; \
	modulus((pBox->y1 - ySrc), stippleHeight, iline); \
	psrcT = (int *) pStipple->devPrivate.ptr; \
	pdstT = addrlBase + (pBox->y1 * nlwidth); \
	x = pBox->x1; \
	minheight = min(height, stippleHeight); \
	\
	while(width > 0) \
	{ \
	    Pdst = pdstT + (x >> PWSH); \
	    modulus((x - xSrc), stippleWidth, rem); \
	    w = min((stippleWidth - rem), width); \
	    w = min(w, PPW - (x & PIM)); \
	    srcx = (rem & PIM); \
	    srcy = iline; \
	    srcinc = rem >> PWSH; \
	    stpsrcp = Stpsrc; \
	    h = minheight; \
	    maskpartialbits((x & PIM), w, mask); \
	    while(h--) \
	    { \
		psrc = psrcT + (srcy * stwidth); \
		srcy = ++srcy < stippleHeight ? srcy : 0; \
		getbits((psrc + srcinc), srcx, w, (*stpsrcp)); \
		*stpsrcp = SCRRIGHT(*stpsrcp, (x & PIM));  \
		stpsrcp++; \
	    } \
	    for (plane = 0; plane < depth; plane++) \
	    { \
		if (cfbplmk[plane]) \
		{ \
		    srcy = 0; h = height;  \
		    pdst = Pdst; psrc = Stpsrc; \
		    while(h--) \
		    { \
			tmpbits = *psrc & cfbfgmk[plane]; \
			*pdst = (*pdst & (~mask | ~(*psrc))) | \
				(ROP(tmpbits, *pdst) & (mask & *psrc)); \
			NextPatBits(Stpsrc); \
			pdst += nlwidth; \
		    } \
		} \
		Pdst += dstpljmp; \
	    } \
	    x += w; \
	    width -= w; \
	} \
	pBox++; \
    }

void
cfbFillBoxStippleOdd (pDrawable, pGC, nBox, pBox, pStipple)
    DrawablePtr	    pDrawable;
    GCPtr	    pGC;
    register int    nBox;	/* number of boxes to fill */
    register BoxPtr pBox;	/* pointer to list of boxes to fill */
    PixmapPtr	    pStipple;	/* pointer to stipple we want to fill with */
{
    register int *psrc;		/* pointer to current word in stipple */
    register int *pdst;		/* pointer to current word in bitmap */
    int *psrcT, *pdstT;
    int *Stpsrc;		/* pointer to list of stipple data */
    register int *stpsrcp;
    int *addrlBase;		/* pointer to start of bitmap */
    int stippleWidth;		/* width of stipple */
    int stippleHeight;		/* height of stipple */
    register int w, h, width;
    int height, minheight;
    int stwidth;		/* width in longwords of stipple */
    int nlwidth;		/* width in longwords of bitmap */
    int iline;			/* first line of stipple to use */
    int xSrc, ySrc;
    int srcx;			/* current stipple x position */
    register int srcy;		/* current stipple y position */
    int x, srcinc, rem;
    int rop;
    int depth;
    unsigned long planemask;
    unsigned long pixel;
    unsigned int mask;


    if (pGC->alu == GXnoop || pGC->planemask == 0)
	return;

    if (pStipple->drawable.depth != 1) {
	FatalError( "Stipple depth not equal to 1!\n" );
    }

    rop = pGC->alu;
    depth = pDrawable->depth;
    planemask = pGC->planemask;
    pixel = pGC->fgPixel;

    stwidth = pStipple->devKind >> 2;
    stippleWidth = pStipple->drawable.width;
    stippleHeight = pStipple->drawable.height;

    if (!(Stpsrc = (int *)ALLOCATE_LOCAL(stippleHeight * sizeof(int))))
	return;

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	addrlBase = (int *)
	    (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devPrivate.ptr);
	nlwidth = (int)
	    (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;

	xSrc = pDrawable->x + pGC->patOrg.x;
	ySrc = pDrawable->y + pGC->patOrg.y;

	sx_check_transparent(rop, ALL_ONE, pixel, planemask);
	sx_plane_select(planemask);

	while (nBox--)
	{
	    width = pBox->x2 - pBox->x1;
	    height = pBox->y2 - pBox->y1;
	    modulus((pBox->y1 - ySrc), stippleHeight, iline);
	    psrcT = (int *) pStipple->devPrivate.ptr;
	    pdstT = addrlBase + (pBox->y1 * nlwidth);
	    x = pBox->x1;
	    minheight = min(height, stippleHeight);

	    while(width > 0)
	    {
		pdst = pdstT + (x >> PWSH);
		modulus((x - xSrc), stippleWidth, rem);
		w = min((stippleWidth - rem), width);
		w = min(w, PPW - (x & PIM));
		srcx = (rem & PIM);
		srcy = iline;
		srcinc = rem >> PWSH;
		stpsrcp = Stpsrc;
		h = minheight;
		maskpartialbits((x & PIM), w, mask);
		while(h--)
		{
		    psrc = psrcT + (srcy * stwidth);
		    srcy = ++srcy < stippleHeight ? srcy : 0;
		    getbits((psrc + srcinc), srcx, w, (*stpsrcp));
		    *stpsrcp = SCRRIGHT(*stpsrcp, (x & PIM)) & mask; 
		    stpsrcp++;
		}
		srcy = 0; h = height; 
		psrc = Stpsrc;
		while(h--)
		{
		    *pdst = *psrc;
		    NextPatBits(Stpsrc);
		    pdst += nlwidth;
		}
		x += w;
		width -= w;
	    }
	    pBox++;
	}
    }
    else
    {
	register int plane;
	int dstpljmp;
	int *Pdst;
	unsigned int tmpbits;
	OmronFbInfoPtr fbInfo;

	addrlBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	nlwidth = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;

	fbInfo = getOmronFbInfo(pDrawable);
	if(isCommonPlane( addrlBase, fbInfo ))
	    addrlBase = (int *) fbInfo->plane;

	dstpljmp = nlwidth * pDrawable->height;

	xSrc = pGC->patOrg.x;
	ySrc = pGC->patOrg.y;

	for (plane = 0; plane < depth; plane++)
	{
	    cfbfgmk[plane] = ((pixel >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	    cfbplmk[plane] = (planemask >> plane) & 0x1;
	}

	switch(rop)
	{
	    case GXcopy:
		BOXSTIPODD(fnCOPY);
		break;
	    case GXxor:
		BOXSTIPODD(fnXOR);
		break;
	    case GXand:
		BOXSTIPODD(fnAND);
		break;
	    case GXor:
		BOXSTIPODD(fnOR);
		break;
	    case GXset:
		BOXSTIPODD(fnSET);
		break;
	    case GXclear:
		BOXSTIPODD(fnCLEAR);
		break;
	    case GXandReverse:
		BOXSTIPODD(fnANDREVERSE);
		break;
	    case GXandInverted:
		BOXSTIPODD(fnANDINVERTED);
		break;
	    case GXnor:
		BOXSTIPODD(fnNOR);
		break;
	    case GXequiv:
		BOXSTIPODD(fnEQUIV);
		break;
	    case GXinvert:
		BOXSTIPODD(fnINVERT);
		break;
	    case GXorReverse:
		BOXSTIPODD(fnORREVERSE);
		break;
	    case GXcopyInverted:
		BOXSTIPODD(fnCOPYINVERTED);
		break;
	    case GXorInverted:
		BOXSTIPODD(fnORINVERTED);
		break;
	    case GXnand:
		BOXSTIPODD(fnNAND);
		break;
	    default:
		break;
	}
    }
    DEALLOCATE_LOCAL(Stpsrc);
}

void
cfbFillBoxOpStipOddCopy (pDrawable, pGC, nBox, pBox, pStipple)
    DrawablePtr	    pDrawable;
    GCPtr	    pGC;
    register int    nBox;	/* number of boxes to fill */
    register BoxPtr pBox;	/* pointer to list of boxes to fill */
    PixmapPtr	    pStipple;	/* pointer to stipple we want to fill with */
{
    register int *psrc;		/* pointer to current word in stipple */
    register int *pdst;		/* pointer to current word in bitmap */
    int *Pdst;
    int *psrcT, *pdstT;
    int *Stpsrc;		/* pointer to list of stipple data */
    register int *stpsrcp;
    int *addrlBase;		/* pointer to start of bitmap */
    int stippleWidth;		/* width of stipple */
    int stippleHeight;		/* height of stipple */
    register int w, h, width;
    int height, minheight;
    int stwidth;		/* width in longwords of stipple */
    int nlwidth;		/* width in longwords of bitmap */
    int iline;			/* first line of stipple to use */
    int xSrc, ySrc;
    int srcx;			/* current stipple x position */
    register int srcy;		/* current stipple y position */
    int x, srcinc, rem;
    int depth, rop;
    unsigned long planemask;
    unsigned long fg, bg;
    unsigned int mask;


    if (pGC->planemask == 0)
	return;

    if (pStipple->drawable.depth != 1) {
	FatalError( "OpaqueStipple depth not equal to 1!\n" );
    }

    depth = pDrawable->depth;
    planemask = pGC->planemask;
    fg = pGC->fgPixel;
    bg = pGC->bgPixel;

    stwidth = pStipple->devKind >> 2;
    stippleWidth = pStipple->drawable.width;
    stippleHeight = pStipple->drawable.height;

    if (!(Stpsrc = (int *)ALLOCATE_LOCAL(stippleHeight * sizeof(int))))
	return;

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	addrlBase = (int *)
	    (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devPrivate.ptr);
	nlwidth = (int)
	    (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;

	xSrc = pDrawable->x + pGC->patOrg.x;
	ySrc = pDrawable->y + pGC->patOrg.y;
	rop = GXcopy;

	while (nBox--)
	{
	    width = pBox->x2 - pBox->x1;
	    height = pBox->y2 - pBox->y1;
	    modulus((pBox->y1 - ySrc), stippleHeight, iline);
	    psrcT = (int *) pStipple->devPrivate.ptr;
	    pdstT = addrlBase + (pBox->y1 * nlwidth);
	    x = pBox->x1;
	    minheight = min(height, stippleHeight);

	    while(width > 0)
	    {
		Pdst = pdstT + (x >> PWSH);
		modulus((x - xSrc), stippleWidth, rem);
		w = min((stippleWidth - rem), width);
		w = min(w, PPW - (x & PIM));
		srcx = (rem & PIM);
		srcy = iline;
		srcinc = rem >> PWSH;
		stpsrcp = Stpsrc;
		h = minheight;
		maskpartialbits((x & PIM), w, mask);
		while(h--)
		{
		    psrc = psrcT + (srcy * stwidth);
		    srcy = ++srcy < stippleHeight ? srcy : 0;
		    getbits((psrc + srcinc), srcx, w, (*stpsrcp));
		    *stpsrcp = SCRRIGHT(*stpsrcp, (x & PIM)) & mask; 
		    stpsrcp++;
		}

		/* draw fgpixel */
		srcy = 0; h = height; 
		psrc = Stpsrc; pdst = Pdst;
		sx_set_transparent(rop, ALL_ONE, fg);
		sx_plane_select(planemask);
		while(h--)
		{
		    *pdst = *psrc;
		    NextPatBits(Stpsrc);
		    pdst += nlwidth;
		}

		/* draw bgpixel */
		srcy = 0; h = height; 
		psrc = Stpsrc; pdst = Pdst;
		sx_set_transparent(rop, ALL_ONE, bg);
		sx_plane_select(planemask);
		while(h--)
		{
		    *pdst = (~(*psrc) & mask);
		    NextPatBits(Stpsrc);
		    pdst += nlwidth;
		}
		x += w;
		width -= w;
	    }
	    pBox++;
	}
    }
    else
    {
	register int plane;
	int dstpljmp;
	unsigned int tmpbits;
	OmronFbInfoPtr fbInfo;

	addrlBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	nlwidth = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;

	fbInfo = getOmronFbInfo(pDrawable);
	if(isCommonPlane( addrlBase, fbInfo ))
	    addrlBase = (int *) fbInfo->plane;

	dstpljmp = nlwidth * pDrawable->height;

	xSrc = pGC->patOrg.x;
	ySrc = pGC->patOrg.y;

	for (plane = 0; plane < depth; plane++)
	{
	    cfbfgmk[plane] = ((fg >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	    cfbbgmk[plane] = ((bg >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	    cfbplmk[plane] = (planemask >> plane) & 0x1;
	}

	while (nBox--)
	{
	    width = pBox->x2 - pBox->x1;
	    height = pBox->y2 - pBox->y1;
	    modulus((pBox->y1 - ySrc), stippleHeight, iline);
	    psrcT = (int *) pStipple->devPrivate.ptr;
	    pdstT = addrlBase + (pBox->y1 * nlwidth);
	    x = pBox->x1;
	    minheight = min(height, stippleHeight);

	    while(width > 0)
	    {
		Pdst = pdstT + (x >> PWSH);
		modulus((x - xSrc), stippleWidth, rem);
		w = min((stippleWidth - rem), width);
		w = min(w, PPW - (x & PIM));
		srcx = (rem & PIM);
		srcy = iline;
		srcinc = rem >> PWSH;
		stpsrcp = Stpsrc;
		h = minheight;
		maskpartialbits((x & PIM), w, mask);
		while(h--)
		{
		    psrc = psrcT + (srcy * stwidth);
		    srcy = ++srcy < stippleHeight ? srcy : 0;
		    getbits((psrc + srcinc), srcx, w, (*stpsrcp));
		    *stpsrcp = SCRRIGHT(*stpsrcp, (x & PIM)); 
		    stpsrcp++;
		}
		for (plane = 0; plane < depth; plane++)
		{
		    if (cfbplmk[plane])
		    {
			srcy = 0; h = height; 
			pdst = Pdst; psrc = Stpsrc;
			while(h--)
			{
			    tmpbits = (*psrc & cfbfgmk[plane]) |
					(~(*psrc) & cfbbgmk[plane]);
			    *pdst = (*pdst & ~mask) | (tmpbits & mask);
			    NextPatBits(Stpsrc);
			    pdst += nlwidth;
			}
		    }
		    Pdst += dstpljmp;
		}
		x += w;
		width -= w;
	    }
	    pBox++;
	}
    }
    DEALLOCATE_LOCAL(Stpsrc);
}

#define BOXOPSTIPODD(ROP) \
    while (nBox--) \
    { \
	width = pBox->x2 - pBox->x1; \
	height = pBox->y2 - pBox->y1; \
	modulus((pBox->y1 - ySrc), stippleHeight, iline); \
	psrcT = (int *) pStipple->devPrivate.ptr; \
	pdstT = addrlBase + (pBox->y1 * nlwidth); \
	x = pBox->x1; \
	minheight = min(height, stippleHeight); \
	\
	while(width > 0) \
	{ \
	    Pdst = pdstT + (x >> PWSH); \
	    modulus((x - xSrc), stippleWidth, rem); \
	    w = min((stippleWidth - rem), width); \
	    w = min(w, PPW - (x & PIM)); \
	    srcx = (rem & PIM); \
	    srcy = iline; \
	    srcinc = rem >> PWSH; \
	    stpsrcp = Stpsrc; \
	    h = minheight; \
	    maskpartialbits((x & PIM), w, mask); \
	    while(h--) \
	    { \
		psrc = psrcT + (srcy * stwidth); \
		srcy = ++srcy < stippleHeight ? srcy : 0; \
		getbits((psrc + srcinc), srcx, w, (*stpsrcp)); \
		*stpsrcp = SCRRIGHT(*stpsrcp, (x & PIM));  \
		stpsrcp++; \
	    } \
	    for (plane = 0; plane < depth; plane++) \
	    { \
		if (cfbplmk[plane]) \
		{ \
		    srcy = 0; h = height;  \
		    pdst = Pdst; psrc = Stpsrc; \
		    while(h--) \
		    { \
			tmpbits = (*psrc & cfbfgmk[plane]) | \
				(~(*psrc) & cfbbgmk[plane]); \
			*pdst = (*pdst & ~mask) | \
				(ROP(tmpbits, *pdst) & mask); \
			NextPatBits(Stpsrc); \
			pdst += nlwidth; \
		    } \
		} \
		Pdst += dstpljmp; \
	    } \
	    x += w; \
	    width -= w; \
	} \
	pBox++; \
    }

#define compute_1toN_operation( alu, fg, bg, pm1, pm2, pm3, pm4, pmask) \
{ \
	unsigned long diff = fg ^ bg; \
	unsigned long same = ~diff; \
	pm1 = same & (~fg); \
	pm2 = diff & bg; \
	pm3 = diff & fg; \
	pm4 = same & fg; \
	if(conv_op_table[alu] == GXnoop) \
		pmask &= ~pm1; \
	if(invert_op_table[alu] == GXnoop) \
		pmask &= ~pm2; \
	if(const_src_op_table[alu] == GXnoop) \
		pmask &= ~pm4; \
}

#define	sx_set_1toN_operation(alu, mask, pm1, pm2, pm3, pm4, pmask) \
{ \
	sx_plane_select(pm1); \
	*sx_conv_op_cont0[alu] = mask; \
	sx_plane_select(pm2); \
	*sx_invert_op_cont0[alu] = mask; \
	sx_plane_select(pm3); \
	*sx_op_cont0[alu] = mask; \
	sx_plane_select(pm4); \
	*sx_const_src_op_cont0[alu] = mask; \
	sx_plane_select(pmask); \
}

void
cfbFillBoxOpStipOdd (pDrawable, pGC, nBox, pBox, pStipple)
    DrawablePtr	    pDrawable;
    GCPtr	    pGC;
    register int    nBox;	/* number of boxes to fill */
    register BoxPtr pBox;	/* pointer to list of boxes to fill */
    PixmapPtr	    pStipple;	/* pointer to stipple we want to fill with */
{
    register int *psrc;		/* pointer to current word in stipple */
    register int *pdst;		/* pointer to current word in bitmap */
    int *psrcT, *pdstT;
    int *Stpsrc;		/* pointer to list of stipple data */
    register int *stpsrcp;
    int *Pdst;
    int *addrlBase;		/* pointer to start of bitmap */
    int stippleWidth;		/* width of stipple */
    int stippleHeight;		/* height of stipple */
    register int w, h, width;
    int height, minheight;
    int stwidth;		/* width in longwords of stipple */
    int nlwidth;		/* width in longwords of bitmap */
    int iline;			/* first line of stipple to use */
    int xSrc, ySrc;
    int srcx;			/* current stipple x position */
    register int srcy;		/* current stipple y position */
    int x, srcinc, rem;
    int rop;
    int depth;
    int drawFg, drawBg;
    unsigned long planemask;
    unsigned long pm1, pm2, pm3, pm4;
    unsigned long fg, bg;
    unsigned int mask;
    unsigned int oldmask = 0;


    if (pGC->alu == GXnoop || pGC->planemask == 0)
	return;

    if (pStipple->drawable.depth != 1) {
	FatalError( "OpaqueStipple depth not equal to 1!\n" );
    }

    rop = pGC->alu;
    depth = pDrawable->depth;
    planemask = pGC->planemask;
    fg = pGC->fgPixel;
    bg = pGC->bgPixel;

    stwidth = pStipple->devKind >> 2;
    stippleWidth = pStipple->drawable.width;
    stippleHeight = pStipple->drawable.height;

    if (!(Stpsrc = (int *)ALLOCATE_LOCAL(stippleHeight * sizeof(int))))
	return;

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	addrlBase = (int *)
	    (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devPrivate.ptr);
	nlwidth = (int)
	    (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;

	xSrc = pDrawable->x + pGC->patOrg.x;
	ySrc = pDrawable->y + pGC->patOrg.y;

	compute_1toN_operation( rop, fg, bg, pm1, pm2, pm3, pm4, planemask);

	while (nBox--)
	{
	    width = pBox->x2 - pBox->x1;
	    height = pBox->y2 - pBox->y1;
	    modulus((pBox->y1 - ySrc), stippleHeight, iline);
	    psrcT = (int *) pStipple->devPrivate.ptr;
	    pdstT = addrlBase + (pBox->y1 * nlwidth);
	    x = pBox->x1;
	    minheight = min(height, stippleHeight);

	    while(width > 0)
	    {
		Pdst = pdstT + (x >> PWSH);
		modulus((x - xSrc), stippleWidth, rem);
		w = min((stippleWidth - rem), width);
		w = min(w, PPW - (x & PIM));
		srcx = (rem & PIM);
		srcy = iline;
		srcinc = rem >> PWSH;
		stpsrcp = Stpsrc;
		h = minheight;
		maskpartialbits((x & PIM), w, mask);
		while(h--)
		{
		    psrc = psrcT + (srcy * stwidth);
		    srcy = ++srcy < stippleHeight ? srcy : 0;
		    getbits((psrc + srcinc), srcx, w, (*stpsrcp));
		    *stpsrcp = SCRRIGHT(*stpsrcp, (x & PIM)) & mask; 
		    stpsrcp++;
		}
		srcy = 0; h = height; 
		psrc = Stpsrc; pdst = Pdst;
		if(oldmask != mask)
		{
		    sx_set_1toN_operation(rop, mask, pm1, pm2, pm3, pm4,
					       planemask);
		    sx_plane_select(planemask);
		    oldmask = mask;
		}
		while(h--)
		{
		    *pdst = *psrc;
		    NextPatBits(Stpsrc);
		    pdst += nlwidth;
		}
		x += w;
		width -= w;
	    }
	    pBox++;
	}
    }
    else
    {
	register int plane;
	int dstpljmp;
	int *Pdst;
	unsigned int tmpbits;
	OmronFbInfoPtr fbInfo;

	addrlBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	nlwidth = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;

	fbInfo = getOmronFbInfo(pDrawable);
	if(isCommonPlane( addrlBase, fbInfo ))
	    addrlBase = (int *) fbInfo->plane;

	dstpljmp = nlwidth * pDrawable->height;

	xSrc = pGC->patOrg.x;
	ySrc = pGC->patOrg.y;

	for (plane = 0; plane < depth; plane++)
	{
	    cfbfgmk[plane] = ((fg >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	    cfbbgmk[plane] = ((bg >> plane) & 0x1) ? ALL_ONE : ALL_ZERO;
	}

	switch(rop)
	{
	    case GXxor:
		BOXOPSTIPODD(fnXOR);
		break;
	    case GXand:
		BOXOPSTIPODD(fnAND);
		break;
	    case GXor:
		BOXOPSTIPODD(fnOR);
		break;
	    case GXset:
		BOXOPSTIPODD(fnSET);
		break;
	    case GXclear:
		BOXOPSTIPODD(fnCLEAR);
		break;
	    case GXandReverse:
		BOXOPSTIPODD(fnANDREVERSE);
		break;
	    case GXandInverted:
		BOXOPSTIPODD(fnANDINVERTED);
		break;
	    case GXnor:
		BOXOPSTIPODD(fnNOR);
		break;
	    case GXequiv:
		BOXOPSTIPODD(fnEQUIV);
		break;
	    case GXinvert:
		BOXOPSTIPODD(fnINVERT);
		break;
	    case GXorReverse:
		BOXOPSTIPODD(fnORREVERSE);
		break;
	    case GXcopyInverted:
		BOXOPSTIPODD(fnCOPYINVERTED);
		break;
	    case GXorInverted:
		BOXOPSTIPODD(fnORINVERTED);
		break;
	    case GXnand:
		BOXOPSTIPODD(fnNAND);
		break;
	    default:
		break;
	}
    }
    DEALLOCATE_LOCAL(Stpsrc);
}
