/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <sys/signal.h>
#ifdef	UniOSU
#include <sys/types.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/sys9100.h>
#else
#include <om68k/sysomron.h>
#endif /* UniOSU */

#include "omronFb.h"

#include "X.h"

#ifdef	UniOSU
#include "osdep.h"
#endif
#include "scrnintstr.h"
#include "servermd.h"
#include "misc.h"
#include "input.h"


extern int omronMouseProc();
extern int omronKbdProc();

extern int omronDtBmCreate();
extern int omronDtBmInit();

static Bool omronScreenInit();
static OmronFbProc omron_fb_proc;

static PixmapFormatRec	MonoFormats[] = {
    1, 1, BITMAP_SCANLINE_PAD,	/* 1-bit deep */
};

static PixmapFormatRec	FourColorFormats4[] = {
    1, 1, BITMAP_SCANLINE_PAD,	/* 1-bit deep */
    4, 4, BITMAP_SCANLINE_PAD,	/* 4-bit deep */
};

static PixmapFormatRec	FourColorFormats8[] = {
    1, 1, BITMAP_SCANLINE_PAD,	/* 1-bit deep */
    4, 8, BITMAP_SCANLINE_PAD,	/* 4-bit deep */
};

static PixmapFormatRec	EightColorFormats[] = {
    1, 1, BITMAP_SCANLINE_PAD,	/* 1-bit deep */
    8, 8, BITMAP_SCANLINE_PAD,	/* 8-bit deep */
};


unsigned long omronGeneration;
OmronFbInfo	omron_fb_info;	
int omronScreenIndex;
int omronBitsPerPixel = 0;

static int NonConsole;

int omronSigIO = 0;

int num_plane = 0;
int scrw = 0;
int scrh = 0;
int full_color = 0;
int QueryFb = 0;
int Version = 0;
char *fb_type = NULL;
int bitsperpixel = 0;


InitOutput(pScreenInfo, argc, argv)
ScreenInfo 	  *pScreenInfo;
int     	  argc;
char    	  **argv;
{
 	int	i;
	int	num_formats;
	PixmapFormatPtr     formats;
	
	if (!omronGetFbInfo(&omron_fb_info))
	{
		ErrorF("Can't get fb information.\n");
		exit(-1);
	}

	ParseFbArg(&omron_fb_info);
#ifdef MONO_ONLY
	omron_fb_info.fb_depth = 1;
#endif

	SetConsoleMode();

	switch(omron_fb_info.fb_depth) {
	case 1 :
		num_formats=1;
		formats=MonoFormats;
		break; 
	case 4 :
		num_formats=2;
		if (bitsperpixel != 0) {
			omronBitsPerPixel = 4;
			formats=FourColorFormats4;
		}
		else {
			formats=FourColorFormats8;
		}
		break; 
	case 8 :
		num_formats=2;
		formats=EightColorFormats;
		break; 
	default :
		break; 
	}

	pScreenInfo->imageByteOrder = IMAGE_BYTE_ORDER;
	pScreenInfo->bitmapScanlineUnit = BITMAP_SCANLINE_UNIT;
	pScreenInfo->bitmapScanlinePad = BITMAP_SCANLINE_PAD;
	pScreenInfo->bitmapBitOrder = BITMAP_BIT_ORDER;

	pScreenInfo->numPixmapFormats = num_formats;
	for (i=0; i< num_formats; i++)
		pScreenInfo->formats[i] = formats[i];
	
	if ( (* omron_fb_proc.CreateProc)(&omron_fb_info) < 0 )
		ErrorF("Can't Create Fb . \n");

	AddScreen(omronScreenInit,argc,argv);

}

ParseFbArg(omron_fb_info)
OmronFbInfoPtr omron_fb_info;
{
	if( QueryFb )
	{
		switch(omron_fb_info->fb_type) {
		case DT_BM :
			puts("      MACHINE  DT");
			puts("      GRAPHIC  BITMAP");
			break;
		case DT_PLASMA :
			puts("      MACHINE  DT");
			puts("      GRAPHIC  PLASMA");
			break;
		case FS_BM :
			puts("      MACHINE  FS");
			puts("      GRAPHIC  BITMAP");
			break;
		case DS_BM :
			puts("      MACHINE  DS");
			puts("      GRAPHIC  BITMAP");
			break;
		}
		printf(" SCREEN_WIDTH  %d\n",omron_fb_info->scr_width);
		printf("SCREEN_HEIGHT  %d\n",omron_fb_info->scr_height);
		printf(" SCREEN_DEPTH  %d\n",omron_fb_info->fb_depth);
		exit(0);
	} 
	if ( scrw )
		omron_fb_info->scr_width = scrw;  
	if ( scrh )
		omron_fb_info->scr_height = scrh;  

	if ( fb_type )
	{
		if( num_plane > 0 )
			omron_fb_info->fb_depth = num_plane;
		return;
	}

	if ( !full_color )
	{
		if( num_plane > 0 && num_plane <= omron_fb_info->fb_depth )
			omron_fb_info->fb_depth = num_plane;
		else
			omron_fb_info->fb_depth = 1;
	}
}


static Bool
omronScreenInit(screenIndex, pScreen, argc, argv)
int		screenIndex;
ScreenPtr	pScreen;
int		argc;
char 		**argv;
{
	if (omronGeneration != serverGeneration)
	{
		if((omronScreenIndex = AllocateScreenPrivateIndex()) <0){
			ErrorF("AllocateScreenPrivateIndex Error\n");
			return FALSE;
		}
		omronGeneration = serverGeneration;
	}
	pScreen->devPrivates[omronScreenIndex].ptr = (pointer)&omron_fb_info;
	return((* omron_fb_proc.InitProc)(screenIndex, pScreen, argc, argv));
}


SetConsoleMode()
{
	char *ttyname();
	char *p = ttyname(2);
	register int i; 

	if (p && strcmp(p,"/dev/console")!=0) 
	{
		NonConsole = TRUE;
		return;
	}

	i = fcntl(2, F_GETFL, 0); 
	if(i >=0)
		i = fcntl(2, F_SETFL, i | FNDELAY);
	if (i < 0) {
		Error("InitOutput:on console can't put stderr in non-block mode\n");
	}
}

ResetConsoleMode()
{
	register int i;

	if ( NonConsole )
	{
		return;
	}
	i = fcntl(2, F_GETFL, 0);
	if (i >= 0)
		(void) fcntl(2, F_SETFL, i & ~FNDELAY);
}

#if	!( defined(UniOSU) ||  defined(NO_SIGIO))
static void
SigIOHandler(sig, code, scp)
    int		code;
    int		sig;
    struct sigcontext *scp;
{
    omronSigIO = 1;
}
#endif


InitInput(argc, argv)
    int     	  argc;
    char    	  **argv;
{
    DevicePtr p, k;
    static int  zero = 0;
    
    p = AddInputDevice(omronMouseProc, TRUE);
    k = AddInputDevice(omronKbdProc, TRUE);

    RegisterPointerDevice(p);
    RegisterKeyboardDevice(k);
    miRegisterPointerDevice(screenInfo.screens[0], p);

#if	defined(UniOSU) || defined(NO_SIGIO)
    /* UniOSU device deriver does not support SIGIO signal */
#else
    signal(SIGIO, SigIOHandler);
#endif

    SetInputCheck (&zero, &omronSigIO);
}

omronGetFbInfo(omron_fb_info)
OmronFbInfoPtr omron_fb_info;
{
	int machine_type;

	if ( fb_type )
	{
		if (strcmp(fb_type, "DT_BM") == 0)
		{
			omron_fb_info->fb_type = DT_BM;
			omron_fb_proc.CreateProc = omronDtBmCreate;
			omron_fb_proc.InitProc = omronDtBmInit;
		} else if (strcmp(fb_type, "DT_PLASMA") == 0) {
			omron_fb_info->fb_type = DT_PLASMA;
			omron_fb_proc.CreateProc = omronDtBmCreate;
			omron_fb_proc.InitProc = omronDtBmInit;
		} else if (strcmp(fb_type, "FS_BM") == 0) {	
			omron_fb_info->fb_type = FS_BM;
		} else if (strcmp(fb_type, "DS_BM") == 0) {
			omron_fb_info->fb_type = DS_BM; 
		} else {
			ErrorF("Unknown fb type. \n");
			return (FALSE);
		}
		if (omron_fb_info->fb_type != DT_PLASMA)
		{
			omron_fb_info->scr_width = SCREEN_WIDTH;  
			omron_fb_info->scr_height = SCREEN_HEIGHT;  
		} else {
			omron_fb_info->scr_width = PLASMA_SCREEN_WIDTH;  
			omron_fb_info->scr_height = PLASMA_SCREEN_HEIGHT;  
		}
		omron_fb_info->fb_width = FB_WIDTH;  
		omron_fb_info->fb_height = FB_HEIGHT;  
		omron_fb_info->fb_depth = 1;
		return (TRUE);
	}
			

#ifdef	UniOSU
	if (sys9100(S91MACHTYPE,&machine_type) < 0)
	{
		ErrorF("sys9100 error.\n");
		return (FALSE);
	}
#else
	if (sysomron(S91MACHTYPE,&machine_type) < 0)
	{
		ErrorF("sysomron error.\n");
		return (FALSE);
	}
#endif

	switch(machine_type & (MACH_MACH | MACH_GRAPHIC_BOARD)) {
	case MACH_DT | MACH_BM : /* we call LUNA */ 
		omron_fb_info->fb_type = DT_BM;
		omron_fb_proc.CreateProc = omronDtBmCreate;
		omron_fb_proc.InitProc = omronDtBmInit;
		break;
	case MACH_DT | MACH_PLASMA  : /* LUNA support plasma display */  
		omron_fb_info->fb_type = DT_PLASMA;
		omron_fb_proc.CreateProc = omronDtBmCreate;
		omron_fb_proc.InitProc = omronDtBmInit;
		break;
	case MACH_DS | MACH_BM : /* we call Mr. */
		omron_fb_info->fb_type = DS_BM;
		break;
	case MACH_FS | MACH_BM : /* we call M90 */
		omron_fb_info->fb_type = FS_BM;
		break;
	default : 
		ErrorF("Can't support machine type. %d\n",machine_type);
		return (FALSE);
	}

	switch(machine_type & MACH_PLANE){
	case MACH_1_PLANE :
		omron_fb_info->fb_depth = 1;
		break;
	case MACH_4_PLANE :
		omron_fb_info->fb_depth = 4;
		break;
	case MACH_8_PLANE :
		omron_fb_info->fb_depth = 8;
		break;
	default :
		ErrorF("Can't support plane number.\n"); 
		return (FALSE);
	}

	if (omron_fb_info->fb_type != DT_PLASMA)
	{
		omron_fb_info->scr_width = SCREEN_WIDTH;  
		omron_fb_info->scr_height = SCREEN_HEIGHT;  
	} else {
		omron_fb_info->scr_width = PLASMA_SCREEN_WIDTH;  
		omron_fb_info->scr_height = PLASMA_SCREEN_HEIGHT;  
	}
	omron_fb_info->fb_width = FB_WIDTH;  
	omron_fb_info->fb_height = FB_HEIGHT;  

	return(TRUE);
}
